//*****************************************************************************
// File: java.cpp
// Description: Nirva Java client interface (c++ part)
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************


#include "nvcc.h"
#include "nvc.h"
#include "java.h"


using namespace std;

#define RESULT_BUFFER_LENGTH	4096



/////////////////////////////////////////////////////////////////////////////
// NvJavaRequestList

NvJavaRequestList::NvJavaRequestList()
{
	m_bUseChecking = true;
}

NvJavaRequestList::~NvJavaRequestList()
{
	Clear();
}



void NvJavaRequestList::Clear()
{
	Lock();
	map<void *,time_t>::iterator pos;
	for(pos = m_vRequests.begin(); pos != m_vRequests.end(); ++pos)
	{
		NvcJavaInt *Request = (NvcJavaInt *)(pos->first);
		if(Request != NULL)
			delete Request;
	}
	m_vRequests.clear();
	Unlock();
}



NvcJavaInt* NvJavaRequestList::AddRequest(int TimeOut)
{
	// Update the request list each time a new request is added
	Update();
	Lock();
	NvcJavaInt *Request = new NvcJavaInt;
	if(Request != NULL)
	{
		Request->m_timeout = TimeOut;
		if(Request->m_timeout < 0)
			Request->m_timeout = 0;
		m_vRequests[(void *)Request] = time(NULL);
	}
	Unlock();
	return Request;
}


void NvJavaRequestList::RemoveRequest(NvcJavaInt *Request)
{
	if(Request == NULL)
		return;

	Lock();
	if(m_vRequests.find((void *)Request) != m_vRequests.end())
	{
		m_vRequests.erase(Request);
		delete Request;
	}
	Unlock();
}


bool NvJavaRequestList::CheckRequest(NvcJavaInt *Request)
{
	if(!m_bUseChecking)
		return true;

	if(Request == NULL)
		return false;

	Lock();
	bool RequestFound = false;
	if(m_vRequests.find((void *)Request) != m_vRequests.end())
	{
		m_vRequests[(void *)Request] = time(NULL);
		Request->m_iInUse++;
		RequestFound = true;
	}
	Unlock();
	return RequestFound;
}

void NvJavaRequestList::FreeRequest(NvcJavaInt *Request)
{
	if(!m_bUseChecking)
		return;

	if(Request == NULL)
		return;

	Lock();
	if(m_vRequests.find((void *)Request) != m_vRequests.end())
	{
		m_vRequests[(void *)Request] = time(NULL);
		if(Request->m_iInUse > 0)
			Request->m_iInUse--;
	}
	Unlock();
}


void NvJavaRequestList::Update()
{
	if(!m_bUseChecking)
		return;
	int i;
	vector<void *> m_vRequestsToClear;

	Lock();
	time_t actual_time = time(NULL);
	
	map<void *,time_t>::iterator pos;
	for(pos = m_vRequests.begin(); pos != m_vRequests.end(); ++pos)
	{
		time_t request_time = pos->second;
		NvcJavaInt *Request = (NvcJavaInt *)(pos->first);
		if(Request == NULL)
			continue;
		if(Request->m_timeout == 0)
			continue;
		if(Request->m_iInUse > 0)
			continue;
		if((actual_time - request_time) > Request->m_timeout)
		{
			m_vRequestsToClear.push_back(pos->first);
			delete Request;
			continue;
		}
	}
	for(i=0; i<(int)m_vRequestsToClear.size(); i++)
		m_vRequests.erase(m_vRequestsToClear[i]);
	Unlock();
}




// Java request class

NvcJavaInt::NvcJavaInt()
{
	ResultBufferSize = 0;
	SetBufferSize(0);
	LastResult = "";
	m_pRequest = NULL;
	m_iInUse = 0;
	m_timeout = 0;
}


NvcJavaInt::~NvcJavaInt()
{
	if(ResultBufferSize)
		delete ResultBuffer;
	if(m_pRequest != NULL)
		NvCloseRequest(m_pRequest);
}

// Request open (the close is automatically made in the destructor)
bool NvcJavaInt::OpenRequest(const char *ConnectString)
{
	if(m_pRequest != NULL)
		return true;			// Ever opened

	m_pRequest = NvOpenRequest(ConnectString);

	if(m_pRequest == NULL)
		return false;


	return true;
}

int NvcJavaInt::Command(const char *CommandString)
{
	if(m_pRequest == NULL)
		return false;			// Not opened

	string CommandStringHere = CommandString;
	EncodeCommandInputString(ENCODING_UTF8, CommandStringHere, m_pRequest);
	strcpy(ResultBuffer, "");
	int DataSize;
	if(!NvCommand2(m_pRequest, CommandStringHere.c_str(), ResultBuffer, ResultBufferSize, &DataSize, ENCODING_UTF8))
		return 0;
	int ResultSetBufferSize = SetBufferSize(DataSize);
	if(ResultSetBufferSize == 0)
		return 0;	// Error
	if(ResultSetBufferSize == 2)
	{
		// We got a reallocation so we must get the last output buffer
		NvCommand2(m_pRequest, "NV_CMD=|LOCAL:GET_LAST_OUTPUT_BUFFER|", ResultBuffer, ResultBufferSize, &DataSize, ENCODING_UTF8);
//		if(!NvCommand2(m_pRequest, CommandString, ResultBuffer, ResultBufferSize, &DataSize, ENCODING_UTF8))
//			return 0;
	}

	LastResult = ResultBuffer;

	return 1;
}



int NvcJavaInt::SetBufferSize(int DataSize)
{
	if(DataSize < ResultBufferSize)
		return 1;

	if(ResultBufferSize == 0)
	{
		ResultBuffer = (char *)malloc(RESULT_BUFFER_LENGTH);
		if(ResultBuffer == NULL)
			return 0;
		ResultBufferSize = RESULT_BUFFER_LENGTH;
	}

	if(DataSize < ResultBufferSize)
		return 1;

	char *NewResultBuffer = (char *)realloc(ResultBuffer, DataSize+1);
	if(NewResultBuffer == NULL)
		return 0;

	ResultBuffer = NewResultBuffer;
	ResultBufferSize = DataSize+1;

	return 2;
}

