// ***************************************************************************************
// File: global.h
// Description: functions used in ejava.cpp and service.cpp
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************

#ifndef _GLOBAL_H
#define _GLOBAL_H

#include "jni.h"
#include "nvsext.h"
#include <string>
#include <vector>
#include <map>



#ifdef WIN32
#define NVTHREADID DWORD
#else
#define NVTHREADID pthread_t
#endif

#define RESULT_BUFFER_LENGTH	256
#ifdef WIN32
#define PATH_SEPARATOR ";"
#else
#define PATH_SEPARATOR ":"
#endif


int ParseMethodName(const char* MethodNameToParse, std::string& DirName, std::string& ClassName, std::string& MethodName, std::string& Type, std::string& TypeName);
bool MyAttachCurrentThread(void **JavaEnv);
void MyDetachCurrentThread();
void SysLog(const char *Message, const char *Extra = NULL, int Type = NV_LOG_INFO, int Level = 1);
int SysGetLogLevel();


/////////////////////////////////////////////////////////////////////////////
// NvJavaInt class
//
// This class encapsulates the thread context for the method
//
class NvJavaInt
{
public:
	std::vector<void *> m_mCommands;	// command stack (used for recursivity)
	void* m_pCommand;					// actual command
	std::string LastResult;							// last string result
	int InUse;									// Used for thread attachement
	
public:
	NvJavaInt();
	virtual ~NvJavaInt();
};

/////////////////////////////////////////////////////////////////////////////
// NvJavaThreadList class
//
// This class manages the correspondance between the thread handle and the NvJavaInt
// class that is the interface to Nirva
//
class NvJavaThreadList : public NvObject
{
public:
	std::map<NVTHREADID, NvJavaInt*> m_mList;	// list of threads using java
	
public:
	NvJavaThreadList();
	virtual ~NvJavaThreadList();

	// Remove all threads from the list
	void RemoveAll();				
	
	// Eventually attach the current thread and retuns a NvJavaInt class that owns the thread context
	NvJavaInt* GetThreadNvIntPtr(void* pCommand = NULL);
	
	// Eventually detach the current thread and remove the thread context
	void FreeThreadNvIntPtr(bool force = false);
};



/////////////////////////////////////////////////////////////////////////////
// NvJavaClassLoader class
//
// This class encapsulates a java class loader
//
class NvJavaClassLoader
{
public:
	jobject Loader;					// Java local class loader
	jmethodID msLoaderFindClass;	// Find class method of the loader
	
public:
	NvJavaClassLoader();
	virtual ~NvJavaClassLoader();

	bool Load(std::string DirName);
	void UnLoad();
};



/////////////////////////////////////////////////////////////////////////////
// NvJavaClassLoaderList class
//
// This class manages the JAVA calss loader list
//
class NvJavaClassLoaderList : public NvObject
{
public:
	std::map<std::string, NvJavaClassLoader*> m_mList;	// list of java environments
	
public:
	NvJavaClassLoaderList();
	virtual ~NvJavaClassLoaderList();

	// Remove all items from the list
	void RemoveAll();				
	
	// Get the NvJavaClassLoader
	NvJavaClassLoader* Get(std::string Type, std::string Name);
	
	// Get the NvJavaClassLoader
	void Add(std::string Type, std::string Name, NvJavaClassLoader* Loader);
	
	// Get or Add the NvJavaClassLoader
	NvJavaClassLoader* GetAdd(std::string Type, std::string Name, std::string DirName);
	
	// Remove the NvJavaClassLoader
	void Remove(std::string Type, std::string Name);
};




/////////////////////////////////////////////////////////////////////////////
// NvJavaClass class
//
// This class encapsulates a java class for the cache
//
class NvJavaClass
{
public:
	jclass TheClass;				// The java class
	jobject Loader;					// Java local class loader
	std::string TypeAndName;		// Type::Name
	std::string PathAndClass;		// DirName::ClassName 
	
public:
	NvJavaClass();
	virtual ~NvJavaClass();

	void UnLoad();
};



/////////////////////////////////////////////////////////////////////////////
// NvJavaClassCache class
//
// This class manages the JAVA class cache
//
class NvJavaClassCache : public NvObject
{
public:
	std::map<std::string, NvJavaClass*> m_mList;	// list of java classes index is Type::Name::DirName::ClassName
	
public:
	NvJavaClassCache();
	virtual ~NvJavaClassCache();

	// Remove all items from the list
	void RemoveAll();				
	
	// Get from cache (must be locked externally)
	NvJavaClass* Get(std::string Type, std::string Name, std::string DirName, std::string ClassName);
	
	// Add to cache (must be locked externally)
	NvJavaClass* Add(std::string Type, std::string Name, std::string DirName, std::string ClassName, jobject Loader, jclass TheClass);
	
	// Remove the cache for the given type and name 
	void Remove(std::string Type, std::string Name);
};



/////////////////////////////////////////////////////////////////////////////
// NvJavaEnv class
//
// This class encapsulates the java environment
//
class NvJavaEnv
{
public:
	NvJavaClassLoader* EnvLoader;		// java environment class loader
	jobject MyLoader;					// Java local class loader
	jmethodID msLoaderFindClass;		// Find class method of the loader
	jobject ClassToRun;					// Pointer to the class object
	std::string sClassToRun;			// Used for log
	jmethodID ms;						// Current method
	std::string CurrentMethod;			// Current method name
	bool bStaticMethod;					// Set to true if the method is static
	bool bOkOne;						// Set to true if method return value of 1 is OK
	jmethodID msRender;					// Render method
	bool bUseEnvClassLoader;			// Set to true if environment loader is to be used (default is false)
	bool bUseClassCache;				// Set to true if class cache is to be used (default is false)
	
public:
	NvJavaEnv();
	virtual ~NvJavaEnv();

	int LoadClass(std::string Type, std::string Name, std::string DirName, std::string ClassName);
	void UnLoadClass();
	int RunMethod(std::string Method, bool BooleanMethod, std::string& JavaException);
	bool DoRender(NvRendererInt *Render);

};

/////////////////////////////////////////////////////////////////////////////
// NvJavaEnvList class
//
// This class manages the JAVA environment list for services
//
class NvJavaEnvList : public NvObject
{
public:
	std::map<std::string, NvJavaEnv*> m_mList;	// list of java environments
	
public:
	NvJavaEnvList();
	virtual ~NvJavaEnvList();

	// Remove all items from the list
	void RemoveAll();				
	
	// Get the NvJavaEnv
	NvJavaEnv* GetJavaEnv(std::string Name);
	
	// Add the NvJavaEnv
	void AddJavaEnv(std::string Name, NvJavaEnv* JavaEnv);
	
	// Remove the NvJavaEnv
	void RemoveJavaEnv(std::string Name);
};


#endif //_GLOBAL_H
