// ***************************************************************************************
// File: container.cpp
// Description: Nirva containers management
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************


#ifdef NV_SERVER
#include "nvs.h"
#else
#include "../nvc/nvcc.h"
#endif

#ifdef NV_SERVER
	#ifdef WIN32
		#ifdef _DEBUG
			#include <stdlib.h>
			#include <crtdbg.h>
			#define NVNEW new( _NORMAL_BLOCK, __FILE__, __LINE__)
		#else // _DEBUG
			#define NVNEW new
		#endif // _DEBUG
	#else	//WIN32
		#define NVNEW new
	#endif	//WIN32
#else	//NV_SERVER
	#define NVNEW new
#endif	//NV_SERVER

using namespace std;

// PM TODO Modif app container en lock write only
// Make a create container command in system service
// allowing to create a lock write only container (for
// app, service or system containers only).
// Create a hash table for the commands allowing to
// get a container in write mode and consult this hash
// in session:GetContainer
// Create a container lock system command to get exclusive
// access to a container (and an unlock command also). This
// should work only with write only locked containers
// and for app, service or system containers)
// All the rest should be OK
// Write only locked containers can be only created from
// the new SYSTEM:CONTAINER:CREATE command

/////////////////////////////////////////////////////////////////////////////
// NvContainerList

NvContainerList::NvContainerList()
{
	m_bUseLocking = false;
	LastContainer = NULL;
	LastContainerName = "";
}

NvContainerList::~NvContainerList()
{
	Clear();
}



void NvContainerList::Clear()
{
	NV_LOCKOBJECT(this);
	map<string,void *>::iterator pos;
	for(pos = m_mContainers.begin(); pos != m_mContainers.end(); ++pos)
	{
		NvRootContainer *Container = (NvRootContainer *)(pos->second);
		if(Container != NULL)
			delete Container;
	}
	m_mContainers.clear();
	LastContainer = NULL;
	LastContainerName = "";
	NV_UNLOCKOBJECT(this);
}


NvContainer* NvContainerList::GetContainer(const char *Name, void *Session, bool Exclusive)
{
	if(Name == NULL)
		return NULL;

	string NameHere = Name;
	NameHere = CleanContainerName(NameHere);
	string OriginalNameHere = NameHere;

	NV_LOCKOBJECT(this);

	if(!m_bUseLocking)
	{
		// Last container cache mechanism
		if((LastContainer != NULL) && (LastContainerName == NameHere))
		{
			NvContainer *ReturnedContainer = LastContainer;
			NV_UNLOCKOBJECT(this);
			return ReturnedContainer;
		}
	}


	NvRootContainer *Container = GetRootContainer(NameHere.c_str());
	if(Container == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return NULL;
	}

	// We get the thread ID because the locking is based on threads
	long ThreadId = NvGetCurrentThreadId();

	if(m_bUseLocking)
	{
		if(Exclusive && (Container->LockMode == CONTAINER_LOCK_WRITE))
		{
			// Now the actual lock mode is RW (exclusive access)
			Container->ActualLockMode = CONTAINER_LOCK_READWRITE;
			// We require the access (this should block other requests)
			Container->ThreadAccessRequired = ThreadId;
			// we must wait for everybody to finish
			WaitSingleThread(Name, ThreadId);
			// Now we record the ref count in order to know when we have to
			// return to write only lock mode
			Container->InitialRefCount = Container->GetRefCount(ThreadId);
		}
		if(Container->ActualLockMode == CONTAINER_LOCK_READWRITE)
		{
			// We are in RW lock mode (Exclusive access)
			if((Container->ThreadOwner == 0) || (Container->ThreadOwner == ThreadId))
			{
				// No other thread owns the container so we get it
				Container->ThreadOwner = ThreadId;
				Container->IncRefCount(ThreadId);
				Container->ThreadAccessRequired = 0;
			}
			else
			{
				// The container is own by somebody else
				// Wait to get the ownership
				Container = WaitSingleThread(Name, ThreadId);
				if(Container == NULL)
				{
					// The container has been removed
					NV_UNLOCKOBJECT(this);
					return NULL;
				}
		
				// Ok, the container is free
				// we can take the ownership
				Container->ThreadOwner = ThreadId;
				Container->IncRefCount(ThreadId);
				Container->ThreadAccessRequired = 0;
			}
		}	// End lock mode read write
		else
		{
			// Lock mode is write only
			// We can give the access
			Container->IncRefCount(ThreadId);
		}
	}

	// Last container cache mechanism
	if((LastContainer != NULL) && (LastContainerName == NameHere))
	{
		NvContainer *ReturnedContainer = LastContainer;
		NV_UNLOCKOBJECT(this);
//printf("req Container=%s - get Container=%s - Session=%s\n", NameHere.c_str(), Container->GetName().c_str(), ((NvSession*)(Session))->m_sIdent.c_str());
		return ReturnedContainer;
	}

	NvContainer *ActualContainer = (NvContainer *)Container;
	string RootName = "";
	while(1)
	{
		NameHere = RemoveRootName(NameHere);
		if(!NameHere.size())
			break;

		RootName = GetRootName(NameHere);
		if(!RootName.length())
			break;	// Should not occur
		
		ActualContainer = ActualContainer->GetSubContainer(RootName.c_str());
		if(ActualContainer == NULL)
		{
			if(m_bUseLocking)
				Container->DecRefCount(ThreadId);
			break;
		}
	}

	if(ActualContainer != NULL)
	{
		LastContainer = ActualContainer;
		LastContainerName = OriginalNameHere;
	}

	NV_UNLOCKOBJECT(this);
	return ActualContainer;
}





NvContainer* NvContainerList::GetCreateContainer(const char *Name, void *Session, bool Exclusive, int LockMode)
{
	// We call first the GetContainer in case the container ever exist
	// This is for performance reason
	NvContainer *ActualContainer = GetContainer(Name, Session, Exclusive);
	if(ActualContainer != NULL)
		return ActualContainer;


	// The container doesn't exist
	// (But maybe the root container exist)

	if(Name == NULL)
		return NULL;

	string NameHere = Name;
	NameHere = CleanContainerName(NameHere);
	string OriginalNameHere = NameHere;

	NV_LOCKOBJECT(this);

	LastContainer = NULL;


	NvRootContainer *Container = GetRootContainer(NameHere.c_str());
	if(Container == NULL)
	{
		// The root container doesn't exist
		// So it is created, the default lock mode is R/W if locking is used
		if(!InternalCreateContainer(Name, LockMode))
		{
			NV_UNLOCKOBJECT(this);
			return NULL;
		}
		Container = GetRootContainer(NameHere.c_str());
		if(Container == NULL)
		{
			// Should not occur because the container just has been created
			NV_UNLOCKOBJECT(this);
			return NULL;
		}
		Container->LockMode = LockMode;
	}
	int PreviousActualLockMode = Container->ActualLockMode;
	int FinalActualLockMode = PreviousActualLockMode;
	if(Exclusive && (Container->LockMode == CONTAINER_LOCK_WRITE))
		FinalActualLockMode = CONTAINER_LOCK_READWRITE;
	Container->ActualLockMode = CONTAINER_LOCK_READWRITE;

	long ThreadId = NvGetCurrentThreadId();
	if(m_bUseLocking)
	{
		// we must wait for everybody to finish
		// We require the access (this should block other requests)
		Container->ThreadAccessRequired = ThreadId;
		// we must wait for everybody to finish
		WaitSingleThread(Name, ThreadId);
		Container->InitialRefCount = Container->GetRefCount(ThreadId);
		if((Container->ThreadOwner == 0) || (Container->ThreadOwner == ThreadId))
		{
			// No other thread owns the container so we get it
			Container->ThreadOwner = ThreadId;
			Container->IncRefCount(ThreadId);
			Container->ThreadAccessRequired = 0;
		}
		else
		{
			// The container is own by somebody else
			// Wait to get the ownership
			Container = WaitSingleThread(Name, ThreadId);
			if(Container == NULL)
			{
				// The container has been removed
				// So we create it again
				if(!InternalCreateContainer(Name, LockMode))
				{
					NV_UNLOCKOBJECT(this);
					return NULL;
				}
				Container = GetRootContainer(NameHere.c_str());
				if(Container == NULL)
				{
					// Should not occur because the container just has been created
					NV_UNLOCKOBJECT(this);
					return NULL;
				}
				Container->LockMode = LockMode;
			}
	
			// Ok, the container is free
			// we can take the ownership
			Container->ThreadOwner = ThreadId;
			Container->IncRefCount(ThreadId);
			Container->ThreadAccessRequired = 0;
		}
	}

	ActualContainer = (NvContainer *)Container;
	string RootName = "";
	while(1)
	{
		NameHere = RemoveRootName(NameHere);
		if(!NameHere.size())
			break;

		RootName = GetRootName(NameHere);
		if(!RootName.length())
			break;	// Should not occur
		
		if(!ActualContainer->CreateSubContainer(RootName.c_str()))
		{
			if(m_bUseLocking)
				Container->DecRefCount(ThreadId);
			Container->ActualLockMode = PreviousActualLockMode;
			NV_UNLOCKOBJECT(this);
			return NULL;
		}
		
		ActualContainer = ActualContainer->GetSubContainer(RootName.c_str());
		if(ActualContainer == NULL)
		{
			if(m_bUseLocking)
				Container->DecRefCount(ThreadId);
			Container->ActualLockMode = PreviousActualLockMode;
			NV_UNLOCKOBJECT(this);
			break;
		}
	}

	Container->ActualLockMode = FinalActualLockMode;
	if(FinalActualLockMode != CONTAINER_LOCK_READWRITE)
	{
		Container->ThreadOwner = 0;
		Container->InitialRefCount = 0;
	}
	if(ActualContainer != NULL)
	{
		LastContainer = ActualContainer;
		LastContainerName = OriginalNameHere;
	}

	NV_UNLOCKOBJECT(this);
	return ActualContainer;
}



void NvContainerList::FreeContainer(const char *Name, void *Session, bool Force)
{
	NV_LOCKOBJECT(this);
	if(!m_bUseLocking)
	{
		NV_UNLOCKOBJECT(this);
		return;
	}

	NvRootContainer *Container = GetRootContainer(Name);
	if(Container == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return;
	}

	long ThreadId = NvGetCurrentThreadId();
	Container->DecRefCount(ThreadId, Force);
	NV_UNLOCKOBJECT(this);
}


bool NvContainerList::CreateContainer(const char *Name, int LockMode)
{
	bool RetVal = false;
	NV_LOCKOBJECT(this);
	RetVal = InternalCreateContainer(Name, LockMode);
	NV_UNLOCKOBJECT(this);
	return RetVal;
}



void NvContainerList::RemoveContainer(const char *Name, void *Session)
{
	if(Name == NULL)
		return;

	string NameHere = Name;
	NameHere = CleanContainerName(NameHere);


	NV_LOCKOBJECT(this);
	NvRootContainer *RootContainer = GetRootContainer(NameHere.c_str());
	if(RootContainer == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return;
	}

	int PreviousActualLockMode = RootContainer->ActualLockMode;
	RootContainer->ActualLockMode = CONTAINER_LOCK_READWRITE;
	long ThreadId = NvGetCurrentThreadId();
	if(m_bUseLocking)
	{

		// we must wait for everybody to finish
		RootContainer->ThreadAccessRequired = ThreadId;
		// we must wait for everybody to finish
		WaitSingleThread(Name, ThreadId);
		if(PreviousActualLockMode == CONTAINER_LOCK_WRITE)
			RootContainer->InitialRefCount = RootContainer->GetRefCount(ThreadId);

		if((RootContainer->ThreadOwner == 0) || (RootContainer->ThreadOwner == ThreadId))
		{
			// No other thread owns the container so we get it
			RootContainer->ThreadOwner = ThreadId;
			RootContainer->IncRefCount(ThreadId);
			RootContainer->ThreadAccessRequired = 0;
		}
		else
		{
			// The container is own by somebody else
			// Wait to get the ownership
			RootContainer = WaitSingleThread(Name, ThreadId);
			if(RootContainer == NULL)
			{
				// The container has been removed
				NV_UNLOCKOBJECT(this);
				return;
			}
	
			// Ok, the container is free
			// we can take the ownership
			RootContainer->ThreadOwner = ThreadId;
			RootContainer->IncRefCount(ThreadId);
			RootContainer->ThreadAccessRequired = 0;
		}
	}


	string SubContainerName = "";
	string RootName = "";
	NvContainer *ActualContainer = (NvContainer *)RootContainer;
	while(1)
	{
		NameHere = RemoveRootName(NameHere);
		if(!NameHere.size())
			break;

		RootName = GetRootName(NameHere);
		if(!RootName.length())
			break;	// Should not occur
		
		ActualContainer = ActualContainer->GetSubContainer(RootName.c_str());
		if(ActualContainer == NULL)
		{
			if(m_bUseLocking)
				RootContainer->DecRefCount(ThreadId);
			RootContainer->ActualLockMode = PreviousActualLockMode;
			break;
		}
		SubContainerName = RootName;
	}
	if(ActualContainer == NULL)
	{
		if(m_bUseLocking)
			RootContainer->DecRefCount(ThreadId);
		RootContainer->ActualLockMode = PreviousActualLockMode;
		NV_UNLOCKOBJECT(this);
		return;
	}

	if((LastContainer != NULL) && (LastContainer == ActualContainer))
	{
		LastContainer = NULL;
		LastContainerName = "";
	}

	if(ActualContainer == (NvContainer *)RootContainer)
	{
		if(m_bUseLocking)
			RootContainer->DecRefCount(ThreadId);
		RootContainer->ActualLockMode = PreviousActualLockMode;

		if(m_mContainers.find(RootContainer->GetName()) != m_mContainers.end())
		{
			m_mContainers.erase(RootContainer->GetName());
			delete RootContainer;
		}
		NV_UNLOCKOBJECT(this);
		return;
	}

	
	NvContainer *ParentContainer = ActualContainer->GetParent();
	if(ParentContainer == NULL)
	{
		// should not occur
		if(m_bUseLocking)
			RootContainer->DecRefCount(ThreadId);
		RootContainer->ActualLockMode = PreviousActualLockMode;
		NV_UNLOCKOBJECT(this);
		return;
	}

	ParentContainer->RemoveSubContainer(SubContainerName.c_str());
	if(m_bUseLocking)
		RootContainer->DecRefCount(ThreadId);
	RootContainer->ActualLockMode = PreviousActualLockMode;
	NV_UNLOCKOBJECT(this);
}



NvRootContainer* NvContainerList::GetRootContainer(const char *Name, bool Erase)
{
	string NameHere = NV_CONTAINER_DEFAULT_NAME;
	if(Name != NULL)
		NameHere = Name;

	if(!NameHere.length())
		NameHere = NV_CONTAINER_DEFAULT_NAME;

	NameHere = GetRootName(NameHere);

	if(!NameHere.length())
		return NULL;

	NvRootContainer *ReturnedContainer = NULL;
	if(m_mContainers.find(NameHere) != m_mContainers.end())
		ReturnedContainer = (NvRootContainer *)(m_mContainers[NameHere]);
	if(Erase && (ReturnedContainer != NULL))
		m_mContainers.erase(ReturnedContainer->GetName());
	return ReturnedContainer;
}


NvRootContainer* NvContainerList::WaitSingleThread(const char *Name, long ThreadId)
{

	string NameHere = Name;
	
	// Get the container
	NvRootContainer *Container = GetRootContainer(NameHere.c_str());
	if(Container == NULL)
		return NULL;

	int RefCountSize = (int)(Container->m_mThreadRefCounts.size());
	if(!RefCountSize)
		return Container;

	if((RefCountSize == 1) && (Container->GetRefCount(ThreadId) > 0))
		return Container;

	int CheckEvery = 1;
	time_t InitialTime = time(NULL);
	while(1)
	{
		// The container is not free so we are still waiting
		NV_UNLOCKOBJECT(this);
		SLEEP(CheckEvery);

		// We first do the check very often but less and less while the time is growing
		CheckEvery += (CheckEvery/10)+1;
		if(CheckEvery > 10000)		// When the check time is 10 seconds, we check every 10 seconds
			CheckEvery = 10000;
		NV_LOCKOBJECT(this);
		
		// Get the container
		Container = GetRootContainer(NameHere.c_str());
		if(Container == NULL)
		{
			// The container has been removed
			return NULL;
		}

		// We must block accesses from other threads if one has
		// requested an exclusive access
		if(Container->ThreadAccessRequired)
			if(Container->ThreadAccessRequired != ThreadId)
				continue;
		RefCountSize = (int)(Container->m_mThreadRefCounts.size());
		if(!RefCountSize)
		{
			// Ok, the container is free (the only owner is us)
			// we can take the ownership
			break;
		}
		if(RefCountSize == 1)
		{
			// One user, check if it's not us
			if(Container->GetRefCount(ThreadId) > 0)
			{
				// this is us
				break;
			}
		}
	}

	return Container;
}


void NvContainerList::SetRootDirectory(const char *RootDirectory)
{
	NV_LOCKOBJECT(this);
	map<string,void *>::iterator pos;
	for(pos = m_mContainers.begin(); pos != m_mContainers.end(); ++pos)
	{
		NvRootContainer *Container = (NvRootContainer *)(pos->second);
		if(Container == NULL)
			continue;
		string ChildRootDirectory = RootDirectory;
		AddDirSuffix(ChildRootDirectory);
		if(ChildRootDirectory.length())
			ChildRootDirectory += Container->GetName();
		Container->SetRootDirectory(ChildRootDirectory.c_str());
	}
	NV_UNLOCKOBJECT(this);
}


bool NvContainerList::Synchronize(int Mode)
{
	bool RetVal = true;
	NV_LOCKOBJECT(this);
	map<string,void *>::iterator pos;
	for(pos = m_mContainers.begin(); pos != m_mContainers.end(); ++pos)
	{
		NvRootContainer *Container = (NvRootContainer *)(pos->second);
		if(Container == NULL)
			continue;
		if(!Container->Synchronize(Mode))
		{
			RetVal = false;
			break;
		}
	}
	NV_UNLOCKOBJECT(this);

	return RetVal;
}



string NvContainerList::GetRootName(string Name)
{
	Name = CleanContainerName(Name);
	if(!Name.length())
		Name = NV_CONTAINER_DEFAULT_NAME;


	// The name may point to sub-container
	// So we get the root name
	string::size_type index;
	index = Name.find('.');
	if(index != string::npos)
		Name = Name.erase(index);

	return Name;
}



string NvContainerList::RemoveRootName(string Name)
{
	Name = CleanContainerName(Name);

	// We remove the root name
	string::size_type index;
	index = Name.find('.');
	if(index != string::npos)
		Name = Name.substr(index+1);
	else
		Name = "";

	return Name;
}



string NvContainerList::CleanContainerName(string Name)
{
	CleanString(Name, true);
	MakeLower(Name);

	if(!Name.size())
		return "";
	
	string::size_type index;
	
	// concatenate 2 consecutive points
	while((index = Name.find("..")) != string::npos)
		Name.replace(index, 2, ".");

	// remove slashes
	while((index = Name.find('/')) != string::npos)
		Name.erase(index, 1);

	// remove backslashes
	while((index = Name.find('\\')) != string::npos)
		Name.erase(index, 1);

	return Name;
}




bool NvContainerList::InternalCreateContainer(const char *Name, int LockMode)
{
	string NameHere = NV_CONTAINER_DEFAULT_NAME;
	if(Name != NULL)
		NameHere = Name;

	if(!NameHere.length())
		NameHere = NV_CONTAINER_DEFAULT_NAME;

	NameHere = CleanContainerName(NameHere);
	if(!NameHere.length())
		return false;

	string RootName = "";

	NvRootContainer *RootContainer = GetRootContainer(NameHere.c_str());
	if(RootContainer != NULL)
	{
		if(m_bUseLocking)
			RootContainer->Lock();
		RootName = RootContainer->GetName();
	}
	else
	{
		RootName = GetRootName(NameHere);
		if(!RootName.length())
			return false;
		RootContainer = NVNEW NvRootContainer;
		if(RootContainer == NULL)
			return false;
		bool bUseLastMechanism = true;
		if(LockMode == CONTAINER_LOCK_WRITE)
			bUseLastMechanism = false;
		RootContainer->bUseLastSubContainer = bUseLastMechanism;
		RootContainer->bUseLastObject = bUseLastMechanism;
		RootContainer->LockMode = LockMode;
		if(!RootContainer->SetName(RootName))
		{
			delete RootContainer;
			return false;
		}
		m_mContainers[RootContainer->GetName()] = ((void *)RootContainer);
		if(m_bUseLocking)
			RootContainer->Lock();
	}

	// Now we have the Root container and its locked

	// We create now the necessary subcontainers if they doesn't exist

	NvContainer *ActualContainer = (NvContainer *)RootContainer;
	while(1)
	{
		NameHere = RemoveRootName(NameHere);
		if(!NameHere.size())
			break;

		RootName = GetRootName(NameHere);
		if(!RootName.length())
			break;	// Should not occur

		
		if(!ActualContainer->CreateSubContainer(RootName.c_str()))
		{
			if(m_bUseLocking)
				RootContainer->Unlock();
			return false;
		}
		
		ActualContainer = ActualContainer->GetSubContainer(RootName.c_str());
		if(ActualContainer == NULL)
		{
			// Should not occur
			if(m_bUseLocking)
				RootContainer->Unlock();
			return false;
		}
	}

	if(m_bUseLocking)
		RootContainer->Unlock();
	return true;
}



void NvContainerList::GetContainerNames(vector<string> *List)
{
	NV_LOCKOBJECT(this);
	map<string,void *>::iterator pos;
	for(pos = m_mContainers.begin(); pos != m_mContainers.end(); ++pos)
		List->push_back(pos->first);
	NV_UNLOCKOBJECT(this);
}



/////////////////////////////////////////////////////////////////////////////
// NvRegistryContainer


NvRegistryContainer::NvRegistryContainer()
	: NvRootContainer()
{
	InUse = 0;
	bRemoveDirRequest = false;
}

bool NvRegistryContainer::GetKey(const char *Key, NvContainer *Destination, int AppendMode, const char *Entries, bool WithSubKeys, bool WithData, const char *FileDir, int PersistTime, bool Recurse)
{
	Lock();

	if(!m_sRootDirectory.length())
	{
		// The registry container should always have a root directory
		Unlock();
		return false;
	}

	string KeyHere = "";
	if(Key != NULL)
		KeyHere = Key;
	CleanString(KeyHere);
	MakeLower(KeyHere);

	KeyHere = NvContainerList::CleanContainerName(KeyHere);
	string ProviKey = GetName();
	if(KeyHere.length())
	{
		ProviKey += ".";
		ProviKey += KeyHere;
		KeyHere = ProviKey;
	}
	
	NvContainer *ActualContainer = (NvContainer *)this;
	string RootName = "";
	string NameHere = KeyHere;
	while(1)
	{
		NameHere = NvContainerList::RemoveRootName(NameHere);
		if(!NameHere.size())
			break;

		RootName = NvContainerList::GetRootName(NameHere);
		if(!RootName.length())
			break;	// Should not occur
		
		ActualContainer = ActualContainer->GetSubContainer(RootName.c_str());
		if(ActualContainer == NULL)
			break;
	}

	if(ActualContainer == NULL)
	{
		Unlock();
		return false;
	}

	// now we have the key container in ActualContainer
	// If the Destination container is NULL this is just to check if the Key/Entry exists
	bool RetVal = true;
	if(Destination == NULL)
	{
		// Entries can then contains only a single value
		if(Entries != NULL && strlen(Entries))
		{
			if(ActualContainer->GetObject(Entries) == NULL)
			{
				Unlock();
				return false;
			}
		}
	}
	else
	{
		// Real GetKey
		RetVal = Destination->CopyFrom(ActualContainer, AppendMode, Entries, WithSubKeys, WithData, FileDir, PersistTime, Recurse);
	}
	Unlock();
	
	return RetVal;
}


bool NvRegistryContainer::SetKey(const char *Key, NvContainer *Source, int AppendMode, const char *Entries, bool WithSubKeys, bool WithData)
{
	Lock();

	if(!m_sRootDirectory.length())
	{
		// The registry container should always have a root directory
		Unlock();
		return false;
	}

	string KeyHere = "";
	if(Key != NULL)
		KeyHere = Key;
	CleanString(KeyHere);
	MakeLower(KeyHere);

	KeyHere = NvContainerList::CleanContainerName(KeyHere);
	string ProviKey = GetName();
	if(KeyHere.length())
	{
		ProviKey += ".";
		ProviKey += KeyHere;
		KeyHere = ProviKey;
	}
	
	NvContainer *ActualContainer = (NvContainer *)this;
	string RootName = "";
	string NameHere = KeyHere;
	while(1)
	{
		NameHere = NvContainerList::RemoveRootName(NameHere);
		if(!NameHere.size())
			break;

		RootName = NvContainerList::GetRootName(NameHere);
		if(!RootName.length())
			break;	// Should not occur
		
		if(!ActualContainer->CreateSubContainer(RootName.c_str()))
			break;
		
		ActualContainer = ActualContainer->GetSubContainer(RootName.c_str());
		if(ActualContainer == NULL)
			break;
	}

	if(ActualContainer == NULL)
	{
		Unlock();
		return false;
	}

	// now we have the key container in ActualContainer

	bool RetVal = true;
	if(Source != NULL)
		RetVal = ActualContainer->CopyFrom(Source, AppendMode, Entries, WithSubKeys, WithData);

	Unlock();
	
	return RetVal;
}


bool NvRegistryContainer::RemoveKey(const char *Key, const char *Entries, bool WithSubKeys)
{
	Lock();

	if(!m_sRootDirectory.length())
	{
		// The registry container should always have a root directory
		Unlock();
		return true;
	}

	string KeyHere = "";
	if(Key != NULL)
		KeyHere = Key;
	CleanString(KeyHere);
	MakeLower(KeyHere);

	KeyHere = NvContainerList::CleanContainerName(KeyHere);
	string ProviKey = GetName();
	if(KeyHere.length())
	{
		ProviKey += ".";
		ProviKey += KeyHere;
		KeyHere = ProviKey;
	}
	
	NvContainer *ActualContainer = (NvContainer *)this;
	string RootName = "";
	string NameHere = KeyHere;
	while(1)
	{
		NameHere = NvContainerList::RemoveRootName(NameHere);
		if(!NameHere.size())
			break;

		RootName = NvContainerList::GetRootName(NameHere);
		if(!RootName.length())
			break;	// Should not occur
		
		ActualContainer = ActualContainer->GetSubContainer(RootName.c_str());
		if(ActualContainer == NULL)
			break;
	}

	if(ActualContainer == NULL)
	{
		Unlock();
		return true;	// no key to remove because it's not found
	}

	// now we have the key container in ActualContainer

	string ObjectsHere = "";
	if(Entries != NULL)
		ObjectsHere = Entries;
	CleanString(ObjectsHere);

	// starts with objects
	NvObjStringList ObjectsToGet;

	bool RetVal = true;

	// if Entries is not null, we must select only requested objects (maybe none)
	// Remark: if object is given but blank, no objects will be get
	// and if object is not given (NULL), all valid objects will be get
	if(Entries != NULL)
	{
		while(ObjectsHere.length())
		{
			// Several objects can be in the list
			// They are separated by a semicolumn
			string ObjectName = ObjectsHere;
			string::size_type index = ObjectName.find(';');
			if(index != string::npos)
			{
				ObjectName = ObjectsHere.substr(0, index);
				ObjectsHere = ObjectsHere.substr(index+1);
			}
			else
				ObjectsHere = "";

			// An object name has been found
			CleanString(ObjectName);
			
			// It should not be empty
			if(!ObjectName.length())
				continue;

			ActualContainer->RemoveObject(ObjectName.c_str());
		}
	}
	else
		ActualContainer->ClearObjects(true);	// We remove all the objects

	
	// Remove all subkeys if requested
	if(WithSubKeys)
		RetVal = ActualContainer->ClearSubcontainers(true);


	if((Entries == NULL) && WithSubKeys)
	{
		// The request is to remove all 
		// So we can remove the Subcontainer itself

		NvContainer *ParentContainer = ActualContainer->GetParent();
		// We remove the key only if it's not the rrot key
		if(ParentContainer != NULL)
			RetVal = ParentContainer->RemoveSubContainer(ActualContainer->GetName().c_str());
	}

	Unlock();
	
	return RetVal;
}



bool NvRegistryContainer::Import(const char *Key, const char *FileName, int AppendMode)
{
	Lock();

	if(!m_sRootDirectory.length())
	{
		// The registry container should always have a root directory
		Unlock();
		return false;
	}

	string KeyHere = "";
	if(Key != NULL)
		KeyHere = Key;
	CleanString(KeyHere);
	MakeLower(KeyHere);

	KeyHere = NvContainerList::CleanContainerName(KeyHere);
	string ProviKey = GetName();
	if(KeyHere.length())
	{
		ProviKey += ".";
		ProviKey += KeyHere;
		KeyHere = ProviKey;
	}
	
	NvContainer *ActualContainer = (NvContainer *)this;
	string RootName = "";
	string NameHere = KeyHere;
	while(1)
	{
		NameHere = NvContainerList::RemoveRootName(NameHere);
		if(!NameHere.size())
			break;

		RootName = NvContainerList::GetRootName(NameHere);
		if(!RootName.length())
			break;	// Should not occur
		
		if(!ActualContainer->CreateSubContainer(RootName.c_str()))
			break;
		
		ActualContainer = ActualContainer->GetSubContainer(RootName.c_str());
		if(ActualContainer == NULL)
			break;
	}

	if(ActualContainer == NULL)
	{
		Unlock();
		return false;
	}

	bool RetVal = ActualContainer->Import(FileName, AppendMode);

	Unlock();

	return RetVal;
}


bool NvRegistryContainer::Export(const char *Key, const char *FileName, bool WithSubContainers)
{
	Lock();

	if(!m_sRootDirectory.length())
	{
		// The registry container should always have a root directory
		Unlock();
		return false;
	}

	string KeyHere = "";
	if(Key != NULL)
		KeyHere = Key;
	CleanString(KeyHere);
	MakeLower(KeyHere);

	KeyHere = NvContainerList::CleanContainerName(KeyHere);
	string ProviKey = GetName();
	if(KeyHere.length())
	{
		ProviKey += ".";
		ProviKey += KeyHere;
		KeyHere = ProviKey;
	}
	
	NvContainer *ActualContainer = (NvContainer *)this;
	string RootName = "";
	string NameHere = KeyHere;
	while(1)
	{
		NameHere = NvContainerList::RemoveRootName(NameHere);
		if(!NameHere.size())
			break;

		RootName = NvContainerList::GetRootName(NameHere);
		if(!RootName.length())
			break;	// Should not occur
		
		ActualContainer = ActualContainer->GetSubContainer(RootName.c_str());
		if(ActualContainer == NULL)
			break;
	}

	if(ActualContainer == NULL)
	{
		Unlock();
		return false;
	}

	// now we have the key container in ActualContainer

	bool RetVal = ActualContainer->Export(FileName, NULL, WithSubContainers);

	Unlock();
	return RetVal;
}





/////////////////////////////////////////////////////////////////////////////
// NvDescriptionContainer

// Loads data from a system, service application or web service description file
bool NvDescriptionContainer::Load(const char *FileName)
{
#ifndef NV_SERVER
	// Used only on server side
	return true;
#else

	if(FileName == NULL)
		return false;
	
	Lock();

	NvContainer::Clear();
	
	FILE *File = Nvfopen(FileName, "r");
	if(File == NULL)
	{
		Unlock();
		return false;
	}


	// We create a subcontainer named provi;
	if(!CreateSubContainer("provi"))
	{
		fclose(File);
		Unlock();
		return false;
	}

	NvContainer* ProviContainer = GetSubContainer("provi");
	if(ProviContainer == NULL)
	{
		fclose(File);
		RemoveSubContainer("provi");
		Unlock();
		return false;
	}

	char Buffer[4096];
	string Line;
	string::size_type index;

	NvObjStringList *StringList = NULL;

	while(freadline(Buffer, sizeof(Buffer), File))
	{
		Line = Buffer;
		CleanString(Line);
		if(!Line.length())
			continue;
		if(Line.find("\\\\") == 0)	// comment
			continue;
		if(Line.find("//") == 0)	// comment
			continue;
		if(Line.find(";") == 0)		// comment
			continue;

		if((Line.find('[') == 0) && ((index = Line.find(']')) != string::npos))
		{
			// This a new object
			// We get its name
			string ObjectName = "";
			if(index > 1)
				ObjectName = Line.substr(1, index-1);
			MakeUpper(ObjectName);
			CleanString(ObjectName);
			if(ObjectName.length())
			{
				// This is a real object name
				StringList = (NvObjStringList *)(ProviContainer->GetObject(ObjectName.c_str()));
				if(StringList == NULL)
				{
					// The object doesn't exist so we create it
					ProviContainer->CreateObject(ObjectName.c_str(), NV_OBJTYPE_STRINGLIST);
					StringList = (NvObjStringList *)(ProviContainer->GetObject(ObjectName.c_str()));
				}
				continue;
			}
		}

		if(StringList == NULL)
			continue;

		StringList->Add(Line);
	}

	fclose(File);

	// Now, we have the section contents in the provi container
	// We can get these ones we are interested by

	int i,j,k;

	bool bUnicode = false;
	bool bWithEncoding = false;

	// We get the INFO section
	StringList = (NvObjStringList *)(ProviContainer->GetObject("INFO"));
	CreateObject("INFO", NV_OBJTYPE_INDSTRINGLIST);
	if(StringList != NULL)
	{
		// First get the ENCODING value if there is one
		for(i=0; i<StringList->GetSize(); i++)
		{
			Line = StringList->GetValue(i);
			string key = "";
			string value = "";
			index = Line.find('=');
			if(index != string::npos)
			{
				key = Line.substr(0, index);
				CleanString(key);
				value = Line.substr(index+1);
				CleanString(value);
			}
			MakeUpper(key);
			MakeUpper(value);
			if(key == "ENCODING")
			{
				// We do not need this entry
				StringList->Remove(i);
				if(value == "UTF-8")
				{
					bWithEncoding = true;
					bUnicode = true;
				}
				if(value == "ISO-8859-1")
					bWithEncoding = true;
				break;
			}
		}

		NvObjIndStringList *IndStringList = (NvObjIndStringList *)(GetObject("INFO"));
		IndStringList->m_bCaseSensitive = true;
		for(i=0; i<StringList->GetSize(); i++)
		{
			Line = StringList->GetValue(i);
			if(bWithEncoding)
				EncodeString(Line, bUnicode, GetMain()->b_Unicode);
			string key = "";
			string value = "";
			index = Line.find('=');
			if(index != string::npos)
			{
				key = Line.substr(0, index);
				CleanString(key);
				value = Line.substr(index+1);
				CleanString(value);
			}
			if(key.length())
				IndStringList->SetValue(key, value);
		}
		ProviContainer->RemoveObject("INFO");	// We don't need any more this object
	}

	// We get the SETTINGS section
	StringList = (NvObjStringList *)(ProviContainer->GetObject("SETTINGS"));
	CreateObject("SETTINGS", NV_OBJTYPE_INDSTRINGLIST);
	if(StringList != NULL)
	{
		NvObjIndStringList *IndStringList = (NvObjIndStringList *)(GetObject("SETTINGS"));
		for(i=0; i<StringList->GetSize(); i++)
		{
			Line = StringList->GetValue(i);
			DecodeUrlString(Line, bUnicode, false);
			string key = "";
			string value = "";
			index = Line.find('=');
			if(index != string::npos)
			{
				key = Line.substr(0, index);
				CleanString(key);
				value = Line.substr(index+1);
				CleanString(value);
			}
			if(key.length())
				IndStringList->SetValue(key, value);
		}
		ProviContainer->RemoveObject("SETTINGS");	// We don't need any more this object
	}

	// We get the PERMISSIONS section
	StringList = (NvObjStringList *)(ProviContainer->GetObject("PERMISSIONS"));
	CreateObject("PERMISSIONS", NV_OBJTYPE_INDSTRINGLIST);
	if(StringList != NULL)
	{
		NvObjIndStringList *IndStringList = (NvObjIndStringList *)(GetObject("PERMISSIONS"));
		for(i=0; i<StringList->GetSize(); i++)
		{
			Line = StringList->GetValue(i);
			DecodeUrlString(Line, bUnicode, false);
			string key = "";
			string value = "";
			index = Line.find('=');
			if(index != string::npos)
			{
				key = Line.substr(0, index);
				CleanString(key);
				value = Line.substr(index+1);
				CleanString(value);
			}
			if(key.length())
				IndStringList->SetValue(key, value);
		}
		ProviContainer->RemoveObject("PERMISSIONS");	// We don't need any more this object
	}

	// We get the COMMAND_CLASSES section
	StringList = (NvObjStringList *)(ProviContainer->GetObject("COMMAND_CLASSES"));
	CreateObject("COMMAND_CLASSES", NV_OBJTYPE_INDSTRINGLIST);
	if(StringList != NULL)
	{
		NvObjIndStringList *IndStringList = (NvObjIndStringList *)(GetObject("COMMAND_CLASSES"));
		for(i=0; i<StringList->GetSize(); i++)
		{
			Line = StringList->GetValue(i);
			DecodeUrlString(Line, bUnicode, false);
			string key = "";
			string value = "";
			index = Line.find('=');
			if(index != string::npos)
			{
				key = Line.substr(0, index);
				CleanString(key);
				value = Line.substr(index+1);
				CleanString(value);
			}
			if(key.length())
			{
				IndStringList->SetValue(key, value);
				// Now, we must also get the corresponding COMMAND_CLASS sections
				string CommandClassObjectName = "COMMAND_CLASS_";
				CommandClassObjectName += key;
				NvObjStringList *StringList2 = (NvObjStringList *)(ProviContainer->GetObject(CommandClassObjectName.c_str()));
				CreateObject(CommandClassObjectName.c_str(), NV_OBJTYPE_INDSTRINGLIST);
				if(StringList2 != NULL)
				{
					NvObjIndStringList *IndStringList2 = (NvObjIndStringList *)(GetObject(CommandClassObjectName.c_str()));
					for(j=0; j<StringList2->GetSize(); j++)
					{
						Line = StringList2->GetValue(j);
						DecodeUrlString(Line, bUnicode, false);
						string key = "";
						string value = "";
						index = Line.find('=');
						if(index != string::npos)
						{
							key = Line.substr(0, index);
							CleanString(key);
							value = Line.substr(index+1);
							CleanString(value);
						}
						if(key.length())
						{
							IndStringList2->SetValue(key, value);
						}
					}
				}
				ProviContainer->RemoveObject(CommandClassObjectName.c_str());	// We don't need any more this object
			}
		}
		ProviContainer->RemoveObject("COMMAND_CLASSES");	// We don't need any more this object
	}

	// We get the ERROR_CLASSES section
	StringList = (NvObjStringList *)(ProviContainer->GetObject("ERROR_CLASSES"));
	CreateObject("ERROR_CLASSES", NV_OBJTYPE_INDSTRINGLIST);
	if(StringList != NULL)
	{
		NvObjIndStringList *IndStringList = (NvObjIndStringList *)(GetObject("ERROR_CLASSES"));
		for(i=0; i<StringList->GetSize(); i++)
		{
			Line = StringList->GetValue(i);
			DecodeUrlString(Line, bUnicode, false);
			string key = "";
			string value = "";
			index = Line.find('=');
			if(index != string::npos)
			{
				key = Line.substr(0, index);
				CleanString(key);
				value = Line.substr(index+1);
				CleanString(value);
			}
			if(key.length())
			{
				IndStringList->SetValue(key, value);
				// Now, we must also get the corresponding ERROR_CLASS sections
				string ErrorClassObjectName = "ERROR_CLASS_";
				ErrorClassObjectName += key;
				NvObjStringList *StringList2 = (NvObjStringList *)(ProviContainer->GetObject(ErrorClassObjectName.c_str()));
				CreateObject(ErrorClassObjectName.c_str(), NV_OBJTYPE_TABLE);
				NvObjTable *Table = (NvObjTable *)(GetObject(ErrorClassObjectName.c_str()));
				Table->InsertColumn("CODE", (const char *)NULL, true);
				// Now, we must create the smae number of columns than the number of languages
				NvObjStringList *StringList3 = (NvObjStringList *)(ProviContainer->GetObject("ERROR_LANGUAGES"));
				if(StringList3 != NULL)
				{
					for(j=0; j<StringList3->GetSize(); j++)
					{
						Line = StringList3->GetValue(j);
						DecodeUrlString(Line, bUnicode, false);
						MakeUpper(Line);
						if(Line.length())
						{
							string ColumnName = "DESC_";
							ColumnName += Line;
							Table->InsertColumn(ColumnName);
						}
					}
				}
				
				if(StringList2 != NULL)
				{
					for(j=0; j<StringList2->GetSize(); j++)
					{
						Line = StringList2->GetValue(j);
						DecodeUrlString(Line, bUnicode, false);
						string key = "";
						string value = "";
						index = Line.find('=');
						if(index != string::npos)
						{
							key = Line.substr(0, index);
							CleanString(key);
							value = Line.substr(index+1);
							CleanString(value);
						}
						if(key.length())
						{
							Table->InsertRow();
							Table->InsertCellLine(j, 0, key);
							for(k=1; k<Table->GetNumColumn(); k++)
							{
								index = value.find(';');
								string Description = "";
								if(index != string::npos)
								{
									Description = value.substr(0, index);
									CleanString(Description);
									value = value.substr(index+1);
								}
								else
								{
									Description = value;
									value = "";
								}
								Table->InsertCellLine(j, k, Description);
							}
						}
					}
				}
				ProviContainer->RemoveObject(ErrorClassObjectName.c_str());	// We don't need any more this object
			}
		}
		ProviContainer->RemoveObject("ERROR_CLASSES");	// We don't need any more this object
	}

	
	// We don't need this sub container any more
	RemoveSubContainer("provi");
	
	Unlock();

	return true;
#endif // NV_SERVER
}

/////////////////////////////////////////////////////////////////////////////
// NvRootContainer

NvRootContainer::NvRootContainer()
	: NvContainer()
{
	ThreadOwner = 0;
	ThreadAccessRequired = 0;
	InitialRefCount = 0;
	LockMode = CONTAINER_LOCK_READWRITE;
	ActualLockMode = LockMode;
}


NvRootContainer::~NvRootContainer()
{
	Clear();
}


void NvRootContainer::Clear()
{
	Lock();
	NvContainer::Clear();
	ThreadOwner = 0;
	ThreadAccessRequired = 0;
	InitialRefCount = 0;
	LockMode = CONTAINER_LOCK_READWRITE;
	ActualLockMode = LockMode;
	Unlock();
}


void NvRootContainer::Lock()
{
	NV_LOCKLOCK(&Protect);
}


void NvRootContainer::Unlock()
{
	NV_UNLOCKLOCK(&Protect);
}

void NvRootContainer::IncRefCount(long ThreadId)
{
	int LocalRefCount = 0;
	if(m_mThreadRefCounts.find(ThreadId) != m_mThreadRefCounts.end())
		LocalRefCount = m_mThreadRefCounts[ThreadId];
	LocalRefCount++;
	if(LocalRefCount > 0)
		m_mThreadRefCounts[ThreadId] = LocalRefCount;
}

void NvRootContainer::DecRefCount(long ThreadId, bool Force)
{
	int LocalRefCount = 0;
	bool bInList = false;
	if(m_mThreadRefCounts.find(ThreadId) != m_mThreadRefCounts.end())
	{
		LocalRefCount = m_mThreadRefCounts[ThreadId];
		bInList = true;
	}
	LocalRefCount--;
	if(LocalRefCount < 0)
		LocalRefCount = 0;
	if(Force)
		LocalRefCount = 0;
	if(LocalRefCount > 0)
	{
		m_mThreadRefCounts[ThreadId] = LocalRefCount;
		bInList = true;
	}
	else if(bInList)
	{
		m_mThreadRefCounts.erase(ThreadId);
		bInList = false;
	}
	if(!m_mThreadRefCounts.size())
	{
		// No more owner possible in this case
		ThreadOwner = 0;
		ActualLockMode = LockMode;
		InitialRefCount = 0;
		bInList = false;
	}
	if(!bInList)
	{
		// If we were the owner, we are not any more
		if(ThreadOwner == ThreadId)
		{
			ThreadOwner = 0;
			ActualLockMode = LockMode;
			InitialRefCount = 0;
		}
	}
	if(ActualLockMode == CONTAINER_LOCK_READWRITE)
	{
		if((m_mThreadRefCounts.size() == 1) && bInList)
		{
			if(LocalRefCount == InitialRefCount)
			{
				// We go back to the previous lock mode (should be WRITE only in this case)
				// (because InitialRefCount is used only when in WRITE only locked container
				// we try to access it in WRITE mode)
				ThreadOwner = 0;
				ActualLockMode = LockMode;
				InitialRefCount = 0;
			}
		}
	}
}

int NvRootContainer::GetRefCount(long ThreadId)
{
	int LocalRefCount = 0;
	if(m_mThreadRefCounts.find(ThreadId) != m_mThreadRefCounts.end())
		LocalRefCount = m_mThreadRefCounts[ThreadId];
	return LocalRefCount;
}


/////////////////////////////////////////////////////////////////////////////
// NvContainer

NvContainer::NvContainer()
	: NvLocalContainer()
{
	m_cParent = NULL;
	LastSubContainer = NULL;
	bUseLastSubContainer = true;
}

NvContainer::NvContainer(string Name)
	: NvLocalContainer(Name)
{
	m_cParent = NULL;
	LastSubContainer = NULL;
	bUseLastSubContainer = true;
}

NvContainer::NvContainer(const char *Name)
	: NvLocalContainer(Name)
{
	m_cParent = NULL;
	LastSubContainer = NULL;
	bUseLastSubContainer = true;
}

NvContainer::NvContainer(string Name, NvContainer *Parent)
	: NvLocalContainer(Name)
{
	SetParent(Parent);
	LastSubContainer = NULL;
	bUseLastSubContainer = true;
}

NvContainer::NvContainer(const char *Name, NvContainer *Parent)
	: NvLocalContainer(Name)
{
	SetParent(Parent);
	LastSubContainer = NULL;
	bUseLastSubContainer = true;
}


NvContainer::NvContainer(NvContainer *Parent)
	: NvLocalContainer()
{
	SetParent(Parent);
	LastSubContainer = NULL;
	bUseLastSubContainer = true;
}


NvContainer::~NvContainer()
{
	Clear();
}


void NvContainer::SetRootDirectory(const char *RootDirectory)
{
	NvLocalContainer::SetRootDirectory(RootDirectory);

	map<string,void *>::iterator pos;
	for(pos = m_mSubcontainers.begin(); pos != m_mSubcontainers.end(); ++pos)
	{
		NvContainer *Container = (NvContainer *)(pos->second);
		if(Container == NULL)
			continue;
		string ChildRootDirectory = RootDirectory;
		AddDirSuffix(ChildRootDirectory);
		if(ChildRootDirectory.length())
			ChildRootDirectory += Container->GetName();
		Container->SetRootDirectory(ChildRootDirectory.c_str());
	}
}



void NvContainer::GetRealSubContainerNames(vector<string> *List)
{
	string ChildRootDirectory = m_sRootDirectory;
	AddDirSuffix(ChildRootDirectory);
	if(!ChildRootDirectory.length())
		return;	// should not occur
	
	char FilesToFind[NV_MAX_PATHLENGTH];

#ifdef WIN32
	strncpy(FilesToFind, ChildRootDirectory.c_str(), NV_MAX_PATHLENGTH-1);
	if(strlen(FilesToFind))
	{
		// We add the last \ character
		if(FilesToFind[strlen(FilesToFind) - 1] != '\\')
			strcat(FilesToFind, "\\");
	}
	strcat(FilesToFind, "*");
	WIN32_FIND_DATA FindData;
	// Try to find the first file
	HANDLE hFileFind = FindFirstFile(FilesToFind, &FindData);
	if(hFileFind == INVALID_HANDLE_VALUE)
		return;	// No file
	while(1)
	{
		// We remove all the found files
		if((FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0)
		{
			if(FindData.cFileName[0] != '.')
			{
				if(strchr(FindData.cFileName, ' ') == NULL)
				{
					// The subdirectory name seems to be OK
					// We must still verify that its a lowercase name
					string Original = FindData.cFileName;
					string Copy = Original;
					MakeLower(Copy);
					if(Copy == Original)
						List->push_back(Original);	// This directory is a good candidate
				}
			}
		}

		if(!FindNextFile(hFileFind, &FindData))
			break;	// no more file

	}
	FindClose(hFileFind);	// we close the finder
#else

	char WorkDirHere[NV_MAX_PATHLENGTH];
	strncpy(WorkDirHere, ChildRootDirectory.c_str(), NV_MAX_PATHLENGTH-1);
	if(strlen(WorkDirHere))
	{
		// We remove the last / character
		if(WorkDirHere[strlen(WorkDirHere) - 1] == '/')
			WorkDirHere[strlen(WorkDirHere) - 1] = '\0';
	}
	else
		strcpy(WorkDirHere, ".");	// we are in the current directory

	// open directory file
	DIR *Dir = opendir(WorkDirHere);
	if(Dir == NULL)
		return;		// cannot open directory file
	
	dirent *DirEntry;
	while((DirEntry = readdir(Dir)) != 0)
	{
		if(DirEntry->d_name[0] == '.')
			continue;
		strcpy(FilesToFind, WorkDirHere);
		if(strlen(FilesToFind))
		{
			// We add the last \ character
			if(FilesToFind[strlen(FilesToFind) - 1] != '/')
				strcat(FilesToFind, "/");
		}
		strcat(FilesToFind, DirEntry->d_name);
		DIR *DirHere = NULL;
		if((DirHere = opendir(FilesToFind)) == NULL)
			continue;
		closedir(DirHere);
		// We have found a real subdirectory
		// The subdirectory name seems to be OK
		// We must still verify that its a lowercase name
		string Original = DirEntry->d_name;
		string Copy = Original;
		MakeLower(Copy);
		if(Copy == Original)
			List->push_back(Original);	// This directory is a good candidate
	}
	closedir(Dir);	// close the directory file
#endif
}



void NvContainer::GetRealObjectNames(vector<string> *List)
{
	string ChildRootDirectory = m_sRootDirectory;
	AddDirSuffix(ChildRootDirectory);
	if(!ChildRootDirectory.length())
		return;	// should not occur
	
	char FilesToFind[NV_MAX_PATHLENGTH];

#ifdef WIN32
	strncpy(FilesToFind, ChildRootDirectory.c_str(), NV_MAX_PATHLENGTH-1);
	if(strlen(FilesToFind))
	{
		// We add the last \ character
		if(FilesToFind[strlen(FilesToFind) - 1] != '\\')
			strcat(FilesToFind, "\\");
	}
	strcat(FilesToFind, "*");
	WIN32_FIND_DATA FindData;
	// Try to find the first file
	HANDLE hFileFind = FindFirstFile(FilesToFind, &FindData);
	if(hFileFind == INVALID_HANDLE_VALUE)
		return;	// No file
	while(1)
	{
		// We remove all the found files
		if((FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0)
		{
			if(!FindNextFile(hFileFind, &FindData))
				break;	// no more file
			continue;
		}
		if(strchr(FindData.cFileName, ' ') != NULL)
		{
			if(!FindNextFile(hFileFind, &FindData))
				break;	// no more file
			continue;
		}
		// The file name seems to be OK
		// We must still verify that its a lowercase name
		string Original = FindData.cFileName;
		string Copy = Original;
		MakeLower(Copy);
		if(Copy != Original)
		{
			if(!FindNextFile(hFileFind, &FindData))
				break;	// no more file
			continue;
		}
		// Now we must verify it's a real registry object file
		// A file has been found
		// get its type
		// And verify several things
		// The file name must be of the form objname_nvdd.obj
		// where dd is the file type on 2 digits
		string::size_type index = Original.rfind("_nv");
		if(index == string::npos)
		{
			if(!FindNextFile(hFileFind, &FindData))
				break;	// no more file
			continue;
		}
		string sType = Original.substr(index+3);
		index = sType.rfind(".obj");
		if(index == string::npos)
		{
			if(!FindNextFile(hFileFind, &FindData))
				break;	// no more file
			continue;
		}
		sType = sType.substr(0, index);
		if(sType.length() != 2)
		{
			if(!FindNextFile(hFileFind, &FindData))
				break;	// no more file
			continue;
		}
		int iType = atoi(sType.c_str());
		if(iType < NV_OBJTYPE_UNKNOWN)
		{
			if(!FindNextFile(hFileFind, &FindData))
				break;	// no more file
			continue;
		}
		if(iType > NV_OBJTYPE_MAXTYPEVALUE)
		{
			if(!FindNextFile(hFileFind, &FindData))
				break;	// no more file
			continue;
		}
		
		index = Original.rfind("_nv");
		if(index == string::npos)
		{
			if(!FindNextFile(hFileFind, &FindData))
				break;	// no more file
			continue;
		}

		Original = Original.substr(0, index);
		CleanString(Original);
		if(!Original.length())
		{
			if(!FindNextFile(hFileFind, &FindData))
				break;	// no more file
			continue;
		}

		List->push_back(Original);	// This file is a good candidate

		if(!FindNextFile(hFileFind, &FindData))
			break;	// no more file

	}
	FindClose(hFileFind);	// we close the finder
#else

	char WorkDirHere[NV_MAX_PATHLENGTH];
	strncpy(WorkDirHere, ChildRootDirectory.c_str(), NV_MAX_PATHLENGTH-1);
	if(strlen(WorkDirHere))
	{
		// We remove the last / character
		if(WorkDirHere[strlen(WorkDirHere) - 1] == '/')
			WorkDirHere[strlen(WorkDirHere) - 1] = '\0';
	}
	else
		strcpy(WorkDirHere, ".");	// we are in the current directory

	// open directory file
	DIR *Dir = opendir(WorkDirHere);
	if(Dir == NULL)
		return;		// cannot open directory file
	
	dirent *DirEntry;
	while((DirEntry = readdir(Dir)) != 0)
	{
		if(DirEntry->d_name[0] == '.')
			continue;
		strncpy(FilesToFind, WorkDirHere, NV_MAX_PATHLENGTH-1);
		if(strlen(FilesToFind))
		{
			// We add the last \ character
			if(FilesToFind[strlen(FilesToFind) - 1] != '/')
				strcat(FilesToFind, "/");
		}
		strcat(FilesToFind, DirEntry->d_name);
		DIR *DirHere = NULL;
		if((DirHere = opendir(FilesToFind)) != NULL)
		{
			closedir(DirHere);
			continue;
		}
		// The file name seems to be OK
		// We must still verify that its a lowercase name
		string Original = DirEntry->d_name;
		string Copy = Original;
		MakeLower(Copy);
		if(Copy != Original)
			continue;
		// Now we must verify it's a real registry object file
		// A file has been found
		// get its type
		// And verify several things
		// The file name must be of the form objname_nvdd.obj
		// where dd is the file type on 2 digits
		string::size_type index = Original.rfind("_nv");
		if(index == string::npos)
			continue;
		string sType = Original.substr(index+3);
		index = sType.rfind(".obj");
		if(index == string::npos)
			continue;
		sType = sType.substr(0, index);
		if(sType.length() != 2)
			continue;
		int iType = atoi(sType.c_str());
		if(iType < NV_OBJTYPE_UNKNOWN)
			continue;
		if(iType > NV_OBJTYPE_MAXTYPEVALUE)
			continue;

		index = Original.rfind("_nv");
		if(index == string::npos)
			continue;

		Original = Original.substr(0, index);
		CleanString(Original);
		if(!Original.length())
			continue;
		List->push_back(Original);	// This file is a good candidate
	}
	closedir(Dir);	// close the directory file
#endif
}


void NvContainer::SetParent(NvContainer *Parent)
{
	m_cParent = Parent;
}


// PM Move mode is not used for now. We use the MoveContainer function instead.
bool NvContainer::CopyFrom(NvContainer *Source, int AppendMode, const char *Objects, bool WithSubContainers, bool WithData, const char *FileDir, int PersistTime, bool Recurse, bool Move, bool MoveFiles)
{
	if(Source == NULL)
		return false;

	if(Source == this)
		return true;

	// Move is not authorised for registry
	if(Move && (Source->m_sRootDirectory.length() || m_sRootDirectory.length()))
		return false;

	string FileDirHere = "";
	if(FileDir != NULL)
		FileDirHere = FileDir;
	CleanString(FileDirHere);

	string ObjectsHere = "";
	if(Objects != NULL)
		ObjectsHere = Objects;
	CleanString(ObjectsHere);

	if(PersistTime < 0)
		PersistTime = -1;


	if(AppendMode == NV_CONTAINER_COPY_NO_APPEND)
	{
		// If not append mode, we must clear all container objects
		// With eventual files for registry containers
		ClearObjects(true);

		// and subcontainers if requested
		if(WithSubContainers)
			if(!ClearSubcontainers(true))
				return false;
	}

	int i;

	// starts with objects
	NvObjStringList ObjectsToGet;

	// if Objects is not null, we must select only requested objects (maybe none)
	// Remark: if object is given but blank, no objects will be get
	// and if object is not given (NULL), all valid objects will be get
	if(Objects != NULL)
	{
		while(ObjectsHere.length())
		{
			// Several objects can be in the list
			// They are separated by a semicolumn
			string ObjectName = ObjectsHere;
			string::size_type index = ObjectName.find(';');
			if(index != string::npos)
			{
				ObjectName = ObjectsHere.substr(0, index);
				ObjectsHere = ObjectsHere.substr(index+1);
			}
			else
				ObjectsHere = "";

			// An object name has been found
			CleanString(ObjectName);
			
			// It should not be empty
			if(!ObjectName.length())
				continue;

			// The object should exist in the source
			if(!Source->GetObject(ObjectName.c_str()))
				return false;

			ObjectsToGet.Add(ObjectName);
		}
	}
	else
	{
		// If source is registry, we must get all real objects
		// and not only these ones in the Source->m_mObjects
		vector<string> RealObjNames;

		if(Source->m_sRootDirectory.length())
		{
			// This is a registry container
			Source->GetRealObjectNames(&RealObjNames);
		}
		else
		{
			map<string,void *>::iterator pos;
			for(pos = Source->m_mObjects.begin(); pos != Source->m_mObjects.end(); ++pos)
			{
				NvObj *Object = (NvObj*)(pos->second);
				if(Object == NULL)
					continue;
				RealObjNames.push_back(Object->GetName());
			}
		}
		for(i=0; i<(int)RealObjNames.size(); i++)
			ObjectsToGet.Add(RealObjNames[i]);
	}

	bool RetVal = true;
	string ChildRootDirectory = m_sRootDirectory;
	AddDirSuffix(ChildRootDirectory);
	for(i=0; i<ObjectsToGet.GetSize(); i++)
	{
		string sObjectInList = ObjectsToGet.GetValue(i);

		NvObj* SourceObject = Source->GetObject(sObjectInList.c_str());
		if(SourceObject == NULL)
			continue;	// should not be the case

		NvObj* DestinationObject = GetObject(sObjectInList.c_str());
		if(DestinationObject != NULL)
		{
			// The destination object ever exists
			if(AppendMode == NV_CONTAINER_COPY_APPEND_NO_REPLACE)
				continue;	// We don't copy this object

			if(AppendMode == NV_CONTAINER_COPY_APPEND_REPLACE_IF_BAD_TYPE)
			{
				// We must check the type
				int DestinationType =  DestinationObject->GetType();
				int SourceType =  SourceObject->GetType();
				if(DestinationType == SourceType)
					continue;	// We don't copy this object
			}
			
			// We must remove the object except for registry otherwise in case of crash we may loose registry entry
			if(!ChildRootDirectory.length() || !WithData)
				RemoveObject(sObjectInList.c_str());
		}

		// now we can create the new object
		if(!CreateObject(sObjectInList.c_str(), SourceObject->GetType()))
			return false;

		DestinationObject = GetObject(sObjectInList.c_str());
		if(DestinationObject == NULL)
			return false;	// Should not occur


		if(SourceObject->GetType() == NV_OBJTYPE_BOOLEAN)
		{
			NvObjBoolean *Sobj = (NvObjBoolean*)SourceObject;
			NvObjBoolean *Dobj = (NvObjBoolean*)DestinationObject;
			if(WithData)
				Dobj->CopyFrom(Sobj);
			if(ChildRootDirectory.length())
			{
				// for registry containers, we must save the object to file
				string FileName = GetObjectFile(DestinationObject->GetName().c_str(), NV_OBJTYPE_BOOLEAN);
				if(FileName.length())
					RetVal = SaveObjToFile(FileName.c_str(), Dobj);
			}
		}
		if(SourceObject->GetType() == NV_OBJTYPE_INTEGER)
		{
			NvObjInteger *Sobj = (NvObjInteger*)SourceObject;
			NvObjInteger *Dobj = (NvObjInteger*)DestinationObject;
			if(WithData)
				Dobj->CopyFrom(Sobj);
			if(ChildRootDirectory.length())
			{
				// for registry containers, we must save the object to file
				string FileName = GetObjectFile(DestinationObject->GetName().c_str(), NV_OBJTYPE_INTEGER);
				if(FileName.length())
					RetVal = SaveObjToFile(FileName.c_str(), Dobj);
			}
		}
		if(SourceObject->GetType() == NV_OBJTYPE_STRING)
		{
			NvObjString *Sobj = (NvObjString*)SourceObject;
			NvObjString *Dobj = (NvObjString*)DestinationObject;
			if(WithData)
				Dobj->CopyFrom(Sobj);
			if(ChildRootDirectory.length())
			{
				// for registry containers, we must save the object to file
				string FileName = GetObjectFile(DestinationObject->GetName().c_str(), NV_OBJTYPE_STRING);
				if(FileName.length())
					RetVal = SaveObjToFile(FileName.c_str(), Dobj);
			}
		}
		if(SourceObject->GetType() == NV_OBJTYPE_STRINGLIST)
		{
			NvObjStringList *Sobj = (NvObjStringList*)SourceObject;
			NvObjStringList *Dobj = (NvObjStringList*)DestinationObject;
			if(WithData)
				Dobj->CopyFrom(Sobj);
			if(ChildRootDirectory.length())
			{
				// for registry containers, we must save the object to file
				string FileName = GetObjectFile(DestinationObject->GetName().c_str(), NV_OBJTYPE_STRINGLIST);
				if(FileName.length())
					RetVal = SaveObjToFile(FileName.c_str(), Dobj);
			}
		}
		if(SourceObject->GetType() == NV_OBJTYPE_INDSTRINGLIST)
		{
			NvObjIndStringList *Sobj = (NvObjIndStringList*)SourceObject;
			NvObjIndStringList *Dobj = (NvObjIndStringList*)DestinationObject;
			if(WithData)
				Dobj->CopyFrom(Sobj);
			if(ChildRootDirectory.length())
			{
				// for registry containers, we must save the object to file
				string FileName = GetObjectFile(DestinationObject->GetName().c_str(), NV_OBJTYPE_INDSTRINGLIST);
				if(FileName.length())
					RetVal = SaveObjToFile(FileName.c_str(), Dobj);
			}
		}
		if(SourceObject->GetType() == NV_OBJTYPE_TABLE)
		{
			NvObjTable *Sobj = (NvObjTable*)SourceObject;
			NvObjTable *Dobj = (NvObjTable*)DestinationObject;
			if(WithData)
				Dobj->CopyFrom(Sobj);
			if(ChildRootDirectory.length())
			{
				// for registry containers, we must save the object to file
				string FileName = GetObjectFile(DestinationObject->GetName().c_str(), NV_OBJTYPE_TABLE);
				if(FileName.length())
					RetVal = SaveObjToFile(FileName.c_str(), Dobj);
			}
		}
		if(SourceObject->GetType() == NV_OBJTYPE_FILE)
		{
			NvObjFile *Sobj = (NvObjFile*)SourceObject;
			NvObjFile *Dobj = (NvObjFile*)DestinationObject;
			Dobj->SetPersistTime(Sobj->m_iPersistTime);
			if(ChildRootDirectory.length())
			{
				// The file is always persistant here because it's a registry file
				Dobj->SetPersistTime();
				// for registry containers, we must save the object to file
				string FileName = GetObjectFile(DestinationObject->GetName().c_str(), NV_OBJTYPE_FILE);
				
				// The object filename is directly the registry file
				Dobj->SetFileName(FileName);
				if(FileName.length())
				{
					if(WithData)
						RetVal = SaveObjToFile(FileName.c_str(), Dobj, Sobj);
					else
						Dobj->Create();
				}
			}
			else if(FileDirHere.length())
			{
				if(Move && !MoveFiles)
				{
					RetVal = Dobj->MoveFrom(Sobj);
				}
				else
				{
					if(Dobj->m_iPersistTime == -1)				// The file is persistant
						Dobj->SetPersistTime(PersistTime);		// we change that to the new persistant value
					RetVal = Dobj->SetAutoFileName(FileDirHere, Sobj->GetFileExtension());
					if(RetVal && WithData)
						RetVal = Dobj->CopyFrom(Sobj);
				}
			}
		}
		if(SourceObject->GetType() == NV_OBJTYPE_BINARY)
		{
			NvObjBinary *Sobj = (NvObjBinary*)SourceObject;
			NvObjBinary *Dobj = (NvObjBinary*)DestinationObject;
			if(WithData)
				RetVal = Dobj->CopyFrom(Sobj);
			if(RetVal && ChildRootDirectory.length())
			{
				// for registry containers, we must save the object to file
				string FileName = GetObjectFile(DestinationObject->GetName().c_str(), NV_OBJTYPE_BINARY);
				if(FileName.length())
					RetVal = SaveObjToFile(FileName.c_str(), Dobj);
			}
		}

		if(!RetVal)
		{
			RemoveObject(sObjectInList.c_str());
			return false;
		}

		if(Move)
			Source->RemoveObject(sObjectInList.c_str());
	}
	
	if(!RetVal)
		return false;

	if(WithSubContainers)
	{
		NvContainer *SourceContainer = NULL;
		NvContainer *DestinationContainer = NULL;

		vector<string> SubNames;

		// For registry source, we must take all real subcontainers
		if(Source->m_sRootDirectory.length())
		{
			// This is a registry container
			Source->GetRealSubContainerNames(&SubNames);
		}
		else
		{
			map<string,void *>::iterator pos;
			for(pos = Source->m_mSubcontainers.begin(); pos != Source->m_mSubcontainers.end(); ++pos)
			{
				SourceContainer = (NvContainer *)(pos->second);
				if(SourceContainer == NULL)
					continue;
				SubNames.push_back(SourceContainer->GetName());
			}
		}
		for(i=0; i<(int)SubNames.size(); i++)
		{
			SourceContainer = Source->GetSubContainer(SubNames[i].c_str());
			if(SourceContainer == NULL)
				continue;
			
			DestinationContainer = GetSubContainer(SourceContainer->GetName().c_str());
			if(DestinationContainer == NULL)
			{
				if(!CreateSubContainer(SourceContainer->GetName().c_str()))
				{
					RetVal = false;
					break;
				}
			}

			if(Recurse)
			{
				DestinationContainer = GetSubContainer(SourceContainer->GetName().c_str());
				if(DestinationContainer == NULL)
				{
					// should not occur
					RetVal = false;
					break;
				}

				// Recursive function that processes all subcontainers
				RetVal = DestinationContainer->CopyFrom(SourceContainer, AppendMode, NULL, true, WithData, FileDir, PersistTime);

				if(!RetVal)
					break;

				if(Move)
					Source->RemoveSubContainer(SubNames[i].c_str());
			}

		}
	}

	return RetVal; 

}

bool NvContainer::SaveObjToFile(const char *FileName, NvObj *Dobj, NvObj *Sobj)
{
	bool RetVal = true;
	int Type = Dobj->GetType();
	string FileNameHere = GetObjectFile(Dobj->GetName().c_str(), Type);
	if(FileNameHere.length())
	{
		string FileNameTmp = FileName;
		FileNameTmp += ".nvtmp";
		unlink(FileNameTmp.c_str());
		switch(Type)
		{
			case NV_OBJTYPE_BOOLEAN:
				RetVal = ((NvObjBoolean *)Dobj)->SaveToFile(FileNameTmp.c_str());
				break;
			case NV_OBJTYPE_INTEGER:
				RetVal = ((NvObjInteger *)Dobj)->SaveToFile(FileNameTmp.c_str());
				break;
			case NV_OBJTYPE_STRING:
				RetVal = ((NvObjString *)Dobj)->SaveToFile(FileNameTmp.c_str());
				break;
			case NV_OBJTYPE_STRINGLIST:
				RetVal = ((NvObjStringList *)Dobj)->SaveToFile(FileNameTmp.c_str());
				break;
			case NV_OBJTYPE_INDSTRINGLIST:
				RetVal = ((NvObjIndStringList *)Dobj)->SaveToFile(FileNameTmp.c_str());
				break;
			case NV_OBJTYPE_TABLE:
				RetVal = ((NvObjTable *)Dobj)->SaveToFile(FileNameTmp.c_str());
				break;
			case NV_OBJTYPE_FILE:
				((NvObjFile *)Dobj)->SetFileName(FileNameTmp.c_str());
				RetVal = ((NvObjFile *)Dobj)->CopyFrom((NvObjFile *)Sobj);
				((NvObjFile *)Dobj)->SetFileName(FileName);
				break;
			case NV_OBJTYPE_BINARY:
				RetVal = ((NvObjBinary *)Dobj)->SaveToFile(FileNameTmp.c_str());
				break;
		}
		if(RetVal)
		{
			unlink(FileNameHere.c_str());
			rename(FileNameTmp.c_str(), FileNameHere.c_str());
		}
	}
	return RetVal;
}



#ifdef NV_SERVER
void NvContainer::Encode(bool bSrcUnicode, bool bDestUnicode, bool bWithSubcontainers, bool bWithNames)
{
//	int i;

	// starts with objects
	map<string,void *>::iterator pos;
	for(pos = m_mObjects.begin(); pos != m_mObjects.end(); ++pos)
	{
		NvObj *Object = (NvObj*)(pos->second);
		if(Object == NULL)
			continue;
		if(bWithNames)
			Object->EncodeName(bSrcUnicode, bDestUnicode);
		if(Object->GetType() == NV_OBJTYPE_STRING)
			((NvObjString*)Object)->Encode(bSrcUnicode, bDestUnicode);
		if(Object->GetType() == NV_OBJTYPE_STRINGLIST)
			((NvObjStringList*)Object)->Encode(bSrcUnicode, bDestUnicode);
		if(Object->GetType() == NV_OBJTYPE_INDSTRINGLIST)
			((NvObjIndStringList*)Object)->Encode(bSrcUnicode, bDestUnicode);
		if(Object->GetType() == NV_OBJTYPE_TABLE)
			((NvObjTable*)Object)->Encode(bSrcUnicode, bDestUnicode);
	}

	if(bWithSubcontainers)
	{
		map<string,void *>::iterator pos;
		for(pos = m_mSubcontainers.begin(); pos != m_mSubcontainers.end(); ++pos)
		{
			NvContainer *Container = (NvContainer *)(pos->second);
			if(Container == NULL)
				continue;
			Container->Encode(bSrcUnicode, bDestUnicode, bWithSubcontainers, bWithNames);
		}
	}

	if(bWithNames)
		EncodeName(bSrcUnicode, bDestUnicode);
}
#endif

bool NvContainer::Synchronize(int Mode)
{
	if(!m_sRootDirectory.length())
		return true;	// Nothing to do for a memory container
	
	// We create the root directory if it doesn't exist
	if(!CreateDir(m_sRootDirectory.c_str()))
		return false;

	bool RetVal = true;
//	int i;
	if(Mode == NV_CONTAINER_SYNCHRONIZE_MEMTODISK)
	{
		map<string,void *>::iterator pos;
		for(pos = m_mObjects.begin(); pos != m_mObjects.end(); ++pos)
		{
			NvObj *Object = (NvObj*)(pos->second);
			if(Object == NULL)
				continue;

			if(Object->GetType() == NV_OBJTYPE_BOOLEAN)
			{
				NvObjBoolean *obj = (NvObjBoolean*)Object;
				string FileName = GetObjectFile(Object->GetName().c_str(), NV_OBJTYPE_BOOLEAN);
				if(FileName.length())
					RetVal = SaveObjToFile(FileName.c_str(), obj);
			}
			if(Object->GetType() == NV_OBJTYPE_INTEGER)
			{
				NvObjInteger *obj = (NvObjInteger*)Object;
				string FileName = GetObjectFile(Object->GetName().c_str(), NV_OBJTYPE_INTEGER);
				if(FileName.length())
					RetVal = SaveObjToFile(FileName.c_str(), obj);
			}
			if(Object->GetType() == NV_OBJTYPE_STRING)
			{
				NvObjString *obj = (NvObjString*)Object;
				string FileName = GetObjectFile(Object->GetName().c_str(), NV_OBJTYPE_STRING);
				if(FileName.length())
					RetVal = SaveObjToFile(FileName.c_str(), obj);
			}
			if(Object->GetType() == NV_OBJTYPE_STRINGLIST)
			{
				NvObjStringList *obj = (NvObjStringList*)Object;
				string FileName = GetObjectFile(Object->GetName().c_str(), NV_OBJTYPE_STRINGLIST);
				if(FileName.length())
					RetVal = SaveObjToFile(FileName.c_str(), obj);
			}
			if(Object->GetType() == NV_OBJTYPE_INDSTRINGLIST)
			{
				NvObjIndStringList *obj = (NvObjIndStringList*)Object;
				string FileName = GetObjectFile(Object->GetName().c_str(), NV_OBJTYPE_INDSTRINGLIST);
				if(FileName.length())
					RetVal = SaveObjToFile(FileName.c_str(), obj);
			}
			if(Object->GetType() == NV_OBJTYPE_TABLE)
			{
				NvObjTable *obj = (NvObjTable*)Object;
				string FileName = GetObjectFile(Object->GetName().c_str(), NV_OBJTYPE_TABLE);
				if(FileName.length())
					RetVal = SaveObjToFile(FileName.c_str(), obj);
			}
			if(Object->GetType() == NV_OBJTYPE_FILE)
			{
				NvObjFile *obj = (NvObjFile*)Object;
				string FileName = GetObjectFile(Object->GetName().c_str(), NV_OBJTYPE_FILE);
				if(FileName.length())
				{
					string FileNameTmp = FileName;
					FileNameTmp += ".nvtmp";
					unlink(FileNameTmp.c_str());
					if(!CopyFileEx(obj->GetFilePathName().c_str(), FileNameTmp.c_str()))
						RetVal = false;
					else
					{
						unlink(FileName.c_str());
						rename(FileNameTmp.c_str(), FileName.c_str());
					}
				}
			}
			if(Object->GetType() == NV_OBJTYPE_BINARY)
			{
				NvObjBinary *obj = (NvObjBinary*)Object;
				string FileName = GetObjectFile(Object->GetName().c_str(), NV_OBJTYPE_BINARY);
				if(FileName.length())
					RetVal = SaveObjToFile(FileName.c_str(), obj);
			}
			
			if(!RetVal)
				break;
		}

		if(!RetVal)
			return false;

		for(pos = m_mSubcontainers.begin(); pos != m_mSubcontainers.end(); ++pos)
		{
			NvContainer *Container = (NvContainer *)(pos->second);
			if(Container == NULL)
				continue;
			RetVal = Container->Synchronize(Mode);
			if(!RetVal)
				break;
		}
	}
	else if(Mode == NV_CONTAINER_SYNCHRONIZE_DISKTOMEM)
	{
		vector<string> RealObjNames;
		GetRealObjectNames(&RealObjNames);
		string SaveRootDirectory = m_sRootDirectory;
		int i;
		for(i=0; i<(int)RealObjNames.size(); i++)
		{
			// We reset temporarly the root directory name in order to remove the memory object
			// But not the file object
			m_sRootDirectory = "";
			RemoveObject(RealObjNames[i].c_str());

			// Now we can recreate the object with information from file
			m_sRootDirectory = SaveRootDirectory;
			GetObject(RealObjNames[i].c_str());
		}

		// Do it now for subcontainers
		vector<string> SubNames;
		GetRealSubContainerNames(&SubNames);
		NvContainer *Container = NULL;
		for(i=0; i<(int)SubNames.size(); i++)
		{
			Container = GetSubContainer(SubNames[i].c_str());
			if(Container == NULL)
				continue;
			
			RetVal = Container->Synchronize(Mode);
			if(!RetVal)
				break;

		}
	}
	else
		return false; // bad method

	

	return RetVal; 

}



void NvContainer::Clear()
{
	ClearSubcontainers();
	NvLocalContainer::Clear();
}



bool NvContainer::IsEmpty()
{
	bool RetVal = false;
	if(!m_mSubcontainers.size())
		if(!m_mObjects.size())
			RetVal = true;
	return RetVal;
}

bool NvContainer::ClearSubcontainers(bool WithFiles)
{
	map<string,void *>::iterator pos;
	for(pos = m_mSubcontainers.begin(); pos != m_mSubcontainers.end(); ++pos)
	{
		NvContainer *Container = (NvContainer *)(pos->second);
		if(Container != NULL)
		{
			if(WithFiles)
			{
				string ChildRootDirectory = Container->m_sRootDirectory;
				AddDirSuffix(ChildRootDirectory);
				if(ChildRootDirectory.length())
				{
					// This is a registry container
					// We must remove the real files
					RemoveDir(ChildRootDirectory.c_str());
//					if(!RemoveDir(ChildRootDirectory.c_str()))
//						return false;
				}
			}
			delete Container;
		}
	}
	m_mSubcontainers.clear();
	if(bUseLastSubContainer)
		LastSubContainer = NULL;
	return true;
}


NvContainer* NvContainer::GetParent()
{
	return m_cParent;
}


NvContainer* NvContainer::GetSubContainer(const char *Name)
{
	if(Name == NULL)
		return NULL;
	string NameHere = Name;
	NameHere = NvContainerList::CleanContainerName(NameHere);


	// Last subcontainer cache mechanism
	if((bUseLastSubContainer) && (LastSubContainer != NULL) && (LastSubContainer->GetName() == NameHere))
	{
		if(LastSubContainer->m_sRootDirectory.length())
		{
			// We try to create the subdirectory
			if(!CreateDir(LastSubContainer->m_sRootDirectory.c_str()))
				return NULL;
		}
		return LastSubContainer;
	}

	NvContainer *ReturnedContainer = NULL;
	if(m_mSubcontainers.find(NameHere) != m_mSubcontainers.end())
		ReturnedContainer = (NvContainer *)(m_mSubcontainers[NameHere]);

	if(ReturnedContainer == NULL)
	{
		// The subcontainer doesn't exist
		// For a registry container, we try to see if the corresponding directory exist
		string ChildRootDirectory = m_sRootDirectory;
		AddDirSuffix(ChildRootDirectory);
		if(ChildRootDirectory.length())
		{
			ChildRootDirectory += NameHere;
			if(DirExist(ChildRootDirectory.c_str()))
			{
				// The subcontainer exist in directory
				// so we add it to the list
				NvContainer *Container = NVNEW NvContainer(this);
				if(Container == NULL)
					return NULL;
				if(!Container->SetName(NameHere))
				{
					delete Container;
					return NULL;
				}
				Container->SetRootDirectory(ChildRootDirectory.c_str());
				m_mSubcontainers[Container->GetName()] = (void *)Container;
				ReturnedContainer = Container;
			}
		}
	}

	if(ReturnedContainer != NULL)
	{
		if(bUseLastSubContainer)
			LastSubContainer = ReturnedContainer;
		if(ReturnedContainer->m_sRootDirectory.length())
		{
			// We try to create the subdirectory
			if(!CreateDir(ReturnedContainer->m_sRootDirectory.c_str()))
				return NULL;
		}
	}
	return ReturnedContainer;
}



bool NvContainer::CreateSubContainer(const char *Name)
{
	if(Name == NULL)
		return false;

	if(!strlen(Name))
		return false;

	// Try to see if the subcontainer ever exist
	NvContainer* Container = GetSubContainer(Name);
	if(Container != NULL)
		return true;

	Container = NVNEW NvContainer(this);
	if(Container == NULL)
		return false;

	if(!Container->SetName(Name))
	{
		delete Container;
		return false;
	}

	Container->bUseLastObject = bUseLastObject;
	Container->bUseLastSubContainer = bUseLastSubContainer;
	
	string ChildRootDirectory = m_sRootDirectory;
	AddDirSuffix(ChildRootDirectory);
	// if this is a registry, we must create the corresponding subdirectory
	if(ChildRootDirectory.length())
	{
		ChildRootDirectory += Container->GetName();
		Container->SetRootDirectory(ChildRootDirectory.c_str());
		// We try to create the subdirectory
		if(!CreateDir(Container->m_sRootDirectory.c_str()))
		{
			delete Container;
			return false;
		}
	}
	m_mSubcontainers[Container->GetName()] = (void *)Container;

	if(bUseLastSubContainer)
		LastSubContainer = Container;
	return true;
}


bool NvContainer::RemoveSubContainer(const char *Name)
{
	if(Name == NULL)
		return false;

	if(!strlen(Name))
		return false;

	string NameHere = Name;

	NvContainer *Container = NULL;
	string ChildRootDirectory = "";
	if(m_mSubcontainers.find(NameHere) != m_mSubcontainers.end())
		Container = (NvContainer *)(m_mSubcontainers[NameHere]);

	if(Container != NULL)
	{
		ChildRootDirectory = Container->m_sRootDirectory;
		AddDirSuffix(ChildRootDirectory);
		m_mSubcontainers.erase(Container->GetName());
		delete Container;
		// The last subcontainer must be reset if it's the deleted one
		if(bUseLastSubContainer && (Container == LastSubContainer))
			LastSubContainer = NULL;
	}

	if(ChildRootDirectory.length())
	{
		// This is a registry container
		// We must remove the real files
		RemoveDir(ChildRootDirectory.c_str());
//		if(!RemoveDir(ChildRootDirectory.c_str()))
//			return false;
	}

	return true;
}

#ifdef NV_SERVER
bool NvContainer::GetXmlObjectDataToStream(const char *Name, FILE *stream, int NumTabs, bool WithSubContainers, bool WithData, bool WithFileData, string NsPrefix, int XmlCoding, NvContainer *Model, void *XmlPtr, bool bUnicode, bool bWebs)
{
	NvXml *XmlPtrHere = (NvXml*)XmlPtr;
	string Tabs = "";
	int i;
	for(i=0; i<NumTabs; i++)
		Tabs += '\t';
	

	bool AllObjects = false;
	string NameHere = "";
	if(Name != NULL)
		NameHere = Name;
	CleanString(NameHere);
	if(NameHere == "NV_ALLCONTAINER_OBJECTS")
		AllObjects = true;


	NvObjStringList ObjectsToGet;
	NvObj *Object;
	if(AllObjects)
	{
		if(Model == NULL)
		{
			map<string,void *>::iterator pos;
			for(pos = m_mObjects.begin(); pos != m_mObjects.end(); ++pos)
			{
				Object = (NvObj*)(pos->second);
				if(Object != NULL)
					ObjectsToGet.Add(Object->GetName());
			}
		}
		else
		{
			map<string,void *>::iterator pos;
			for(pos = Model->m_mObjects.begin(); pos != Model->m_mObjects.end(); ++pos)
			{
				Object = (NvObj*)(pos->second);
				if(Object != NULL)
					ObjectsToGet.Add(Object->GetName());
			}
		}
	}
	else
	{
		if(NameHere.length())
			ObjectsToGet.Add(NameHere);
	}


	bool RetVal = true;
	for(i=0; i<ObjectsToGet.GetSize(); i++)
	{
		Object = GetObject(ObjectsToGet.GetValue(i).c_str());
		
		if(Object == NULL)
		{
			if(Model == NULL)
				return false;	// The object doesn't exists

			// For a web service, we create the object
			Object = Model->GetObject(ObjectsToGet.GetValue(i).c_str());
			if(Object == NULL)
				return false;
			int ObjectType = Object->GetType();
			CreateObject(ObjectsToGet.GetValue(i).c_str(), ObjectType);
			Object = GetObject(ObjectsToGet.GetValue(i).c_str());
			if(Object == NULL)
				return false;		// should not occur here
		}

		if(Object->GetType() == NV_OBJTYPE_BOOLEAN)
		{
			NvObjBoolean *ObjectHere = (NvObjBoolean *)Object;
			RetVal = ObjectHere->GetXmlDataToStream(stream, NumTabs+1, WithData, NsPrefix, XmlCoding, XmlPtr, bUnicode, bWebs);
		}
		if(Object->GetType() == NV_OBJTYPE_INTEGER)
		{
			NvObjInteger *ObjectHere = (NvObjInteger *)Object;
			RetVal = ObjectHere->GetXmlDataToStream(stream, NumTabs+1, WithData, NsPrefix, XmlCoding, XmlPtr, bUnicode, bWebs);
		}
		if(Object->GetType() == NV_OBJTYPE_STRING)
		{
			NvObjString *ObjectHere = (NvObjString *)Object;
			RetVal = ObjectHere->GetXmlDataToStream(stream, NumTabs+1, WithData, NsPrefix, XmlCoding, XmlPtr, bUnicode, bWebs);
		}
		if(Object->GetType() == NV_OBJTYPE_STRINGLIST)
		{
			NvObjStringList *ObjectHere = (NvObjStringList *)Object;
			RetVal = ObjectHere->GetXmlDataToStream(stream, NumTabs+1, WithData, NsPrefix, XmlCoding, XmlPtr, bUnicode, bWebs);
		}
		if(Object->GetType() == NV_OBJTYPE_INDSTRINGLIST)
		{
			NvObjIndStringList *ObjectHere = (NvObjIndStringList *)Object;
			RetVal = ObjectHere->GetXmlDataToStream(stream, NumTabs+1, WithData, NsPrefix, XmlCoding, XmlPtr, bUnicode, bWebs);
		}
		if(Object->GetType() == NV_OBJTYPE_TABLE)
		{
			NvObjTable *ObjectHere = (NvObjTable *)Object;
			RetVal = ObjectHere->GetXmlDataToStream(stream, NumTabs+1, WithData, NsPrefix, XmlCoding, XmlPtr, bUnicode, bWebs);
		}
		if(Object->GetType() == NV_OBJTYPE_FILE)
		{
			NvObjFile *ObjectHere = (NvObjFile *)Object;
			RetVal = ObjectHere->GetXmlDataToStream(stream, NumTabs+1, WithFileData, NsPrefix, XmlCoding, XmlPtr, bUnicode, bWebs);
		}
		if(Object->GetType() == NV_OBJTYPE_BINARY)
		{
			NvObjBinary *ObjectHere = (NvObjBinary *)Object;
			RetVal = ObjectHere->GetXmlDataToStream(stream, NumTabs+1, WithFileData, NsPrefix, XmlCoding, XmlPtr, bUnicode, bWebs);
		}

		if(!RetVal)
			break;
	}

	if(!RetVal)
		return false;

	if(WithSubContainers)
	{
		
		map<std::string,void *> *m_mSubcontainersHere = &m_mSubcontainers;
		if(Model != NULL)
			m_mSubcontainersHere = &(Model->m_mSubcontainers);
		map<string,void *>::iterator pos;
		NvContainer *SubContainer;
		NvContainer *ModelSubContainer;
		for(pos = m_mSubcontainersHere->begin(); pos != m_mSubcontainersHere->end(); ++pos)
		{
			ModelSubContainer = NULL;
			if(Model == NULL)
				SubContainer = (NvContainer*)(pos->second);
			else
			{
				ModelSubContainer = (NvContainer*)(pos->second);
				SubContainer = GetSubContainer(ModelSubContainer->GetName().c_str());
				if(SubContainer == NULL)
				{
					// We create the subcontainer if not existing
					CreateSubContainer(ModelSubContainer->GetName().c_str());
					SubContainer = GetSubContainer(ModelSubContainer->GetName().c_str());
					if(SubContainer == NULL)
						return false;
				}
			}
			string XmlData = Tabs;
			XmlData += "<";
			XmlData += NsPrefix;
			if(XmlCoding == XML_MODEL_SIMPLE)
			{
				XmlData += XmlPtrHere->EncodeOutputString(SubContainer->GetName(), bUnicode);
				XmlData += " type=\"container\"";
			}
			else if(XmlCoding == XML_MODEL_COMPACT)
			{
				XmlData += XmlPtrHere->EncodeOutputString(SubContainer->GetName(), bUnicode);
				XmlData += " type=\"container\"";
			}
			else if(XmlCoding == XML_MODEL_TINY)
			{
				XmlData += XmlPtrHere->EncodeOutputString(SubContainer->GetName(), bUnicode);
			}
			else
			{
				XmlData += "NVCONTAINER name=\"";
				XmlData += XmlPtrHere->EncodeOutputString(SubContainer->GetName(), bUnicode);
				XmlData += "\"";
			}
			XmlData += ">\n";
			if(fwrite(XmlData.c_str(), 1, XmlData.length(), stream) != XmlData.length())
				return false;
			RetVal = SubContainer->GetXmlObjectDataToStream("NV_ALLCONTAINER_OBJECTS", stream, NumTabs+1, true, WithData, WithFileData, NsPrefix, XmlCoding, ModelSubContainer, XmlPtr, bUnicode, bWebs);
			if(!RetVal)
				break;
			XmlData = Tabs;
			XmlData += "</";
			XmlData += NsPrefix;
			if(XmlCoding == XML_MODEL_SIMPLE)
				XmlData += XmlPtrHere->EncodeOutputString(SubContainer->GetName(), bUnicode);
			else if(XmlCoding == XML_MODEL_COMPACT)
				XmlData += XmlPtrHere->EncodeOutputString(SubContainer->GetName(), bUnicode);
			else if(XmlCoding == XML_MODEL_TINY)
				XmlData += XmlPtrHere->EncodeOutputString(SubContainer->GetName(), bUnicode);
			else
				XmlData += "NVCONTAINER";
			XmlData += ">\n";
			if(fwrite(XmlData.c_str(), 1, XmlData.length(), stream) != XmlData.length())
				return false;
		}
	}


	return RetVal;
}


bool NvContainer::GetXmlObjectDataToString(const char *Name, string& Output, int NumTabs, bool WithSubContainers, bool WithData, bool WithFileData, string NsPrefix, int XmlCoding, NvContainer *Model, void *XmlPtr, bool bUnicode, bool bWebs)
{
	NvXml *XmlPtrHere = (NvXml*)XmlPtr;
	string Tabs = "";
	int i;
	for(i=0; i<NumTabs; i++)
		Tabs += '\t';
	

	bool AllObjects = false;
	string NameHere = "";
	if(Name != NULL)
		NameHere = Name;
	CleanString(NameHere);
	if(NameHere == "NV_ALLCONTAINER_OBJECTS")
		AllObjects = true;


	NvObjStringList ObjectsToGet;
	NvObj *Object;
	if(AllObjects)
	{
		if(Model == NULL)
		{
			map<string,void *>::iterator pos;
			for(pos = m_mObjects.begin(); pos != m_mObjects.end(); ++pos)
			{
				Object = (NvObj*)(pos->second);
				if(Object != NULL)
					ObjectsToGet.Add(Object->GetName());
			}
		}
		else
		{
			map<string,void *>::iterator pos;
			for(pos = Model->m_mObjects.begin(); pos != Model->m_mObjects.end(); ++pos)
			{
				Object = (NvObj*)(pos->second);
				if(Object != NULL)
					ObjectsToGet.Add(Object->GetName());
			}
		}
	}
	else
	{
		if(NameHere.length())
			ObjectsToGet.Add(NameHere);
	}

	bool RetVal = true;
	for(i=0; i<ObjectsToGet.GetSize(); i++)
	{
		Object = GetObject(ObjectsToGet.GetValue(i).c_str());
		
		if(Object == NULL)
		{
			if(Model == NULL)
				return false;	// The object doesn't exists

			// For a web service, we create the object
			Object = Model->GetObject(ObjectsToGet.GetValue(i).c_str());
			if(Object == NULL)
				return false;
			int ObjectType = Object->GetType();
			CreateObject(ObjectsToGet.GetValue(i).c_str(), ObjectType);
			Object = GetObject(ObjectsToGet.GetValue(i).c_str());
			if(Object == NULL)
				return false;		// should not occur here
		}

		if(Object->GetType() == NV_OBJTYPE_BOOLEAN)
		{
			NvObjBoolean *ObjectHere = (NvObjBoolean *)Object;
			RetVal = ObjectHere->GetXmlDataToString(Output, NumTabs+1, WithData, NsPrefix, XmlCoding, XmlPtr, bUnicode, bWebs);
		}
		if(Object->GetType() == NV_OBJTYPE_INTEGER)
		{
			NvObjInteger *ObjectHere = (NvObjInteger *)Object;
			RetVal = ObjectHere->GetXmlDataToString(Output, NumTabs+1, WithData, NsPrefix, XmlCoding, XmlPtr, bUnicode, bWebs);
		}
		if(Object->GetType() == NV_OBJTYPE_STRING)
		{
			NvObjString *ObjectHere = (NvObjString *)Object;
			RetVal = ObjectHere->GetXmlDataToString(Output, NumTabs+1, WithData, NsPrefix, XmlCoding, XmlPtr, bUnicode, bWebs);
		}
		if(Object->GetType() == NV_OBJTYPE_STRINGLIST)
		{
			NvObjStringList *ObjectHere = (NvObjStringList *)Object;
			RetVal = ObjectHere->GetXmlDataToString(Output, NumTabs+1, WithData, NsPrefix, XmlCoding, XmlPtr, bUnicode, bWebs);
		}
		if(Object->GetType() == NV_OBJTYPE_INDSTRINGLIST)
		{
			NvObjIndStringList *ObjectHere = (NvObjIndStringList *)Object;
			RetVal = ObjectHere->GetXmlDataToString(Output, NumTabs+1, WithData, NsPrefix, XmlCoding, XmlPtr, bUnicode, bWebs);
		}
		if(Object->GetType() == NV_OBJTYPE_TABLE)
		{
			NvObjTable *ObjectHere = (NvObjTable *)Object;
			RetVal = ObjectHere->GetXmlDataToString(Output, NumTabs+1, WithData, NsPrefix, XmlCoding, XmlPtr, bUnicode, bWebs);
		}
		if(Object->GetType() == NV_OBJTYPE_FILE)
		{
			NvObjFile *ObjectHere = (NvObjFile *)Object;
			RetVal = ObjectHere->GetXmlDataToString(Output, NumTabs+1, WithFileData, NsPrefix, XmlCoding, XmlPtr, bUnicode, bWebs);
		}
		if(Object->GetType() == NV_OBJTYPE_BINARY)
		{
			NvObjBinary *ObjectHere = (NvObjBinary *)Object;
			RetVal = ObjectHere->GetXmlDataToString(Output, NumTabs+1, WithFileData, NsPrefix, XmlCoding, XmlPtr, bUnicode, bWebs);
		}

		if(!RetVal)
			break;
	}

	if(!RetVal)
		return false;

	if(WithSubContainers)
	{
		map<std::string,void *> *m_mSubcontainersHere = &m_mSubcontainers;
		if(Model != NULL)
			m_mSubcontainersHere = &(Model->m_mSubcontainers);
		map<string,void *>::iterator pos;

		NvContainer *SubContainer;
		NvContainer *ModelSubContainer = NULL;
		for(pos = m_mSubcontainersHere->begin(); pos != m_mSubcontainersHere->end(); ++pos)
		{
			ModelSubContainer = NULL;
			if(Model == NULL)
				SubContainer = (NvContainer*)(pos->second);
			else
			{
				ModelSubContainer = (NvContainer*)(pos->second);
				SubContainer = GetSubContainer(ModelSubContainer->GetName().c_str());
				if(SubContainer == NULL)
				{
					// We create the subcontainer if not existing
					CreateSubContainer(ModelSubContainer->GetName().c_str());
					SubContainer = GetSubContainer(ModelSubContainer->GetName().c_str());
					if(SubContainer == NULL)
						return false;
				}
			}
			string XmlData = Tabs;
			XmlData += "<";
			XmlData += NsPrefix;
			if(XmlCoding == XML_MODEL_SIMPLE)
			{
				XmlData += XmlPtrHere->EncodeOutputString(SubContainer->GetName(), bUnicode);
				XmlData += " type=\"container\"";
			}
			else if(XmlCoding == XML_MODEL_COMPACT)
			{
				XmlData += XmlPtrHere->EncodeOutputString(SubContainer->GetName(), bUnicode);
				XmlData += " type=\"container\"";
			}
			else if(XmlCoding == XML_MODEL_TINY)
			{
				XmlData += XmlPtrHere->EncodeOutputString(SubContainer->GetName(), bUnicode);
			}
			else
			{
				XmlData += "NVCONTAINER name=\"";
				XmlData += XmlPtrHere->EncodeOutputString(SubContainer->GetName(), bUnicode);
				XmlData += "\"";
			}
			XmlData += ">\n";
			Output += XmlData;
			RetVal = SubContainer->GetXmlObjectDataToString("NV_ALLCONTAINER_OBJECTS", Output, NumTabs+1, true, WithData, WithFileData, NsPrefix, XmlCoding, ModelSubContainer, XmlPtr, bUnicode, bWebs);
			if(!RetVal)
				break;
			XmlData = Tabs;
			XmlData += "</";
			XmlData += NsPrefix;
			if(XmlCoding == XML_MODEL_SIMPLE)
				XmlData += XmlPtrHere->EncodeOutputString(SubContainer->GetName(), bUnicode);
			else if(XmlCoding == XML_MODEL_COMPACT)
				XmlData += XmlPtrHere->EncodeOutputString(SubContainer->GetName(), bUnicode);
			else if(XmlCoding == XML_MODEL_TINY)
				XmlData += XmlPtrHere->EncodeOutputString(SubContainer->GetName(), bUnicode);
			else
				XmlData += "NVCONTAINER";
			XmlData += ">\n";
			Output += XmlData;
		}
	}


	return RetVal;
}


bool NvContainer::GetXmlObjectDataSize(const char *Name, NvInt64& Size, bool WithSubContainers, bool WithFileData, NvContainer *Model)
{
	bool AllObjects = false;
	string NameHere = "";
	if(Name != NULL)
		NameHere = Name;
	CleanString(NameHere);
	if(NameHere == "NV_ALLCONTAINER_OBJECTS")
		AllObjects = true;


	NvObjStringList ObjectsToGet;
	NvObj *Object;
	int i;
	if(AllObjects)
	{
		if(Model == NULL)
		{
			map<string,void *>::iterator pos;
			for(pos = m_mObjects.begin(); pos != m_mObjects.end(); ++pos)
			{
				Object = (NvObj*)(pos->second);
				if(Object != NULL)
					ObjectsToGet.Add(Object->GetName());
			}
		}
		else
		{
			map<string,void *>::iterator pos;
			for(pos = Model->m_mObjects.begin(); pos != Model->m_mObjects.end(); ++pos)
			{
				Object = (NvObj*)(pos->second);
				if(Object != NULL)
					ObjectsToGet.Add(Object->GetName());
			}
		}
	}
	else
	{
		if(NameHere.length())
			ObjectsToGet.Add(NameHere);
	}

	bool RetVal = true;
	for(i=0; i<ObjectsToGet.GetSize(); i++)
	{
		Object = GetObject(ObjectsToGet.GetValue(i).c_str());
		
		if(Object == NULL)
			return false;	// The object doesn't exists

		if(Object->GetType() == NV_OBJTYPE_BOOLEAN)
		{
			NvObjBoolean *ObjectHere = (NvObjBoolean *)Object;
			Size += (NvInt64)ObjectHere->GetXmlDataSize();
		}
		if(Object->GetType() == NV_OBJTYPE_INTEGER)
		{
			NvObjInteger *ObjectHere = (NvObjInteger *)Object;
			Size += (NvInt64)ObjectHere->GetXmlDataSize();
		}
		if(Object->GetType() == NV_OBJTYPE_STRING)
		{
			NvObjString *ObjectHere = (NvObjString *)Object;
			Size += (NvInt64)ObjectHere->GetXmlDataSize();
		}
		if(Object->GetType() == NV_OBJTYPE_STRINGLIST)
		{
			NvObjStringList *ObjectHere = (NvObjStringList *)Object;
			Size += (NvInt64)ObjectHere->GetXmlDataSize();
		}
		if(Object->GetType() == NV_OBJTYPE_INDSTRINGLIST)
		{
			NvObjIndStringList *ObjectHere = (NvObjIndStringList *)Object;
			Size += (NvInt64)ObjectHere->GetXmlDataSize();
		}
		if(Object->GetType() == NV_OBJTYPE_TABLE)
		{
			NvObjTable *ObjectHere = (NvObjTable *)Object;
			Size += (NvInt64)ObjectHere->GetXmlDataSize();
		}
		if(Object->GetType() == NV_OBJTYPE_FILE)
		{
			NvObjFile *ObjectHere = (NvObjFile *)Object;
			Size += (NvInt64)ObjectHere->GetXmlDataSize();
			if(WithFileData)
				Size += (NvInt64)ObjectHere->GetSize();
		}
		if(Object->GetType() == NV_OBJTYPE_BINARY)
		{
			NvObjBinary *ObjectHere = (NvObjBinary *)Object;
			Size += (NvInt64)ObjectHere->GetXmlDataSize();
			if(WithFileData)
				Size += (NvInt64)ObjectHere->GetSize();
		}

		if(!RetVal)
			break;
	}

	if(!RetVal)
		return false;

	if(WithSubContainers)
	{
		map<std::string,void *> *m_mSubcontainersHere = &m_mSubcontainers;
		if(Model != NULL)
			m_mSubcontainersHere = &(Model->m_mSubcontainers);
		map<string,void *>::iterator pos;

		NvContainer *SubContainer;
		NvContainer *ModelSubContainer;
		for(pos = m_mSubcontainersHere->begin(); pos != m_mSubcontainersHere->end(); ++pos)
		{
			ModelSubContainer = NULL;
			if(Model == NULL)
				SubContainer = (NvContainer*)(pos->second);
			else
			{
				ModelSubContainer = (NvContainer*)(pos->second);
				SubContainer = GetSubContainer(ModelSubContainer->GetName().c_str());
			}
			if(SubContainer != NULL)
				RetVal = SubContainer->GetXmlObjectDataSize("NV_ALLCONTAINER_OBJECTS", Size, true, WithFileData, ModelSubContainer);
			if(!RetVal)
				break;
		}
	}


	return RetVal;
}



void NvContainer::SetWsdlMessage(string& Output, int NumTabs, bool Input, void *XmlPtr)
{
	NvXml *XmlPtrHere = (NvXml*)XmlPtr;
	string Tabs = "";
	int i;
	for(i=0; i<NumTabs; i++)
		Tabs += '\t';
	
	NvObj *Object;
	map<string,void *>::iterator pos;
	for(pos = m_mObjects.begin(); pos != m_mObjects.end(); ++pos)
	{
		Object = (NvObj*)(pos->second);
		
		if(Object == NULL)
			continue;	// The object doesn't exists

		string ObjTypeName = "string";
		if(Object->GetType() == NV_OBJTYPE_BOOLEAN)
			ObjTypeName = "boolean";
		if(Object->GetType() == NV_OBJTYPE_INTEGER)
			ObjTypeName = "integer";
		if(Object->GetType() == NV_OBJTYPE_STRINGLIST)
			ObjTypeName = "stringlist";
		if(Object->GetType() == NV_OBJTYPE_INDSTRINGLIST)
			ObjTypeName = "indstringlist";
		if(Object->GetType() == NV_OBJTYPE_TABLE)
		{
			if(Input)
				ObjTypeName = "in_table";
			else
				ObjTypeName = "out_table";
		}
		if(Object->GetType() == NV_OBJTYPE_FILE)
		{
			if(Input)
				ObjTypeName = "in_file";
			else
				ObjTypeName = "out_file";
		}
		if(Object->GetType() == NV_OBJTYPE_BINARY)
		{
			if(Input)
				ObjTypeName = "in_binary";
			else
				ObjTypeName = "out_binary";
		}
		
		string XmlData = Tabs;
		XmlData += "<xs:element name=\"";
		XmlData += XmlPtrHere->EncodeOutputString(Object->GetName(), true);
		XmlData += "\" type=\"nvt:";
		XmlData += ObjTypeName;
		XmlData += "\" />\n";
		Output += XmlData;
		ReserveString(Output);
	}

	NvContainer *SubContainer;
	for(pos = m_mSubcontainers.begin(); pos != m_mSubcontainers.end(); ++pos)
	{
		SubContainer = (NvContainer*)(pos->second);

		bool bSubContainerEmpty = SubContainer->IsEmpty();

		string XmlData = Tabs;
		XmlData += "<xs:element name=\"";
		XmlData += XmlPtrHere->EncodeOutputString(SubContainer->GetName(), true);
		XmlData += "\">\n";
		XmlData += Tabs;
		XmlData += "\t<xs:complexType>\n";
		XmlData += Tabs;
		XmlData += "\t\t<xs:all>\n";
		Output += XmlData;
		ReserveString(Output);
		SubContainer->SetWsdlMessage(Output, NumTabs+3, Input, XmlPtr);
		XmlData = Tabs;
		XmlData += "\t\t</xs:all>\n";
//		if(bSubContainerEmpty)
//		{
			XmlData += Tabs;
			XmlData += "\t\t<xs:attribute name=\"type\" type=\"xs:string\" fixed=\"container\"></xs:attribute>\n";
//		}
		XmlData += Tabs;
		XmlData += "\t</xs:complexType>\n";
		XmlData += Tabs;
		XmlData += "</xs:element>\n";
		Output += XmlData;
		ReserveString(Output);
	}
}



bool NvContainer::SetSoapSample(std::string& Output, int NumTabs, std::string NsPrefix, void *XmlPtr, bool bUnicode)
{
	NvXml *XmlPtrHere = (NvXml*)XmlPtr;
	string Tabs = "";
	int i;
	for(i=0; i<NumTabs; i++)
		Tabs += '\t';
	

	bool AllObjects = true;

	NvObjStringList ObjectsToGet;
	NvObj *Object;
	if(AllObjects)
	{
		map<string,void *>::iterator pos;
		for(pos = m_mObjects.begin(); pos != m_mObjects.end(); ++pos)
		{
			Object = (NvObj*)(pos->second);
			ObjectsToGet.Add(Object->GetName());
		}
	}

	bool RetVal = true;
	for(i=0; i<ObjectsToGet.GetSize(); i++)
	{
		Object = GetObject(ObjectsToGet.GetValue(i).c_str());
		
		if(Object == NULL)
			return false;	// The object doesn't exists

		if(Object->GetType() == NV_OBJTYPE_BOOLEAN)
		{
			NvObjBoolean *ObjectHere = (NvObjBoolean *)Object;
			RetVal = ObjectHere->SetSoapSample(Output, NumTabs+1, NsPrefix, XmlPtr, bUnicode);
		}
		if(Object->GetType() == NV_OBJTYPE_INTEGER)
		{
			NvObjInteger *ObjectHere = (NvObjInteger *)Object;
			RetVal = ObjectHere->SetSoapSample(Output, NumTabs+1, NsPrefix, XmlPtr, bUnicode);
		}
		if(Object->GetType() == NV_OBJTYPE_STRING)
		{
			NvObjString *ObjectHere = (NvObjString *)Object;
			RetVal = ObjectHere->SetSoapSample(Output, NumTabs+1, NsPrefix, XmlPtr, bUnicode);
		}
		if(Object->GetType() == NV_OBJTYPE_STRINGLIST)
		{
			NvObjStringList *ObjectHere = (NvObjStringList *)Object;
			RetVal = ObjectHere->SetSoapSample(Output, NumTabs+1, NsPrefix, XmlPtr, bUnicode);
		}
		if(Object->GetType() == NV_OBJTYPE_INDSTRINGLIST)
		{
			NvObjIndStringList *ObjectHere = (NvObjIndStringList *)Object;
			RetVal = ObjectHere->SetSoapSample(Output, NumTabs+1, NsPrefix, XmlPtr, bUnicode);
		}
		if(Object->GetType() == NV_OBJTYPE_TABLE)
		{
			NvObjTable *ObjectHere = (NvObjTable *)Object;
			RetVal = ObjectHere->SetSoapSample(Output, NumTabs+1, NsPrefix, XmlPtr, bUnicode);
		}
		if(Object->GetType() == NV_OBJTYPE_FILE)
		{
			NvObjFile *ObjectHere = (NvObjFile *)Object;
			RetVal = ObjectHere->SetSoapSample(Output, NumTabs+1, NsPrefix, XmlPtr, bUnicode);
		}
		if(Object->GetType() == NV_OBJTYPE_BINARY)
		{
			NvObjBinary *ObjectHere = (NvObjBinary *)Object;
			RetVal = ObjectHere->SetSoapSample(Output, NumTabs+1, NsPrefix, XmlPtr, bUnicode);
		}

		if(!RetVal)
			break;
	}

	if(!RetVal)
		return false;

	NvContainer *SubContainer;
	map<string,void *>::iterator pos;
	for(pos = m_mSubcontainers.begin(); pos != m_mSubcontainers.end(); ++pos)
	{
		SubContainer = (NvContainer*)(pos->second);
		string XmlData = Tabs;
		XmlData += "<";
		XmlData += NsPrefix;
		XmlData += XmlPtrHere->EncodeOutputString(SubContainer->GetName(), bUnicode);
		XmlData += " type=\"container";
		XmlData += "\">\n";
		Output += XmlData;
		RetVal = SubContainer->SetSoapSample(Output, NumTabs+1, NsPrefix, XmlPtr, bUnicode);
		if(!RetVal)
			break;
		XmlData = Tabs;
		XmlData += "</";
		XmlData += NsPrefix;
		XmlData += XmlPtrHere->EncodeOutputString(SubContainer->GetName(), bUnicode);
		XmlData += ">\n";
		Output += XmlData;
	}

	return RetVal;
}
#endif // NV_SERVER


bool NvContainer::Import(const char *FileName, int AppendMode, const char *FileDir, int PersistTime, const char *Object, NvInt64 *Size, NvInt64 Offset, NvInt64 BytesToCopy)
{
	if(FileName == NULL)
		return false;

	FILE *stream = Nvfopen(FileName, "r");
	if(stream == NULL)
		return false;

	char Buffer[1024];
	if(freadline(Buffer, sizeof(Buffer), stream) == 0)
	{
		// This is not a container export file
		fclose(stream);
		return false;
	}

	Buffer[sizeof(Buffer)-1] = '\0';
	string Line = Buffer;
	if(Line != NV_CONTAINER_EXPORT_IDENTIFIER)
	{
		// This is not a container export file
		fclose(stream);
		return false;
	}

	bool RetVal = Import(stream, AppendMode, FileDir, PersistTime, Object, Size, Offset, BytesToCopy);

	fclose(stream);

	return RetVal;
}


bool NvContainer::Import(FILE *stream, int AppendMode, const char *FileDir, int PersistTime, const char *Object, NvInt64 *Size, NvInt64 Offset, NvInt64 BytesToCopy)
{

	string FileDirHere = "";
	if(FileDir != NULL)
		FileDirHere = FileDir;
	CleanString(FileDirHere);

	if(Size != NULL)
		*Size = 0;

	if(PersistTime < 0)
		PersistTime = -1;

	if(AppendMode == NV_CONTAINER_COPY_NO_APPEND)
	{
		// If not append mode, we must clear all container objects
		// With eventual files for registry containers
		ClearObjects(true);

		// and subcontainers
		if(!ClearSubcontainers(true))
			return false;
	}

	string ObjectHere = "";
	if(Object != NULL)
		ObjectHere = Object;
	NvContainerList::CleanContainerName(ObjectHere);

	bool ToSkip = false;

	// If there is a specific object requested
	// We must verify if it's in this container or not
	if(ObjectHere.length())
	{
		string::size_type index = ObjectHere.find('.');
		if(index != string::npos)
		{
			ToSkip = true;	// by default we skip this container
			// We have found a subcontainer name
			string RootName = NvContainerList::GetRootName(ObjectHere);
			if(RootName == GetName())
			{
				// This is our current container
				// So we remove it from ObjectHere
				ObjectHere = NvContainerList::RemoveRootName(ObjectHere);
				index = ObjectHere.find('.');
				if(index == string::npos)
				{
					// We are in the requested subcontainer
					// so we do not skip
					// and ObjectHere contains the object name to get
					ToSkip = false;
				}
			}
		}
	}

	char Buffer[1024];
	string Line;
	string::size_type index;
	while(freadline(Buffer, sizeof(Buffer), stream))
	{
		Line = Buffer;
		if(Line == "[OBJECT]")
		{
			// Subcontainer entry
			// We get its name
			if(!freadline(Buffer, sizeof(Buffer), stream))
				return false;
			Line = Buffer;
			index = Line.find("Name=");
			if(index == string::npos)
				return false;
			string ObjectName = Line.substr(5);
			CleanString(ObjectName);
			MakeLower(ObjectName);
			if(!ObjectName.length())
				continue;
		
			bool TakeObject = true;
			if(ToSkip)
				TakeObject = false;

			if(ObjectHere.length() && (!ToSkip))
			{
				string ObjectHereHere = ObjectHere;
				MakeLower(ObjectHereHere);
				if(ObjectHereHere != ObjectName)
					TakeObject = false;
			}
		
			// Get type
			if(!freadline(Buffer, sizeof(Buffer), stream))
				return false;
			Line = Buffer;
			index = Line.find("Type=");
			if(index == string::npos)
				return false;
			string ObjectType = Line.substr(5);
			CleanString(ObjectType);
			MakeUpper(ObjectType);
			if(!ObjectType.length())
				continue;
			int iObjectType = NV_OBJTYPE_UNKNOWN;
			if(ObjectType == "BOOLEAN")
				iObjectType = NV_OBJTYPE_BOOLEAN;
			if(ObjectType == "INTEGER")
				iObjectType = NV_OBJTYPE_INTEGER;
			if(ObjectType == "STRING")
				iObjectType = NV_OBJTYPE_STRING;
			if(ObjectType == "STRINGLIST")
				iObjectType = NV_OBJTYPE_STRINGLIST;
			if(ObjectType == "INDSTRINGLIST")
				iObjectType = NV_OBJTYPE_INDSTRINGLIST;
			if(ObjectType == "TABLE")
				iObjectType = NV_OBJTYPE_TABLE;
			if(ObjectType == "FILE")
				iObjectType = NV_OBJTYPE_FILE;
			if(ObjectType == "BINARY")
				iObjectType = NV_OBJTYPE_BINARY;

			// Get length
			if(!freadline(Buffer, sizeof(Buffer), stream))
				return false;
			Line = Buffer;
			index = Line.find("Length=");
			if(index == string::npos)
				return false;
			string ObjectLength = Line.substr(7);
			CleanString(ObjectLength);
			MakeUpper(ObjectLength);
			if(!ObjectLength.length())
				continue;
			NvInt64 iObjectLength = NvAtoi64(ObjectLength.c_str());

			if(!TakeObject)
			{
				// We skip this object
				// We must seek of the object size
				// But special management before for file objects
				if(iObjectType == NV_OBJTYPE_FILE)
				{
					// Get object file name
					if(!freadline(Buffer, sizeof(Buffer), stream))
						return false;

					// Get object origin value
					if(!freadline(Buffer, sizeof(Buffer), stream))
						return false;
					Line = Buffer;
					index = Line.find("Origin=");
					if(index != string::npos)
					{
						// Read next line for persist time
						// This is because old NIRVA versions (before 2.0.109)
						// were not storing origin in export files
						if(!freadline(Buffer, sizeof(Buffer), stream))
							return false;
					}
				}

				// We must seek of the object size now
				if(Nvfseek(stream, iObjectLength, SEEK_CUR) != 0)
					return false;
				continue;	// We don't copy this object
			}
			

			string ChildRootDirectory = m_sRootDirectory;
			AddDirSuffix(ChildRootDirectory);

			NvObj* DestinationObject = GetObject(ObjectName.c_str());
			if(DestinationObject != NULL)
			{
				// The destination object ever exists
				if(AppendMode == NV_CONTAINER_COPY_APPEND_NO_REPLACE)
				{
					// We must seek of the object size
					if(Nvfseek(stream, iObjectLength, SEEK_CUR) != 0)
						return false;
					continue;	// We don't copy this object
				}
	
				// We must remove the object except for registry otherwise in case of crash we may loose registry entry
				if(!ChildRootDirectory.length())
					RemoveObject(ObjectName.c_str());
			}

			// now we can create the new object
			if(!CreateObject(ObjectName.c_str(), iObjectType))
				return false;

			DestinationObject = GetObject(ObjectName.c_str());
			if(DestinationObject == NULL)
				return false;

			if(iObjectType == NV_OBJTYPE_BOOLEAN)
			{
				NvObjBoolean *Dobj = (NvObjBoolean*)DestinationObject;
				if(!Dobj->Import(stream, (int)iObjectLength))
					return false;
				if(Size != NULL)
					*Size = 1;
				if(ChildRootDirectory.length())
				{
					// for registry containers, we must save the object to file
					string FileName = GetObjectFile(DestinationObject->GetName().c_str(), NV_OBJTYPE_BOOLEAN);
					if(!FileName.length())
						return false;
					if(!SaveObjToFile(FileName.c_str(), Dobj))
						return false;
				}
			}
			if(iObjectType == NV_OBJTYPE_INTEGER)
			{
				NvObjInteger *Dobj = (NvObjInteger*)DestinationObject;
				if(!Dobj->Import(stream, (int)iObjectLength))
					return false;
				if(Size != NULL)
					*Size = sizeof(int);
				if(ChildRootDirectory.length())
				{
					// for registry containers, we must save the object to file
					string FileName = GetObjectFile(DestinationObject->GetName().c_str(), NV_OBJTYPE_INTEGER);
					if(!FileName.length())
						return false;
					if(!SaveObjToFile(FileName.c_str(), Dobj))
						return false;
				}
			}
			if(iObjectType == NV_OBJTYPE_STRING)
			{
				NvObjString *Dobj = (NvObjString*)DestinationObject;
				if(!Dobj->Import(stream, (int)iObjectLength))
					return false;
				if(Size != NULL)
					*Size = (NvInt64)(Dobj->GetValue().length());
				if(ChildRootDirectory.length())
				{
					// for registry containers, we must save the object to file
					string FileName = GetObjectFile(DestinationObject->GetName().c_str(), NV_OBJTYPE_STRING);
					if(!FileName.length())
						return false;
					if(!SaveObjToFile(FileName.c_str(), Dobj))
						return false;
				}
			}
			if(iObjectType == NV_OBJTYPE_STRINGLIST)
			{
				NvObjStringList *Dobj = (NvObjStringList*)DestinationObject;
				if(!Dobj->Import(stream, (int)iObjectLength))
					return false;
				if(Size != NULL)
					*Size = (NvInt64)(Dobj->GetSize());
				if(ChildRootDirectory.length())
				{
					// for registry containers, we must save the object to file
					string FileName = GetObjectFile(DestinationObject->GetName().c_str(), NV_OBJTYPE_STRINGLIST);
					if(!FileName.length())
						return false;
					if(!SaveObjToFile(FileName.c_str(), Dobj))
						return false;
				}
			}
			if(iObjectType == NV_OBJTYPE_INDSTRINGLIST)
			{
				NvObjIndStringList *Dobj = (NvObjIndStringList*)DestinationObject;
				if(!Dobj->Import(stream, (int)iObjectLength))
					return false;
				if(Size != NULL)
					*Size = (NvInt64)(Dobj->GetSize());
				if(ChildRootDirectory.length())
				{
					// for registry containers, we must save the object to file
					string FileName = GetObjectFile(DestinationObject->GetName().c_str(), NV_OBJTYPE_INDSTRINGLIST);
					if(!FileName.length())
						return false;
					if(!SaveObjToFile(FileName.c_str(), Dobj))
						return false;
				}
			}
			if(iObjectType == NV_OBJTYPE_TABLE)
			{
				NvObjTable *Dobj = (NvObjTable*)DestinationObject;
				if(!Dobj->Import(stream, (int)iObjectLength))
					return false;
				if(Size != NULL)
					*Size = (NvInt64)(Dobj->GetNumRow());
				if(ChildRootDirectory.length())
				{
					// for registry containers, we must save the object to file
					string FileName = GetObjectFile(DestinationObject->GetName().c_str(), NV_OBJTYPE_TABLE);
					if(!FileName.length())
						return false;
					if(!SaveObjToFile(FileName.c_str(), Dobj))
						return false;
				}
			}
			if(iObjectType == NV_OBJTYPE_FILE)
			{
				NvObjFile *Dobj = (NvObjFile*)DestinationObject;

				// Get object file name
				if(!freadline(Buffer, sizeof(Buffer), stream))
					return false;
				Line = Buffer;
				index = Line.find("FileName=");
				if(index == string::npos)
					return false;
				string ObjectFileName = Line.substr(9);
				CleanString(ObjectFileName);

				// Get object origin value
				string Origin = "";
				if(!freadline(Buffer, sizeof(Buffer), stream))
					return false;
				Line = Buffer;
				index = Line.find("Origin=");
				if(index != string::npos)
				{
					Origin = Line.substr(7);

					// Read next line for persist time
					// This is because old NIRVA versions (before 2.0.109)
					// were not storing origin in export files
					if(!freadline(Buffer, sizeof(Buffer), stream))
						return false;
					Dobj->SetOrigin(Origin);
				}
				
				// Get object persist value
				Line = Buffer;
				index = Line.find("Persist=");
				if(index == string::npos)
					return false;
				string Persist = Line.substr(8);
				CleanString(Persist);
				int iPersist = atoi(Persist.c_str());
				Dobj->SetPersistTime(iPersist);
				if(Size != NULL)
					*Size = iObjectLength;

				if(ChildRootDirectory.length())
				{
					// The file is always persistant here because it's a registry file
					Dobj->SetPersistTime();
					// for registry containers, we must save the object to file
					string FileName = GetObjectFile(DestinationObject->GetName().c_str(), NV_OBJTYPE_FILE);
					
					// The object filename is directly the registry file
					Dobj->SetFileName(FileName);
					if(!FileName.length())
						return false;

					string FileNameTmp = FileName;
					FileNameTmp += ".nvtmp";
					unlink(FileNameTmp.c_str());
					Dobj->SetFileName(FileNameTmp.c_str());

					if(!Dobj->Import(stream, iObjectLength, Offset, BytesToCopy))
					{
						Dobj->SetFileName(FileName);
						return false;
					}
					Dobj->SetFileName(FileName.c_str());
					unlink(FileName.c_str());
					rename(FileNameTmp.c_str(), FileName.c_str());
				}
				else
				{
					if(FileDirHere.length())
					{
						if(Dobj->m_iPersistTime == -1)				// The file is persistant
							Dobj->SetPersistTime(PersistTime);		// we change that to the new persistant value

						// Get file extension if there is one
						string FileExtension = "";
						index = ObjectFileName.rfind('.');
						if(index != string::npos)
							FileExtension = ObjectFileName.substr(index+1);
						if(!Dobj->SetAutoFileName(FileDirHere, FileExtension))
							return false;
						if(!Dobj->Import(stream, iObjectLength, Offset, BytesToCopy))
							return false;
					}
					else
					{
						// We must seek of the object size
						if(Nvfseek(stream, iObjectLength, SEEK_CUR) != 0)
							return false;
					}
				}
			}
			if(iObjectType == NV_OBJTYPE_BINARY)
			{
				NvObjBinary *Dobj = (NvObjBinary*)DestinationObject;
				if(!Dobj->Import(stream, (int)iObjectLength, (int)Offset, (int)BytesToCopy))
					return false;
				if(Size != NULL)
					*Size = iObjectLength;
				if(ChildRootDirectory.length())
				{
					// for registry containers, we must save the object to file
					string FileName = GetObjectFile(DestinationObject->GetName().c_str(), NV_OBJTYPE_BINARY);
					if(!FileName.length())
						return false;
					if(!SaveObjToFile(FileName.c_str(), Dobj))
						return false;
				}
			}
		
			// If we are here an the request is a specific object
			// We have found it at this step so we can leave successfully
			if(ObjectHere.length() && (!ToSkip))
				return true;
		}	// End processing object
		
		if(Line == "[SUBCONTAINER]")
		{
			// Subcontainer entry
			// We get its name
			if(!freadline(Buffer, sizeof(Buffer), stream))
				return false;
			Line = Buffer;
			index = Line.find("Name=");
			if(index == string::npos)
				return false;

			string ContainerName = Line.substr(5);
			CleanString(ContainerName);
			MakeLower(ContainerName);
			if(!ContainerName.length())
				return false;

			// We have a container name now
	
			NvContainer *DestinationContainer = NULL;
				
			DestinationContainer = GetSubContainer(ContainerName.c_str());
			if(DestinationContainer == NULL)
			{
				if(!CreateSubContainer(ContainerName.c_str()))
					return false;

				DestinationContainer = GetSubContainer(ContainerName.c_str());
				if(DestinationContainer == NULL)
					return false;
			}

			// Recursive function that processes all subcontainers
			if(!DestinationContainer->Import(stream, AppendMode, FileDir, PersistTime, ObjectHere.c_str()))
				return false;
		}	// End processing subcontainer
	
		if(Line == "[END_SUBCONTAINER]")
			break;

	}
	
	return true;
}


bool NvContainer::Export(const char *FileName, const char *Objects, bool WithSubContainers)
{
	FILE *stream = Nvfopen(FileName, "w");
	if(stream == NULL)
		return false;

	string Line = NV_CONTAINER_EXPORT_IDENTIFIER;
	AddEndLine(Line);
	Line += "Version=";
	Line += NV_CONTAINER_EXPORT_VERSION;
	AddEndLine(Line);
	AddEndLine(Line);

	if(fwrite(Line.c_str(), 1, Line.length(), stream) != Line.length())
	{
		fclose(stream);
		return false;
	}

	bool RetVal = Export(stream, Objects, WithSubContainers);
	fclose(stream);

	return RetVal;
}



bool NvContainer::Export(FILE *stream, const char *Objects, bool WithSubContainers)
{

	string ObjectsHere = "";
	if(Objects != NULL)
		ObjectsHere = Objects;
	CleanString(ObjectsHere);

	// starts with objects
	NvObjStringList ObjectsToGet;

	// if Objects is not null, we must select only requested objects (maybe none)
	// Remark: if object is given but blank, no objects will be get
	// and if object is not given (NULL), all valid objects will be get
	int i;
	if(Objects != NULL)
	{
		while(ObjectsHere.length())
		{
			// Several objects can be in the list
			// They are separated by a semicolumn
			string ObjectName = ObjectsHere;
			string::size_type index = ObjectName.find(';');
			if(index != string::npos)
			{
				ObjectName = ObjectsHere.substr(0, index);
				ObjectsHere = ObjectsHere.substr(index+1);
			}
			else
				ObjectsHere = "";

			// An object name has been found
			CleanString(ObjectName);
			
			// It should not be empty
			if(!ObjectName.length())
				continue;

			// The object should exist in the source
			if(!GetObject(ObjectName.c_str()))
				continue;

			ObjectsToGet.Add(ObjectName);
		}
	}
	else
	{
		// If source is registry, we must get all real objects
		// and not only these ones in the Source->m_mObjects
		vector<string> RealObjNames;

		if(m_sRootDirectory.length())
		{
			// This is a registry container
			GetRealObjectNames(&RealObjNames);
		}
		else
		{
			map<string,void *>::iterator pos;
			for(pos = m_mObjects.begin(); pos != m_mObjects.end(); ++pos)
			{
				NvObj *Object = (NvObj*)(pos->second);
				if(Object == NULL)
					continue;
				RealObjNames.push_back(Object->GetName());
			}
		}
		for(i=0; i<(int)RealObjNames.size(); i++)
			ObjectsToGet.Add(RealObjNames[i]);
	}


	NvObj *Object;
	bool RetVal = true;
	for(i=0; i<ObjectsToGet.GetSize(); i++)
	{
		string sObjectInList = ObjectsToGet.GetValue(i);

		Object = GetObject(sObjectInList.c_str());
		
		if(Object == NULL)
			continue;	// Should not occur

		if(Object->GetType() == NV_OBJTYPE_BOOLEAN)
		{
			NvObjBoolean *ObjectHere = (NvObjBoolean *)Object;
			RetVal = ObjectHere->Export(stream);
		}
		if(Object->GetType() == NV_OBJTYPE_INTEGER)
		{
			NvObjInteger *ObjectHere = (NvObjInteger *)Object;
			RetVal = ObjectHere->Export(stream);
		}
		if(Object->GetType() == NV_OBJTYPE_STRING)
		{
			NvObjString *ObjectHere = (NvObjString *)Object;
			RetVal = ObjectHere->Export(stream);
		}
		if(Object->GetType() == NV_OBJTYPE_STRINGLIST)
		{
			NvObjStringList *ObjectHere = (NvObjStringList *)Object;
			RetVal = ObjectHere->Export(stream);
		}
		if(Object->GetType() == NV_OBJTYPE_INDSTRINGLIST)
		{
			NvObjIndStringList *ObjectHere = (NvObjIndStringList *)Object;
			RetVal = ObjectHere->Export(stream);
		}
		if(Object->GetType() == NV_OBJTYPE_TABLE)
		{
			NvObjTable *ObjectHere = (NvObjTable *)Object;
			RetVal = ObjectHere->Export(stream);
		}
		if(Object->GetType() == NV_OBJTYPE_FILE)
		{
			NvObjFile *ObjectHere = (NvObjFile *)Object;
			RetVal = ObjectHere->Export(stream);
		}
		if(Object->GetType() == NV_OBJTYPE_BINARY)
		{
			NvObjBinary *ObjectHere = (NvObjBinary *)Object;
			RetVal = ObjectHere->Export(stream);
		}

		if(!RetVal)
			break;
	}

	if(!RetVal)
		return false;


	if(WithSubContainers)
	{
		NvContainer *SourceContainer = NULL;

		vector<string> SubNames;

		// For registry source, we must take all real subcontainers
		if(m_sRootDirectory.length())
		{
			// This is a registry container
			GetRealSubContainerNames(&SubNames);
		}
		else
		{
			map<string,void *>::iterator pos;
			for(pos = m_mSubcontainers.begin(); pos != m_mSubcontainers.end(); ++pos)
			{
				SourceContainer = (NvContainer *)(pos->second);
				if(SourceContainer == NULL)
					continue;
				SubNames.push_back(SourceContainer->GetName());
			}
		}
		for(i=0; i<(int)SubNames.size(); i++)
		{
			SourceContainer = GetSubContainer(SubNames[i].c_str());
			if(SourceContainer == NULL)
				continue;

			string Line = "[SUBCONTAINER]";
			AddEndLine(Line);
			Line += "Name=";
			Line += SourceContainer->GetName();
			AddEndLine(Line);
			AddEndLine(Line);
			if(fwrite(Line.c_str(), 1, Line.length(), stream) != Line.length())
				return false;
			

			RetVal = SourceContainer->Export(stream, NULL, WithSubContainers);
			if(!RetVal)
				break;
	
			Line = "[END_SUBCONTAINER]";
			AddEndLine(Line);
			AddEndLine(Line);
			if(fwrite(Line.c_str(), 1, Line.length(), stream) != Line.length())
				return false;
		}
	}

	return RetVal;
}



// Populates a container from the content of a directory
// Subdirectories becomes subcontainers and files are file objects
// The container is cleared
// Each Subcontainer has a string object named DIR_NAME that contains the real directory name (with respect to original case)
bool NvContainer::Dir(const char *DirName, const char *Files, const char *SubDirs, bool Recurse, int PersistTime, bool NameStandard)
{
	string DirNameHere = "";
	if(DirName != NULL)
		DirNameHere = DirName;
	CleanString(DirNameHere);
	if(!DirNameHere.length())
		return false;

	if(PersistTime < 0)
		PersistTime = -1;

	// We always clear the container
	Clear();

	// Used for standard object and subcontainers naming
	int iDirCounter = 1;
	int iFileCounter = 1;

	bool WithFiles = true;
	bool WithSubDir = true;

	char FilesToAdd[NV_MAX_PATHLENGTH];
	char DirsToAdd[NV_MAX_PATHLENGTH];
	strcpy(FilesToAdd, "");
	strcpy(DirsToAdd, "");

	if(SubDirs != NULL)
	{
		if(strlen(SubDirs))
			strcpy(DirsToAdd, SubDirs);
		else
			WithSubDir = false;
	}
	if(Files != NULL)
	{
		if(strlen(Files))
			strcpy(FilesToAdd, Files);
		else
			WithFiles = false;
	}
	if(strcmp(DirsToAdd, "*") == 0)
		strcpy(DirsToAdd, "");
	if(strcmp(FilesToAdd, "*") == 0)
		strcpy(FilesToAdd, "");

	if(!WithSubDir)
		Recurse = false;



	char FileToAdd[NV_MAX_PATHLENGTH];
#ifdef WIN32
	// We use functions for finding files given a path and wildcards
	char FilesToFind[NV_MAX_PATHLENGTH];
	strncpy(FilesToFind, DirName, NV_MAX_PATHLENGTH-1);
	if(strlen(FilesToFind))
	{
		// We add the last \ character
		if(FilesToFind[strlen(FilesToFind) - 1] != '\\')
			strcat(FilesToFind, "\\");
	}
	strcat(FilesToFind, "*");
	WIN32_FIND_DATA FindData;
	// Try to find the first file
	HANDLE hFileFind = FindFirstFile(FilesToFind, &FindData);
	if(hFileFind == INVALID_HANDLE_VALUE)
		return false;	// No file
	while(1)
	{
		// We remove all the found files
		strncpy(FileToAdd, DirName, NV_MAX_PATHLENGTH-1);
		if(strlen(FileToAdd))
		{
			// We add the last \ character
			if(FileToAdd[strlen(FileToAdd) - 1] != '\\')
				strcat(FileToAdd, "\\");
		}
		strcat(FileToAdd, FindData.cFileName);
		if((FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0)
		{
			if(WithFiles)
			{
				bool ToAdd = false;
				if(strlen(FilesToAdd))
				{
					// Case insensitive for windows
					if(strstrnocase(FindData.cFileName, FilesToAdd) == FindData.cFileName)
						ToAdd = true;
				}
				else
					ToAdd = true;
				if(ToAdd)
				{
					string ObjectName = FindData.cFileName;
					if(NameStandard)
					{
						char Buffer[20];
						sprintf(Buffer, "file%d", iFileCounter);
						ObjectName = Buffer;
						iFileCounter++;
					}
					else
					{
						CleanString(ObjectName);
						MakeLower(ObjectName);
						// replace point with underscore
						string::size_type index;
						while((index = ObjectName.find('.')) != string::npos)
							ObjectName.replace(index, 1, "_p_");
					}


					NvObjFile* DestinationObject = (NvObjFile*)GetObject(ObjectName.c_str());
					if(DestinationObject != NULL)
					{
						// We must remove the object
						RemoveObject(ObjectName.c_str());
					}

					// now we can create the new object
					if(!CreateObject(ObjectName.c_str(), NV_OBJTYPE_FILE))
					{
						FindClose(hFileFind);	// we close the finder
						return false;
					}

					DestinationObject = (NvObjFile*)GetObject(ObjectName.c_str());
					if(DestinationObject == NULL)
					{
						FindClose(hFileFind);	// we close the finder
						return false;
					}

					DestinationObject->SetPersistTime(PersistTime);
					DestinationObject->SetFileName(FileToAdd);
					DestinationObject->SetOrigin(FindData.cFileName);
				}
			}
		}
		else if(FindData.cFileName[0] != '.')
		{
			if(WithSubDir)
			{
				bool ToAdd = false;
				if(strlen(DirsToAdd))
				{
					// Case insensitive for windows
					if(strstrnocase(FindData.cFileName, DirsToAdd) == FindData.cFileName)
						ToAdd = true;
				}
				else
					ToAdd = true;
				
				if(ToAdd)
				{
					// Create the new subcontainer here

					NvContainer *DestinationContainer = NULL;
					
					string ContainerName = FindData.cFileName;
					if(NameStandard)
					{
						char Buffer[20];
						sprintf(Buffer, "dir%d", iDirCounter);
						ContainerName = Buffer;
						iDirCounter++;
					}
					else
					{
						CleanString(ContainerName);
						MakeLower(ContainerName);
						// replace point with underscore
						string::size_type index;
						while((index = ContainerName.find('.')) != string::npos)
							ContainerName.replace(index, 1, "_p_");
					}
						
					DestinationContainer = GetSubContainer(ContainerName.c_str());
					if(DestinationContainer == NULL)
					{
						if(!CreateSubContainer(ContainerName.c_str()))
						{
							FindClose(hFileFind);	// we close the finder
							return false;
						}

						DestinationContainer = GetSubContainer(ContainerName.c_str());
						if(DestinationContainer == NULL)
						{
							FindClose(hFileFind);	// we close the finder
							return false;
						}
					}
					

					if(Recurse)
					{
						if(!DestinationContainer->Dir(FileToAdd, Files, SubDirs, Recurse, PersistTime, NameStandard))
						{
							FindClose(hFileFind);	// we close the finder
							return false;
						}
					}

					// Write now the DIR_NAME object that contains the real directory name
					NvObjString* DestinationObject = (NvObjString*)(DestinationContainer->GetObject("DIR_NAME"));
					if(DestinationObject != NULL)
					{
						// We must remove the object
						RemoveObject("DIR_NAME");
					}

					// now we can create the new object
					if(!DestinationContainer->CreateObject("DIR_NAME", NV_OBJTYPE_STRING))
					{
						FindClose(hFileFind);	// we close the finder
						return false;
					}

					DestinationObject = (NvObjString*)(DestinationContainer->GetObject("DIR_NAME"));
					if(DestinationObject == NULL)
					{
						FindClose(hFileFind);	// we close the finder
						return false;
					}

					DestinationObject->SetValue(FindData.cFileName);
				}
			}
		}
		if(!FindNextFile(hFileFind, &FindData))
			break;	// no more file
	}
	FindClose(hFileFind);	// we close the finder
#else

	// for unix, we open the directory file and we search for entries
	// corresponding to our temporary file names
	char WorkDirHere[NV_MAX_PATHLENGTH];
	strncpy(WorkDirHere, DirName, NV_MAX_PATHLENGTH-1);
	if(strlen(WorkDirHere))
	{
		// We remove the last / character
		if(WorkDirHere[strlen(WorkDirHere) - 1] == '/')
			WorkDirHere[strlen(WorkDirHere) - 1] = '\0';
	}
	else
		strcpy(WorkDirHere, ".");	// we are in the current directory

	// open directory file
	DIR *Dir = opendir(WorkDirHere);
	if(Dir == NULL)
		return false;		// cannot open directory file
	
	dirent *DirEntry;
	while((DirEntry = readdir(Dir)) != 0)
	{
		// we remove files
		strncpy(FileToAdd, WorkDirHere, NV_MAX_PATHLENGTH-1);
		if(strlen(FileToAdd))
		{
			// We add the last \ character
			if(FileToAdd[strlen(FileToAdd) - 1] != '/')
				strcat(FileToAdd, "/");
		}
		strcat(FileToAdd, DirEntry->d_name);
		if(DirExist(FileToAdd))
		{
			// Its a directory
			if(WithSubDir && (DirEntry->d_name[0] != '.'))
			{
				bool ToAdd = false;
				if(strlen(DirsToAdd))
				{
					if(strstr(DirEntry->d_name, DirsToAdd) == DirEntry->d_name)
						ToAdd = true;
				}
				else
					ToAdd = true;
				if(ToAdd)
				{
					// Create the new subcontainer here

					NvContainer *DestinationContainer = NULL;
					
					string ContainerName = DirEntry->d_name;
					if(NameStandard)
					{
						char Buffer[20];
						sprintf(Buffer, "dir%d", iDirCounter);
						ContainerName = Buffer;
						iDirCounter++;
					}
					else
					{
						CleanString(ContainerName);
						MakeLower(ContainerName);
						// replace point with underscore
						string::size_type index;
						while((index = ContainerName.find('.')) != string::npos)
							ContainerName.replace(index, 1, "_p_");
					}
						
					DestinationContainer = GetSubContainer(ContainerName.c_str());
					if(DestinationContainer == NULL)
					{
						if(!CreateSubContainer(ContainerName.c_str()))
						{
							closedir(Dir);	// close the directory file
							return false;
						}

						DestinationContainer = GetSubContainer(ContainerName.c_str());
						if(DestinationContainer == NULL)
						{
							closedir(Dir);	// close the directory file
							return false;
						}
					}


					if(Recurse)
					{
						if(!DestinationContainer->Dir(FileToAdd, Files, SubDirs, Recurse, PersistTime, NameStandard))
						{
							closedir(Dir);	// close the directory file
							return false;
						}
					}

					// Write now the DIR_NAME object that contains the real directory name
					NvObjString* DestinationObject = (NvObjString*)(DestinationContainer->GetObject("DIR_NAME"));
					if(DestinationObject != NULL)
					{
						// We must remove the object
						RemoveObject("DIR_NAME");
					}

					// now we can create the new object
					if(!DestinationContainer->CreateObject("DIR_NAME", NV_OBJTYPE_STRING))
					{
						closedir(Dir);	// close the directory file
						return false;
					}

					DestinationObject = (NvObjString*)(DestinationContainer->GetObject("DIR_NAME"));
					if(DestinationObject == NULL)
					{
						closedir(Dir);	// close the directory file
						return false;
					}

					DestinationObject->SetValue(DirEntry->d_name);
				}
			}
		}
		else
		{
			//it's a file
			if(WithFiles)
			{
				bool ToAdd = false;
				if(strlen(FilesToAdd))
				{
					if(strstr(DirEntry->d_name, FilesToAdd) == DirEntry->d_name)
						ToAdd = true;
				}
				else
					ToAdd = true;
				if(ToAdd)
				{
					string ObjectName = DirEntry->d_name;
					if(NameStandard)
					{
						char Buffer[20];
						sprintf(Buffer, "file%d", iFileCounter);
						ObjectName = Buffer;
						iFileCounter++;
					}
					else
					{
						CleanString(ObjectName);
						MakeLower(ObjectName);
						// replace point with underscore
						string::size_type index;
						while((index = ObjectName.find('.')) != string::npos)
							ObjectName.replace(index, 1, "_p_");
					}

					NvObjFile* DestinationObject = (NvObjFile*)GetObject(ObjectName.c_str());
					if(DestinationObject != NULL)
					{
						// We must remove the object
						RemoveObject(ObjectName.c_str());
					}

					// now we can create the new object
					if(!CreateObject(ObjectName.c_str(), NV_OBJTYPE_FILE))
					{
						closedir(Dir);	// close the directory file
						return false;
					}

					DestinationObject = (NvObjFile*)GetObject(ObjectName.c_str());
					if(DestinationObject == NULL)
					{
						closedir(Dir);	// close the directory file
						return false;
					}

					DestinationObject->SetPersistTime(PersistTime);
					DestinationObject->SetFileName(FileToAdd);
					DestinationObject->SetOrigin(DirEntry->d_name);
				}
			}
		}
	}
	closedir(Dir);	// close the directory file
#endif
	
	return true;
}





bool NvContainer::View(NvObjTable *Content, bool WithRootName)
{
	Content->Clear();	
	Content->InsertColumn("NAME", "Name");
	Content->InsertColumn("TYPE", "Type");
	Content->InsertColumn("INFO", "Info");

	// Get the complete container name
	string CompleteParentName = "";
	NvContainer *Parent = GetParent();;
	vector<string> Parents;
	while(Parent != NULL)
	{
		Parents.push_back(Parent->GetName());
		Parent = Parent->GetParent();
	}
	int i;
	bool bHasParent = false;
	if(Parents.size())
		bHasParent = true;
	if(bHasParent)
	{
		// The container is not the root one
		if(WithRootName)
			CompleteParentName=Parents[Parents.size()-1];
	}
	for(i=(int)Parents.size()-2; i>=0; i--)
	{
		if(CompleteParentName.length())
			CompleteParentName += ".";
		CompleteParentName += Parents[i];
	}

	string CompleteName = CompleteParentName;
	if(CompleteName.length())
		CompleteName += ".";
	CompleteName += GetName();
	CompleteName += ".";


	// First get the parent
	if(bHasParent)
	{
		Content->InsertRow();
		Content->InsertCellLine(-1, 0, GetParent()->GetName());
		Content->InsertCellLine(-1, 1, "PARENT");
		Content->InsertCellLine(-1, 2, CompleteParentName);
	}

	// process subcontainers
	// We use a map in order to automatically sort the subcontainers
	map<string,NvContainer*> SubContainers;
	NvContainer *SourceContainer = NULL;
	vector<string> SubNames;
	// For registry source, we must take all real subcontainers
	if(m_sRootDirectory.length())
	{
		// This is a registry container
		GetRealSubContainerNames(&SubNames);
	}
	else
	{
		map<string,void *>::iterator pos;
		for(pos = m_mSubcontainers.begin(); pos != m_mSubcontainers.end(); ++pos)
		{
			SourceContainer = (NvContainer *)(pos->second);
			if(SourceContainer == NULL)
				continue;
			SubNames.push_back(SourceContainer->GetName());
		}
	}
	for(i=0; i<(int)SubNames.size(); i++)
	{
		SourceContainer = GetSubContainer(SubNames[i].c_str());
		if(SourceContainer == NULL)
			continue;
		SubContainers[SourceContainer->GetName()] = SourceContainer;
	}
	SubNames.clear();
	
	map<string,NvContainer*>::iterator pos;
	for(pos = SubContainers.begin(); pos != SubContainers.end(); ++pos)
	{
		SourceContainer = pos->second;
		Content->InsertRow();
		Content->InsertCellLine(-1, 0, pos->first);
		Content->InsertCellLine(-1, 1, "CHILD");
		string CompleteChildName = CompleteName;
		CompleteChildName += pos->first;
		Content->InsertCellLine(-1, 2, CompleteChildName);
	}


	// process objects
	// We use a map in order to automatically sort the objects
	map<string,NvObj*> Objects;
	NvObj *SourceObject = NULL;
	vector<string> ObjectNames;
	// For registry source, we must take all real subcontainers
	if(m_sRootDirectory.length())
	{
		// This is a registry container
		GetRealObjectNames(&ObjectNames);
	}
	else
	{
		map<string,void *>::iterator pos;
		for(pos = m_mObjects.begin(); pos != m_mObjects.end(); ++pos)
		{
			NvObj *Object = (NvObj*)(pos->second);
			if(Object == NULL)
				continue;
			ObjectNames.push_back(Object->GetName());
		}
	}
	for(i=0; i<(int)ObjectNames.size(); i++)
	{
		SourceObject = GetObject(ObjectNames[i].c_str());
		if(SourceObject == NULL)
			continue;
		Objects[SourceObject->GetName()] = SourceObject;
	}
	ObjectNames.clear();
	
	map<string,NvObj*>::iterator pos2;
	for(pos2 = Objects.begin(); pos2 != Objects.end(); ++pos2)
	{
		SourceObject = pos2->second;
		Content->InsertRow();
		Content->InsertCellLine(-1, 0, pos2->first);
		Content->InsertCellLine(-1, 1, "OBJECT");
		string ObjectType = "STRING";
		int iObjType = SourceObject->GetType();
		switch(iObjType)
		{
			case NV_OBJTYPE_BOOLEAN:
				ObjectType = "BOOLEAN";
				break;
			case NV_OBJTYPE_INTEGER:
				ObjectType = "INTEGER";
				break;
			case NV_OBJTYPE_STRING:
				ObjectType = "STRING";
				break;
			case NV_OBJTYPE_STRINGLIST:
				ObjectType = "STRINGLIST";
				break;
			case NV_OBJTYPE_INDSTRINGLIST:
				ObjectType = "INDSTRINGLIST";
				break;
			case NV_OBJTYPE_TABLE:
				ObjectType = "TABLE";
				break;
			case NV_OBJTYPE_FILE:
				ObjectType = "FILE";
				break;
			case NV_OBJTYPE_BINARY:
				ObjectType = "BINARY";
				break;
		}
		Content->InsertCellLine(-1, 2, ObjectType);
	}

	return true;
}


bool NvContainer::MoveObject(NvContainer *DestinationContainer, const char *ObjName, bool Replace)
{
	if(DestinationContainer == NULL)
		return false;
	if(DestinationContainer == this)
		return true; // Cannot move on ourself
	
	NvObj* Object = GetObject(ObjName);
	if(Object == NULL)
		return false; // The object doesn't exist

	NvObj* DObject = DestinationContainer->GetObject(ObjName);
	if(DObject != NULL)
	{
		// The destination object exist
		if(Replace)
		{
			// Remove it in replace mode
			DestinationContainer->RemoveObject(ObjName);
		}
		else
		{
			// Return an error if no replace mode
			return false; // The object doesn't exist
		}
	}

	// We transfert the object reference to the destination container
	DestinationContainer->m_mObjects[Object->GetName()] = (void *)(Object);
	if(DestinationContainer->bUseLastObject)
		DestinationContainer->LastObject = Object;
	
	// We erase the object reference from the actual container
	m_mObjects.erase(Object->GetName());
	if((bUseLastObject) && (Object == LastObject))
		LastObject = NULL;

	return true;
}



bool NvContainer::MoveContainer(NvContainer *SourceContainer)
{
	if(SourceContainer == NULL)
		return false;
	if(SourceContainer == this)
		return true; // Cannot move on ourself

	// Move not autorised on registry containers
	if(SourceContainer->m_sRootDirectory.length() || m_sRootDirectory.length())
		return false;

	// Verify that we do not break the branch we sit on (the dest container is a subcontainer of the source container)
	bool bRecurseFound = false;
	NvContainer *ContainerHere = this;
	while(ContainerHere != NULL)
	{
		ContainerHere = ContainerHere->GetParent();
		if(ContainerHere == SourceContainer)
		{
			bRecurseFound = true;
			break;
		}
	}
	if(bRecurseFound)
		return false;

	// Now check if the source container is a subcontainer of the destination container
	// This is autorised but we must proceed differently
	ContainerHere = SourceContainer;
	while(ContainerHere != NULL)
	{
		ContainerHere = ContainerHere->GetParent();
		if(ContainerHere == this)
		{
			bRecurseFound = true;
			break;
		}
	}

	if(bRecurseFound)
	{
		// We must remove the source container from the list of subcontainers
		SourceContainer->GetParent()->m_mSubcontainers.erase(SourceContainer->GetName());
	}


	// Clear this container
	Clear();

	// We transfert the object table to the destination container
	m_mObjects = SourceContainer->m_mObjects;
	// We transfert the subcontainer table to the destination container
	m_mSubcontainers = SourceContainer->m_mSubcontainers;
	map<string,void *>::iterator pos;
	for(pos = m_mSubcontainers.begin(); pos != m_mSubcontainers.end(); ++pos)
	{
		NvContainer *Container = (NvContainer *)(pos->second);
		if(Container == NULL)
			continue;
		Container->SetParent(this);
	}
	SourceContainer->m_mObjects.clear();
	SourceContainer->m_mSubcontainers.clear();
	SourceContainer->bUseLastObject = false;
	SourceContainer->bUseLastSubContainer = false;
	SourceContainer->LastObject = NULL;
	SourceContainer->LastSubContainer = NULL;

	return true;
}


/////////////////////////////////////////////////////////////////////////////
// NvLocalContainer

NvLocalContainer::NvLocalContainer()
{
	m_sName = NV_CONTAINER_DEFAULT_NAME;
	m_sRootDirectory = "";
	LastObject = NULL;
	bUseLastObject = true;
}

NvLocalContainer::NvLocalContainer(string Name)
{
	SetName(Name);
	m_sRootDirectory = "";
	LastObject = NULL;
	bUseLastObject = true;
}

NvLocalContainer::NvLocalContainer(const char *Name)
{
	SetName(Name);
	m_sRootDirectory = "";
	LastObject = NULL;
	bUseLastObject = true;
}

NvLocalContainer::~NvLocalContainer()
{
	Clear();
}


string NvLocalContainer::GetName()
{
	return m_sName;
}

bool NvLocalContainer::SetName(string Name)
{
	CleanString(Name, true);
	MakeLower(Name);
	
	string::size_type index;
	
	// remove slashes
	while((index = Name.find('/')) != string::npos)
		Name.erase(index, 1);
	// remove backslashes
	while((index = Name.find('\\')) != string::npos)
		Name.erase(index, 1);
	// remove points
	while((index = Name.find('.')) != string::npos)
		Name.erase(index, 1);

	if(!Name.length())
		return false;

	m_sName = Name;
	return true;
}


bool NvLocalContainer::SetName(const char *Name)
{
	if(Name == NULL)
		return false;
	string NameHere = Name;
	return SetName(NameHere);
}



#ifdef NV_SERVER
void NvLocalContainer::EncodeName(bool bSrcUnicode, bool bDestUnicode)
{
	EncodeString(m_sName, bSrcUnicode, bDestUnicode); 
}
#endif // NV_SERVER


void NvLocalContainer::SetRootDirectory(const char *RootDirectory)
{
	if(RootDirectory != NULL)
		m_sRootDirectory = RootDirectory;
	CleanString(m_sRootDirectory);

	AddDirSuffix(m_sRootDirectory);

}



string NvLocalContainer::GetObjectFile(const char *Name, int Type)
{
	string RetVal = "";
	if(!m_sRootDirectory.length())
		return RetVal;
	if(Name == NULL)
		return RetVal;

	NvObj ProviObj(Name);
	if(!ProviObj.GetName().length())
		return RetVal;

	RetVal = m_sRootDirectory;
	AddDirSuffix(RetVal);
	RetVal += ProviObj.GetName();
	char Buf[50];
	sprintf(Buf, "_nv%02d.obj", Type);
	RetVal += Buf;

	return RetVal;
}


void NvLocalContainer::Clear()
{
	ClearObjects();
}

void NvLocalContainer::ClearObjects(bool WithFiles)
{
//	int i;

	map<string,void *>::iterator pos;
	for(pos = m_mObjects.begin(); pos != m_mObjects.end(); ++pos)
	{
		NvObj *Object = (NvObj *)(pos->second);
		if(Object != NULL)
		{
			string sObjName = Object->GetName();
			int iObjType = Object->GetType();
			switch(iObjType)
			{
				case NV_OBJTYPE_BOOLEAN:
					delete ((NvObjBoolean *)Object);
					break;
				case NV_OBJTYPE_INTEGER:
					delete ((NvObjInteger *)Object);
					break;
				case NV_OBJTYPE_STRING:
					delete ((NvObjString *)Object);
					break;
				case NV_OBJTYPE_STRINGLIST:
					delete ((NvObjStringList *)Object);
					break;
				case NV_OBJTYPE_INDSTRINGLIST:
					delete ((NvObjIndStringList *)Object);
					break;
				case NV_OBJTYPE_TABLE:
					delete ((NvObjTable *)Object);
					break;
				case NV_OBJTYPE_FILE:
					delete ((NvObjFile *)Object);
					break;
				case NV_OBJTYPE_BINARY:
					delete ((NvObjBinary *)Object);
					break;
				default:
					delete Object;
					break;
			}
			if(WithFiles)
			{
				// For a registry container, we must also remove the corresponding file
				string FileName = GetObjectFile(sObjName.c_str(), iObjType);
				if(FileName.length())
					RemoveFile(FileName.c_str());
			}
		}
	}
	m_mObjects.clear();
	if(bUseLastObject)
		LastObject = NULL;
}


// Put data into the object
// The content of data depends of the object itself
bool NvLocalContainer::PopulateObject(const char *Name, char *Data, unsigned long DataSize)
{
	if(Name == NULL)
		return false;

	if(!strlen(Name))
		return false;

	NvObj *Object = GetObject(Name);
	
	if(Object == NULL)
		return false;	// The object doesn't exists

	bool RetVal = true;

	if(Object->GetType() == NV_OBJTYPE_BOOLEAN)
	{
		NvObjBoolean *ObjectHere = (NvObjBoolean *)Object;
		RetVal = ObjectHere->Populate(Data, DataSize);
	}
	if(Object->GetType() == NV_OBJTYPE_INTEGER)
	{
		NvObjInteger *ObjectHere = (NvObjInteger *)Object;
		RetVal = ObjectHere->Populate(Data, DataSize);
	}
	if(Object->GetType() == NV_OBJTYPE_STRING)
	{
		NvObjString *ObjectHere = (NvObjString *)Object;
		RetVal = ObjectHere->Populate(Data, DataSize);
	}
	if(Object->GetType() == NV_OBJTYPE_STRINGLIST)
	{
		NvObjStringList *ObjectHere = (NvObjStringList *)Object;
		RetVal = ObjectHere->Populate(Data, DataSize);
	}
	if(Object->GetType() == NV_OBJTYPE_INDSTRINGLIST)
	{
		NvObjIndStringList *ObjectHere = (NvObjIndStringList *)Object;
		RetVal = ObjectHere->Populate(Data, DataSize);
	}
	if(Object->GetType() == NV_OBJTYPE_TABLE)
	{
		NvObjTable *ObjectHere = (NvObjTable *)Object;
		RetVal = ObjectHere->Populate(Data, DataSize);
	}
	if(Object->GetType() == NV_OBJTYPE_FILE)
	{
		NvObjFile *ObjectHere = (NvObjFile *)Object;
		RetVal = ObjectHere->Populate(Data, DataSize);
	}
	if(Object->GetType() == NV_OBJTYPE_BINARY)
	{
		NvObjBinary *ObjectHere = (NvObjBinary *)Object;
		RetVal = ObjectHere->Populate(Data, DataSize);
	}

	return RetVal;
}


// Put data into the object from file
// The content of data depends of the object itself
bool NvLocalContainer::PopulateObjectFromFile(const char *Name, const char *FileName, bool bUseDecodeFromFile)
{
	if(Name == NULL)
		return false;

	if(!strlen(Name))
		return false;

	NvObj *Object = GetObject(Name);
	
	if(Object == NULL)
		return false;	// The object doesn't exists

	bool RetVal = true;

	if(Object->GetType() == NV_OBJTYPE_BOOLEAN)
	{
		NvObjBoolean *ObjectHere = (NvObjBoolean *)Object;
		RetVal = ObjectHere->LoadFromFile(FileName, bUseDecodeFromFile);
	}
	if(Object->GetType() == NV_OBJTYPE_INTEGER)
	{
		NvObjInteger *ObjectHere = (NvObjInteger *)Object;
		RetVal = ObjectHere->LoadFromFile(FileName, bUseDecodeFromFile);
	}
	if(Object->GetType() == NV_OBJTYPE_STRING)
	{
		NvObjString *ObjectHere = (NvObjString *)Object;
		RetVal = ObjectHere->LoadFromFile(FileName, bUseDecodeFromFile);
	}
	if(Object->GetType() == NV_OBJTYPE_STRINGLIST)
	{
		NvObjStringList *ObjectHere = (NvObjStringList *)Object;
		RetVal = ObjectHere->LoadFromFile(FileName, bUseDecodeFromFile);
	}
	if(Object->GetType() == NV_OBJTYPE_INDSTRINGLIST)
	{
		NvObjIndStringList *ObjectHere = (NvObjIndStringList *)Object;
		RetVal = ObjectHere->LoadFromFile(FileName, bUseDecodeFromFile);
	}
	if(Object->GetType() == NV_OBJTYPE_TABLE)
	{
		NvObjTable *ObjectHere = (NvObjTable *)Object;
		RetVal = ObjectHere->LoadFromFile(FileName, bUseDecodeFromFile);
	}
	if(Object->GetType() == NV_OBJTYPE_FILE)
	{
		NvObjFile *ObjectHere = (NvObjFile *)Object;
		RetVal = ObjectHere->LoadFromFile(FileName);
	}
	if(Object->GetType() == NV_OBJTYPE_BINARY)
	{
		NvObjBinary *ObjectHere = (NvObjBinary *)Object;
		RetVal = ObjectHere->LoadFromFile(FileName);
	}

	return RetVal;
}


// Take the object data and put it into data
// The content of data depends of the object itself
bool NvLocalContainer::GetObjectData(const char *Name, char **Data, unsigned long *DataSize)
{
	*Data = NULL;

	if(Name == NULL)
		return false;

	if(!strlen(Name))
		return false;

	NvObj *Object = GetObject(Name);
	
	if(Object == NULL)
		return false;	// The object doesn't exists

	bool RetVal = false;

	if(Object->GetType() == NV_OBJTYPE_BOOLEAN)
	{
		NvObjBoolean *ObjectHere = (NvObjBoolean *)Object;
		RetVal = ObjectHere->GetData(Data, DataSize);
	}
	if(Object->GetType() == NV_OBJTYPE_INTEGER)
	{
		NvObjInteger *ObjectHere = (NvObjInteger *)Object;
		RetVal = ObjectHere->GetData(Data, DataSize);
	}
	if(Object->GetType() == NV_OBJTYPE_STRING)
	{
		NvObjString *ObjectHere = (NvObjString *)Object;
		RetVal = ObjectHere->GetData(Data, DataSize);
	}
	if(Object->GetType() == NV_OBJTYPE_STRINGLIST)
	{
		NvObjStringList *ObjectHere = (NvObjStringList *)Object;
		RetVal = ObjectHere->GetData(Data, DataSize);
	}
	if(Object->GetType() == NV_OBJTYPE_INDSTRINGLIST)
	{
		NvObjIndStringList *ObjectHere = (NvObjIndStringList *)Object;
		RetVal = ObjectHere->GetData(Data, DataSize);
	}
	if(Object->GetType() == NV_OBJTYPE_TABLE)
	{
		NvObjTable *ObjectHere = (NvObjTable *)Object;
		RetVal = ObjectHere->GetData(Data, DataSize);
	}
	if(Object->GetType() == NV_OBJTYPE_FILE)
	{
		NvObjFile *ObjectHere = (NvObjFile *)Object;
		RetVal = ObjectHere->GetData(Data, DataSize);
	}
	if(Object->GetType() == NV_OBJTYPE_BINARY)
	{
		NvObjBinary *ObjectHere = (NvObjBinary *)Object;
		RetVal = ObjectHere->GetData(Data, DataSize);
	}

	return RetVal;
}




// Creates the given object if it doesn't exist
bool NvLocalContainer::CreateObject(const char *Name, int Type)
{
	if(Name == NULL)
		return false;

	if(!strlen(Name))
		return false;

	if(GetObject(Name) != NULL)
		return true;	// The object ever exists

	NvObj *Object = NULL;
	
	switch(Type)
	{
		case NV_OBJTYPE_BOOLEAN:
			Object = NVNEW NvObjBoolean;
			break;
		case NV_OBJTYPE_INTEGER:
			Object = NVNEW NvObjInteger;
			break;
		case NV_OBJTYPE_STRING:
			Object = NVNEW NvObjString;
			break;
		case NV_OBJTYPE_STRINGLIST:
			Object = NVNEW NvObjStringList;
			break;
		case NV_OBJTYPE_INDSTRINGLIST:
			Object = NVNEW NvObjIndStringList;
			break;
		case NV_OBJTYPE_TABLE:
			Object = NVNEW NvObjTable;
			break;
		case NV_OBJTYPE_FILE:
			Object = NVNEW NvObjFile;
			break;
		case NV_OBJTYPE_BINARY:
			Object = NVNEW NvObjBinary;
			break;
		default:
			Object = NVNEW NvObj;
			break;
	}

	if(Object == NULL)
		return false;

	// Set the garbage collector for file objects
	if(Type == NV_OBJTYPE_FILE)
	{
		NvObjFile *ObjFile = (NvObjFile *)Object;
		ObjFile->Garbage = GetMain()->Garbage;
	}


	if(!Object->SetName(Name))
	{
		switch(Object->GetType())
		{
			case NV_OBJTYPE_BOOLEAN:
				delete ((NvObjBoolean *)Object);
				break;
			case NV_OBJTYPE_INTEGER:
				delete ((NvObjInteger *)Object);
				break;
			case NV_OBJTYPE_STRING:
				delete ((NvObjString *)Object);
				break;
			case NV_OBJTYPE_STRINGLIST:
				delete ((NvObjStringList *)Object);
				break;
			case NV_OBJTYPE_INDSTRINGLIST:
				delete ((NvObjIndStringList *)Object);
				break;
			case NV_OBJTYPE_TABLE:
				delete ((NvObjTable *)Object);
				break;
			case NV_OBJTYPE_FILE:
				delete ((NvObjFile *)Object);
				break;
			case NV_OBJTYPE_BINARY:
				delete ((NvObjBinary *)Object);
				break;
			default:
				delete Object;
				break;
		}
		return false;
	}

	m_mObjects[Object->GetName()] = (void *)(Object);
	if(bUseLastObject)
		LastObject = Object;
	return true;
}




NvObj* NvLocalContainer::GetObject(const char *Name)
{
	if(Name == NULL)
		return NULL;

	if(!strlen(Name))
		return NULL;

	string NameHere = Name;
	NvObj::FormatObjectName(NameHere);
	if(!NameHere.length())
		return NULL;

	// Last object cache mechanism
	if((bUseLastObject)&& (LastObject != NULL) && (LastObject->GetName() == NameHere))
		return LastObject;


	NvObj *Object = NULL;
	if(m_mObjects.find(NameHere) != m_mObjects.end())
		Object = (NvObj *)(m_mObjects[NameHere]);

	if(Object == NULL)
	{
		// The object doesn't exist
		// For a registry container, we try to see if the corresponding file exist
		string ChildRootDirectory = m_sRootDirectory;
		AddDirSuffix(ChildRootDirectory);
		if(!ChildRootDirectory.length())	// Should be always the case
			return NULL;

		NvObj ProviObj(NameHere);
		string FileToSearch = ProviObj.GetName();
		FileToSearch += "_nv";
		string FileName = GetFirstFileDir(ChildRootDirectory.c_str(), FileToSearch.c_str());
		if(!FileName.length())
			return NULL;

		// A file has been found
		// get its type
		// And verify several things
		// The file name must be of the form objname_nvdd.obj
		// where dd is the file type on 2 digits
		string::size_type index = FileName.rfind("_nv");
		if(index == string::npos)
			return NULL;
		string sType = FileName.substr(index+3);
		index = sType.rfind(".obj");
		if(index == string::npos)
			return NULL;
		sType = sType.substr(0, index);
		if(sType.length() != 2)
			return NULL;
		int iType = atoi(sType.c_str());
		if(iType < NV_OBJTYPE_UNKNOWN)
			return NULL;
		if(iType > NV_OBJTYPE_MAXTYPEVALUE)
			return NULL;

		NvObj *NewObject = NULL;
		
		switch(iType)
		{
			case NV_OBJTYPE_BOOLEAN:
				NewObject = NVNEW NvObjBoolean;
				if(NewObject == NULL)
					return NULL;
				if(!NewObject->SetName(NameHere))
				{
					delete ((NvObjBoolean *)NewObject);
					return NULL;
				}
				if(!((NvObjBoolean *)NewObject)->LoadFromFile(FileName.c_str(), true))
				{
					delete ((NvObjBinary *)NewObject);
					return NULL;
				}
				break;
			case NV_OBJTYPE_INTEGER:
				NewObject = NVNEW NvObjInteger;
				if(NewObject == NULL)
					return NULL;
				if(!NewObject->SetName(NameHere))
				{
					delete ((NvObjInteger *)NewObject);
					return NULL;
				}
				if(!((NvObjInteger *)NewObject)->LoadFromFile(FileName.c_str(), true))
				{
					delete ((NvObjBinary *)NewObject);
					return NULL;
				}
				break;
			case NV_OBJTYPE_STRING:
				NewObject = NVNEW NvObjString;
				if(NewObject == NULL)
					return NULL;
				if(!NewObject->SetName(NameHere))
				{
					delete ((NvObjString *)NewObject);
					return NULL;
				}
				if(!((NvObjString *)NewObject)->LoadFromFile(FileName.c_str(), true))
				{
					delete ((NvObjBinary *)NewObject);
					return NULL;
				}
				break;
			case NV_OBJTYPE_STRINGLIST:
				NewObject = NVNEW NvObjStringList;
				if(NewObject == NULL)
					return NULL;
				if(!NewObject->SetName(NameHere))
				{
					delete ((NvObjStringList *)NewObject);
					return NULL;
				}
				if(!((NvObjStringList *)NewObject)->LoadFromFile(FileName.c_str(), true))
				{
					delete ((NvObjBinary *)NewObject);
					return NULL;
				}
				break;
			case NV_OBJTYPE_INDSTRINGLIST:
				NewObject = NVNEW NvObjIndStringList;
				if(NewObject == NULL)
					return NULL;
				if(!NewObject->SetName(NameHere))
				{
					delete ((NvObjIndStringList *)NewObject);
					return NULL;
				}
				if(!((NvObjIndStringList *)NewObject)->LoadFromFile(FileName.c_str(), true))
				{
					delete ((NvObjBinary *)NewObject);
					return NULL;
				}
				break;
			case NV_OBJTYPE_TABLE:
				NewObject = NVNEW NvObjTable;
				if(NewObject == NULL)
					return NULL;
				if(!NewObject->SetName(NameHere))
				{
					delete ((NvObjTable *)NewObject);
					return NULL;
				}
				if(!((NvObjTable *)NewObject)->LoadFromFile(FileName.c_str(), true))
				{
					delete ((NvObjBinary *)NewObject);
					return NULL;
				}
				break;
			case NV_OBJTYPE_FILE:
				NewObject = NVNEW NvObjFile;
				if(NewObject == NULL)
					return NULL;
				((NvObjFile *)NewObject)->SetPersistTime();  // a registry file object is always persistent
				if(!NewObject->SetName(NameHere))
				{
					delete ((NvObjFile *)NewObject);
					return NULL;
				}
				// The file object directly points to the registry file
				((NvObjFile *)NewObject)->SetFileName(FileName);
				break;
			case NV_OBJTYPE_BINARY:
				NewObject = NVNEW NvObjBinary;
				if(NewObject == NULL)
					return NULL;
				if(!NewObject->SetName(NameHere))
				{
					delete ((NvObjBinary *)NewObject);
					return NULL;
				}
				if(!((NvObjBinary *)NewObject)->LoadFromFile(FileName.c_str()))
				{
					delete ((NvObjBinary *)NewObject);
					return NULL;
				}
				break;
			default:
				return NULL;
				break;
		}
		// Now, the object has been loaded from file correctly
		// so we add it to the list of objects
		m_mObjects[NewObject->GetName()] = (void *)NewObject;
		Object = NewObject;
	}

	if((bUseLastObject) && (Object != NULL))
		LastObject = Object;
	return Object;
}


void NvLocalContainer::RemoveObject(const char *Name)
{
	if(Name == NULL)
		return;

	if(!strlen(Name))
		return;

	string NameHere = Name;
	NvObj::FormatObjectName(NameHere);
	if(!NameHere.length())
		return;

	NvObj *Object = NULL;
	if(m_mObjects.find(NameHere) != m_mObjects.end())
		Object = (NvObj *)(m_mObjects[NameHere]);
	if(Object != NULL)
	{
		m_mObjects.erase(NameHere);
		int iObjType = Object->GetType();
		switch(iObjType)
		{
			case NV_OBJTYPE_BOOLEAN:
				delete ((NvObjBoolean *)Object);
				break;
			case NV_OBJTYPE_INTEGER:
				delete ((NvObjInteger *)Object);
				break;
			case NV_OBJTYPE_STRING:
				delete ((NvObjString *)Object);
				break;
			case NV_OBJTYPE_STRINGLIST:
				delete ((NvObjStringList *)Object);
				break;
			case NV_OBJTYPE_INDSTRINGLIST:
				delete ((NvObjIndStringList *)Object);
				break;
			case NV_OBJTYPE_TABLE:
				delete ((NvObjTable *)Object);
				break;
			case NV_OBJTYPE_FILE:
				delete ((NvObjFile *)Object);
				break;
			case NV_OBJTYPE_BINARY:
				delete ((NvObjBinary *)Object);
				break;
			default:
				delete Object;
				break;
		}
		
		// For a registry container, we must also remove the corresponding file
		string FileName = GetObjectFile(NameHere.c_str(), iObjType);
		if(FileName.length())
			RemoveFile(FileName.c_str());

		// The last object must be reset if it's the deleted one
		if((bUseLastObject) && (Object != NULL) && (Object == LastObject))
			LastObject = NULL;
		
	}

	if(Object == NULL)
	{
		// if not found, we must verify if a file object doesn't exist for a registry container
		
		// The object doesn't exist
		// For a registry container, we try to see if the corresponding file exist
		string ChildRootDirectory = m_sRootDirectory;
		AddDirSuffix(ChildRootDirectory);
		if(!ChildRootDirectory.length())
			return;

		NvObj ProviObj(NameHere);
		string FileToSearch = ProviObj.GetName();
		FileToSearch += "_nv";
		string FileName = GetFirstFileDir(ChildRootDirectory.c_str(), FileToSearch.c_str());
		if(!FileName.length())
			return;

		// A file has been found
		// get its type
		// And verify several things
		// The file name must be of the form objname_nvdd.obj
		// where dd is the file type on 2 digits
		string::size_type index = FileName.rfind("_nv");
		if(index == string::npos)
			return;
		string sType = FileName.substr(index+3);
		index = sType.rfind(".obj");
		if(index == string::npos)
			return;
		sType = sType.substr(0, index);
		if(sType.length() != 2)
			return;
		int iType = atoi(sType.c_str());
		if(iType < NV_OBJTYPE_UNKNOWN)
			return;
		if(iType > NV_OBJTYPE_MAXTYPEVALUE)
			return;

		// The file is OK
		// We remove it
		RemoveFile(FileName.c_str());
	}
}


/////////////////////////////////////////////////////////////////////////////
// NvRegistryContainerList

NvRegistryContainerList::NvRegistryContainerList()
{
}

NvRegistryContainerList::~NvRegistryContainerList()
{
	Clear();
}



void NvRegistryContainerList::Clear()
{
	NV_LOCKOBJECT(this);
	map<string,NvRegistryContainer *>::iterator pos;
	for(pos = m_mRegistries.begin(); pos != m_mRegistries.end(); ++pos)
	{
		NvRegistryContainer *Registry = (NvRegistryContainer *)(pos->second);
		if(Registry != NULL)
			delete Registry;
	}
	m_mRegistries.clear();
	m_mDirectories.clear();
	NV_UNLOCKOBJECT(this);
}


string NvRegistryContainerList::CleanRegistryName(string Name)
{
	CleanString(Name);
	MakeLower(Name);

	if(!Name.size())
		return "";

	return Name;
}


NvRegistryContainer* NvRegistryContainerList::OpenRegistry(string Name, string Directory, bool Create)
{
	CleanRegistryName(Name);

	if(!Name.length())
		return NULL;

	CleanString(Directory);
	SetDirName(Directory);
	if(!Directory.length())
		return NULL;
	AddDirSuffix(Directory);
	string DirectoryHere = Directory;
	MakeLower(DirectoryHere);

	NvRegistryContainer *Registry;
	NV_LOCKOBJECT(this);

	if(m_mRegistries.find(Name) != m_mRegistries.end())
	{
		// The registry is ever opened so we increase its in use counter and return it
		// but before that, we check if the directory name is the same
		string NameInList = "";
		if(m_mDirectories.find(DirectoryHere) != m_mDirectories.end())
			NameInList = m_mDirectories[DirectoryHere];
		if(NameInList.length() && (NameInList != Name))
		{
			// A registry with another name points to the same directory
			// This is forbidden
			NV_UNLOCKOBJECT(this);
			return NULL;

		}
		Registry = m_mRegistries[Name];
		if(Registry->bRemoveDirRequest)
		{
			// An other user requested to remove the registry directory
			// So we do not authorize opening in this case
			NV_UNLOCKOBJECT(this);
			return NULL;
		}
		Registry->InUse++;
		NV_UNLOCKOBJECT(this);
		return Registry;
	}

	// The registry object doesn't exist
	// so we create it if requested
	if(!Create)
	{
		// No creation required
		NV_UNLOCKOBJECT(this);
		return NULL;

	}

	if(m_mDirectories.find(DirectoryHere) != m_mDirectories.end())
	{
		// A registry exists that points to the same directory
		// This is forbidden
		NV_UNLOCKOBJECT(this);
		return NULL;

	}


	// This directory must also exist so we create it if not existing
	if(!DirExist(Directory.c_str()))
	{
		if(!CreateDir(Directory.c_str(), false))
		{
			NV_UNLOCKOBJECT(this);
			return NULL;
		}
	}


	// We create a new registry entry
	Registry = new NvRegistryContainer;
	Registry->SetRootDirectory(Directory.c_str());
	Registry->InUse++;

	m_mRegistries[Name] = Registry;
	m_mDirectories[DirectoryHere] = Name;

	NV_UNLOCKOBJECT(this);

	return Registry;
}


// Get an opened registry
NvRegistryContainer* NvRegistryContainerList::GetRegistry(string Name)
{
	CleanRegistryName(Name);

	if(!Name.length())
		return NULL;

	NvRegistryContainer *Registry;
	NV_LOCKOBJECT(this);

	if(m_mRegistries.find(Name) == m_mRegistries.end())
	{
		// The registry is not opened
		NV_UNLOCKOBJECT(this);
		return NULL;
	}

	Registry = m_mRegistries[Name];

	NV_UNLOCKOBJECT(this);

	return Registry;
}


bool NvRegistryContainerList::CloseRegistry(string Name, bool Remove, bool RemoveDir, int WaitTime)
{
	CleanRegistryName(Name);

	if(!Name.length())
		return true;

	NV_LOCKOBJECT(this);
	NvRegistryContainer *Registry = NULL;

	if(m_mRegistries.find(Name) == m_mRegistries.end())
	{
		NV_UNLOCKOBJECT(this);
		return true;
	}

	Registry = m_mRegistries[Name];
	if(Registry->InUse)
		Registry->InUse--;
	if(!Registry->InUse)
	{
		// The registry must be removed
		string Directory = Registry->m_sRootDirectory;
		CleanString(Directory);
		SetDirName(Directory);
		AddDirSuffix(Directory);
		MakeLower(Directory);
		if(m_mDirectories.find(Directory) != m_mDirectories.end())
			m_mDirectories.erase(Directory);
		if(Remove && (!Registry->bRemoveDirRequest))
			::RemoveDir(Directory.c_str(), RemoveDir);
		delete Registry;
		Registry = NULL;
		m_mRegistries.erase(Name);
		NV_UNLOCKOBJECT(this);
		return true;
	}

	if(!Remove)
	{
		NV_UNLOCKOBJECT(this);
		return true;
	}

	if(Registry->bRemoveDirRequest)
	{
		// Somebody else request to remove the directory
		NV_UNLOCKOBJECT(this);
		return true;
	}

	// We must physically remove the registry directory
	// First wait for the registry to not be in use

	string Directory = Registry->m_sRootDirectory;
	CleanString(Directory);
	SetDirName(Directory);
	AddDirSuffix(Directory);
	MakeLower(Directory);

	time_t initial_time = time(NULL);

	while(1)
	{
		if(m_mDirectories.find(Directory) != m_mDirectories.end())
		{
			NV_UNLOCKOBJECT(this);
			SLEEP(10);
			NV_LOCKOBJECT(this);
			if((int)(time(NULL) - initial_time) > WaitTime)
				break;	// TIme out
			continue;
		}
	}
	if(m_mDirectories.find(Directory) != m_mDirectories.end())
	{
		// Still in use so we cannot remove it
		// We reset it to an open state for us
		if(m_mRegistries.find(Name) != m_mRegistries.end())
		{
			Registry = m_mRegistries[Name];
			Registry->InUse++;
		}
		NV_UNLOCKOBJECT(this);
		return false;
	}
	::RemoveDir(Directory.c_str(), RemoveDir);

	NV_UNLOCKOBJECT(this);

	return true;
}
