// ***************************************************************************************
// File: garbage.cpp
// Description: file garbage collector
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************


#ifdef NV_SERVER
#include "nvs.h"
#else
#include "../nvc/nvcc.h"
#endif
#include "garbage.h"

#ifdef NV_SERVER
	#ifdef WIN32
		#ifdef _DEBUG
			#include <stdlib.h>
			#include <crtdbg.h>
			#define NVNEW new( _NORMAL_BLOCK, __FILE__, __LINE__)
		#else // _DEBUG
			#define NVNEW new
		#endif // _DEBUG
	#else	//WIN32
		#define NVNEW new
	#endif	//WIN32
#else	//NV_SERVER
	#define NVNEW new
#endif	//NV_SERVER

using namespace std;



/////////////////////////////////////////////////////////////////////////////
// NvGarbageEntry

NvGarbageEntry::NvGarbageEntry()
{
	m_sFileName = "";
	m_iMaxSeconds = 0;
	m_InitialTime = time(NULL);
}

NvGarbageEntry::~NvGarbageEntry()
{
}



/////////////////////////////////////////////////////////////////////////////
// NvGarbage

NvGarbage::NvGarbage()
{
}

NvGarbage::~NvGarbage()
{
	Clear();
}

void NvGarbage::Clear()
{
	NV_LOCKOBJECT(this);
	NvGarbageEntry *Entry;
	int i;
	for(i=0; i<(int)Entries.size(); i++)
	{
		Entry = (NvGarbageEntry*)(Entries[i]);
		unlink(Entry->m_sFileName.c_str());
		delete Entry;
	}
	Entries.clear();
	NV_UNLOCKOBJECT(this);
}

NvGarbageEntry* NvGarbage::GetEntry(std::string FileName)
{
	CleanString(FileName);
	int i=0;
	NvGarbageEntry *RetVal = NULL;
	NvGarbageEntry *Entry;
	for(i=0; i<(int)Entries.size(); i++)
	{
		Entry = (NvGarbageEntry*)(Entries[i]);
		if(FileName == Entry->m_sFileName)
		{
			RetVal = Entry;
			break;
		}
	}
	return RetVal;
}

int NvGarbage::GetEntryIndex(std::string FileName)
{
	CleanString(FileName);
	int i=0;
	int RetVal = -1;
	NvGarbageEntry *Entry;
	for(i=0; i<(int)Entries.size(); i++)
	{
		Entry = (NvGarbageEntry*)(Entries[i]);
		if(FileName == Entry->m_sFileName)
		{
			RetVal = i;
			break;
		}
	}
	return RetVal;
}

	
void NvGarbage::AddEntry(std::string FileName, int MaxSecond)
{
	CleanString(FileName);
	if(!FileName.length())
		return;

	if(!FileExist(FileName.c_str()))
		return;

	if(MaxSecond < 0)
		MaxSecond = 0;

	time_t InitialTime = time(NULL);
	NV_LOCKOBJECT(this);
	NvGarbageEntry *Entry = GetEntry(FileName);
	if(Entry != NULL)
	{
		// If the entry ever exists, we just update its initial time
		Entry->m_InitialTime = InitialTime;
		NV_UNLOCKOBJECT(this);
		return;
	}

	Entry = NVNEW NvGarbageEntry;
	if(Entry == NULL)
		return;

	Entry->m_iMaxSeconds = MaxSecond;
	Entry->m_sFileName = FileName;

	// We add the entry
	Entries.push_back((void *)Entry);

	NV_UNLOCKOBJECT(this);

}


// Remove an entry without removing the file
void NvGarbage::RemoveEntry(std::string FileName)
{
	CleanString(FileName);
	if(!FileName.length())
		return;
	NV_LOCKOBJECT(this);
	int Index = GetEntryIndex(FileName);
	if(Index == -1)
	{
		// The entry doesn't exist
		// So we have nothing to do
		NV_UNLOCKOBJECT(this);
		return;
	}

	// We remove the entry
	NvGarbageEntry *Entry = (NvGarbageEntry*)(Entries[Index]);
	delete Entry;
	Entries.erase(Entries.begin() + Index);


	NV_UNLOCKOBJECT(this);
}

void NvGarbage::RemoveExpired()
{
	NV_LOCKOBJECT(this);
	int i=0;
	NvGarbageEntry *Entry;
	vector<void *> NewEntries;
	for(i=0; i<(int)Entries.size(); i++)
	{
		Entry = (NvGarbageEntry*)(Entries[i]);
		time_t ActualTime = time(NULL);
		if((int)(ActualTime - Entry->m_InitialTime) <= Entry->m_iMaxSeconds)
		{
			NewEntries.push_back((void *)Entry);
			continue;
		}

		if(FileExist(Entry->m_sFileName.c_str()))
		{
			if(unlink(Entry->m_sFileName.c_str()) != 0)
			{
				NewEntries.push_back((void *)Entry);
				continue;
			}
			else
				delete Entry;
		}
	}
	
	if(NewEntries.size() != Entries.size())
	{
		Entries.clear();
		for(i=0; i<(int)NewEntries.size(); i++)
			Entries.push_back(NewEntries[i]);
	}
	NewEntries.clear();
	NV_UNLOCKOBJECT(this);
}

bool NvGarbage::IsInGarbage(std::string FileName)
{
	CleanString(FileName);
	if(!FileName.length())
		return false;

	bool RetVal = false;

	NV_LOCKOBJECT(this);
	if(GetEntryIndex(FileName) != -1)
		RetVal = true;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}




