#include <stdlib.h>
#include "lzma.h"
#include "LzmaDec.h"
#include "LzmaEnc.h"
#include "7zFile.h"
#include "../task.h"
#include <string>

using namespace std;

static void *SzAlloc(void *p, size_t size) { p = p; return malloc(size); }
static void SzFree(void *p, void *address) { p = p; free(address); }
static ISzAlloc g_Alloc = { SzAlloc, SzFree };

#define IN_BUF_SIZE (1 << 16)
#define OUT_BUF_SIZE (1 << 16)

#define FOOTER_LZMA "::NIRVA:COMPRESS=LZMA"


static SRes Decode2(CLzmaDec *state, ISeqOutStream *outStream, ISeqInStream *inStream, UInt64 unpackSize)
{
  int thereIsSize = (unpackSize != (UInt64)(Int64)-1);
  Byte inBuf[IN_BUF_SIZE];
  Byte outBuf[OUT_BUF_SIZE];
  size_t inPos = 0, inSize = 0, outPos = 0;
  LzmaDec_Init(state);
  for (;;)
  {
    if (inPos == inSize)
    {
      inSize = IN_BUF_SIZE;
      RINOK(inStream->Read(inStream, inBuf, &inSize));
      inPos = 0;
    }
    {
      SRes res;
      SizeT inProcessed = inSize - inPos;
      SizeT outProcessed = OUT_BUF_SIZE - outPos;
      ELzmaFinishMode finishMode = LZMA_FINISH_ANY;
      ELzmaStatus status;
      if (thereIsSize && outProcessed > unpackSize)
      {
        outProcessed = (SizeT)unpackSize;
        finishMode = LZMA_FINISH_END;
      }
      
      res = LzmaDec_DecodeToBuf(state, outBuf + outPos, &outProcessed,
        inBuf + inPos, &inProcessed, finishMode, &status);
      inPos += inProcessed;
      outPos += outProcessed;
      unpackSize -= outProcessed;
      
      if (outStream)
        if (outStream->Write(outStream, outBuf, outPos) != outPos)
          return SZ_ERROR_WRITE;
        
      outPos = 0;
      
      if (res != SZ_OK || thereIsSize && unpackSize == 0)
        return res;
      
      if (inProcessed == 0 && outProcessed == 0)
      {
        if (thereIsSize || status != LZMA_STATUS_FINISHED_WITH_MARK)
          return SZ_ERROR_DATA;
        return res;
      }
    }
  }
}

static SRes Decode(ISeqOutStream *outStream, ISeqInStream *inStream)
{
  UInt64 unpackSize;
  int i;
  SRes res = 0;

  CLzmaDec state;

  /* header: 5 bytes of LZMA properties and 8 bytes of uncompressed size */
  unsigned char header[LZMA_PROPS_SIZE + 8];

  /* Read and parse header */

  RINOK(SeqInStream_Read(inStream, header, sizeof(header)));

  unpackSize = 0;
  for (i = 0; i < 8; i++)
    unpackSize += (UInt64)header[LZMA_PROPS_SIZE + i] << (i * 8);

  LzmaDec_Construct(&state);
  RINOK(LzmaDec_Allocate(&state, header, LZMA_PROPS_SIZE, &g_Alloc));
  res = Decode2(&state, outStream, inStream, unpackSize);
  LzmaDec_Free(&state, &g_Alloc);
  return res;
}

static SRes Encode(ISeqOutStream *outStream, ISeqInStream *inStream, UInt64 fileSize, char *rs)
{
  CLzmaEncHandle enc;
  SRes res;
  CLzmaEncProps props;

  rs = rs;

  enc = LzmaEnc_Create(&g_Alloc);
  if (enc == 0)
    return SZ_ERROR_MEM;

  LzmaEncProps_Init(&props);
  res = LzmaEnc_SetProps(enc, &props);

  if (res == SZ_OK)
  {
    Byte header[LZMA_PROPS_SIZE + 8];
    size_t headerSize = LZMA_PROPS_SIZE;
    int i;

    res = LzmaEnc_WriteProperties(enc, header, &headerSize);
    for (i = 0; i < 8; i++)
      header[headerSize++] = (Byte)(fileSize >> (8 * i));
    if (outStream->Write(outStream, header, headerSize) != headerSize)
      res = SZ_ERROR_WRITE;
    else
    {
      if (res == SZ_OK)
        res = LzmaEnc_Encode(enc, outStream, inStream, NULL, &g_Alloc, &g_Alloc);
    }
  }
  LzmaEnc_Destroy(enc, &g_Alloc, &g_Alloc);
  return res;
}


bool lzmaCompressDecompress(const char* infile, const char* outfile, bool DecodeMode)
{
	CFileSeqInStream inStream;
	CFileOutStream outStream;
	int res = SZ_OK;
	Bool useOutFile = true;

	FileSeqInStream_CreateVTable(&inStream);
	File_Construct(&inStream.file);

	FileOutStream_CreateVTable(&outStream);
	File_Construct(&outStream.file);

	{
		size_t t4 = sizeof(UInt32);
		size_t t8 = sizeof(UInt64);
		if (t4 != 4 || t8 != 8)
			return false;
	}

	if (InFile_Open(&inStream.file, infile) != 0)
		return false;

	if (OutFile_Open(&outStream.file, outfile) != 0)
	{
		File_Close(&inStream.file);
		return false;
	}

	bool GotException = false;
	try
	{
		if (DecodeMode)
		{
			res = Decode(&outStream.s, useOutFile ? &inStream.s : NULL);
		}
		else
		{
			UInt64 fileSize;
			File_GetLength(&inStream.file, &fileSize);
			char rs;
			res = Encode(&outStream.s, &inStream.s, fileSize, &rs);
			if (res == SZ_OK)
			{
				// Add nirva header at the end of the file in order to distinguish the compression type
				string Footer = FOOTER_LZMA;
				outStream.s.Write(&outStream.s, Footer.c_str(), Footer.length());
			}
		}
	}
	catch(...)
	{
		GotException = true;
	}
	File_Close(&outStream.file);
	File_Close(&inStream.file);

	if ((res != SZ_OK) || GotException)
		return false;

	return true;
}

bool lzmaCompress(const char* infile, const char* outfile)
{
	return lzmaCompressDecompress(infile, outfile, false);
}

bool lzmaDecompress(const char* infile, const char* outfile)
{
	return lzmaCompressDecompress(infile, outfile, true);
}

bool lzmaIsLzmaCompressedByNirva(const char* infile)
{
	FILE *File;
	File = Nvfopen(infile, "rb");
	if(File == NULL)
		return false;
	Nvfseek(File, 0, SEEK_END);
	NvInt64 Length = Nvftell(File);
	string Footer = FOOTER_LZMA;
	int FooterLength = (int)(Footer.length());
	if(Length < (NvInt64)(FooterLength))
	{
		fclose(File);
		return false;
	}
	Nvfseek(File, 0-FooterLength, SEEK_CUR);
	char *Buffer = (char*)malloc(FooterLength+1);
	fread(Buffer, 1, FooterLength, File);
	fclose(File);
	Buffer[FooterLength] = '\0';
	string FooterRead = Buffer;
	if(FooterRead != Footer)
		return false;
	return true;
}
