// ***************************************************************************************
// File: service.h
// Description: external service classes
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************

#ifndef _SERVICE_H
#define _SERVICE_H


// Classes declared in this file
class MyNvRendererInt;
class NvService;
class NvServiceList;

#include "task.h"
#include "util.h"
#include "container.h"
#include "command.h"
#include "nvsext.h"
#include <string>

// service command entry point
typedef int (*NVSCOMMANDENTRY)(void *Command);
// service render entry point
typedef int (*NVSRENDERENTRY)(void *RenderInterface, const char *ServiceName);	// ServiceName is used only for Java and Dotnet internally

// Start modes
#define START_MODE_FLAG	0
#define START_MODE_NOSTART	1
#define START_MODE_START	2

// Languages
#define LANGUAGE_CPP	0
#define LANGUAGE_JAVA	1
#define LANGUAGE_DOTNET	2

/////////////////////////////////////////////////////////////////////////////
// MyNvRendererInt class
// Interface to service renderer
class MyNvRendererInt : public NvRendererInt
{
public:
	char *m_pInputData;
	unsigned long m_uInputDataSize;
	std::string m_sInputEncoding;
	std::string m_sPage;
	char *m_pOutputData;
	unsigned long m_uOutputDataSize;
	std::string m_sReason;
	std::string m_sContentType;
	std::string m_sSessionId;
	std::string m_sAppName;
	int i_Encoding;				// Service encoding

	MyNvRendererInt();
	virtual ~MyNvRendererInt();

	const char* GetInputData();
	unsigned long GetInputDataSize();
	const char* GetInputEncoding();
	const char* GetPageName();
	char* AllocOutputBuffer(unsigned long Size);
	char* GetOutputBuffer();
	unsigned long GetOutputBufferSize();
	const char* GetSessionId();
	const char* GetAppName();
	void SetReason(const char* Reason);
	void SetContentType(const char* ContentType);
	void* GetExtraData(const char* DataName);		// For futur extension
	void FreeInputData();
};


/////////////////////////////////////////////////////////////////////////////
// NvService class
class NvService : public NvObject
{
private:
	int InUse;									// In use counter
	NvMutex ProtectInUse;						// protect locking of service for updating the in use counter
	bool Initialized;
	bool Started;
	HMODULE hLib;								// Handle to the service dynamic library
	NVSCOMMANDENTRY LibCommand;					// for Command function of the service
	bool m_bEnable;								// Service enabled or not
	std::string LastSessionId;					// This is used for multithread management of the service
	bool SetSkeletonFile(NvObjFile *File, bool WithSessions, int Langage, const char *SessionId, const char *GuidProject = NULL, const char *GuidSolution = NULL, const char *GuidTypelib = NULL);		// used for automatic source code generation
	bool CreateServiceDirectories();			// Creates service directories if they don't exist
	bool m_bProtectContainers;					// protect service containers from beeing accessed by somebody else
	std::string m_sVariableIdent;				// Default string for command variable ident (default is #)

public:
	NVSRENDERENTRY LibRender;					// for a renderer service (null otherwise)
	std::string m_sName;						// Service name
	std::string m_sDesc;						// Service description
	std::string m_sPath;						// Service path
	std::string m_sRegistryDir;					// Config directory
	std::string m_sDocDir;						// Doc directory
	std::string m_sFileDir;						// File directory
	std::string m_sLogDir;						// Log directory
	std::string m_sProcDir;						// Procedure directory
	std::string m_sBinDir;						// Procedure directory
	std::string m_sWrootDir;					// Web directory
	bool m_bMultithread;						// Multithread or not
	bool m_bLoadOnStart;						// Load on start flag
	bool m_bIsRenderer;							// Renderer or not
//	NvObjTable ErrorCodes;						// Error codes and descrition of the service
	NvObjIndStringList Sessions;				// List of sessions actually connected to the service
	NvRegistryContainer SrvRegistry;			// Service parameters (registry)
	NvDescriptionContainer m_Info;				// Service information
	NvLogList Logs;								// Service logs
	bool bInInstall;							// In installation process
	NvContainerList	Containers;					// Service container list
	int m_iLanguage;							// Java, Dotnet or C++ service
	NvRegistryContainerList Registries;			// Service user registries
	int m_iEncoding;							// Service encoding (C++ only)
	NvTestSetList TestSets;						// Test sets
		
	NvService();
	virtual ~NvService();

	bool SetParameters(const char *Name, const char *Desc, const char *Path, bool Multithread = true, bool LoadOnStart = false, int Language = LANGUAGE_CPP);
	bool LoadInfo();
	bool CreateSourceSkeleton(bool WithSessions, int Language, const char *SessionId);
	bool Package(const char *DescriptionFile, const char *PackageFile, std::string& Info);
	bool Install(const char *PackageFile, std::string& Info);
	bool Init();
	bool Start();
	bool Stop(int WaitTime = 60);
	bool IsInitialized();
	bool IsStarted();

	bool IncInuse();
	void DecInUse();
	int GetInUse();

	bool Enable(int StartIt = START_MODE_NOSTART);
	bool Disable(const char *SessionId, int WaitTime = 60);
	bool IsEnable();
	bool IsRenderer();
	void SetDisableFlag();
	bool CheckRequiredForApplications();

	std::string GetName();
	std::string GetDesc();
	std::string GetPath();
	std::string GetRealPath(bool bLockIt = false);
	std::string GetRegistryDir();
	std::string GetDocDir();
	std::string GetFileDir();
	std::string GetLogDir();
	std::string GetProcDir();
	std::string GetBinDir();
	std::string GetWrootDir();
	std::string GetVariableIdent();
	static void CleanSrvName(std::string& Name);

	void SetProtectContainers(bool ProtectContainers);
	bool GetProtectContainers();

	// Get sessions ids connected to the service
	bool GetSessions(NvObjStringList *ConnectedSessions);

	// Service command
	bool Command(MyNvServiceCommand *pCommand);

	// Service render (for rendering services)
	bool Render(MyNvRendererInt *RendererInterface, NvSession *Session);
};


/////////////////////////////////////////////////////////////////////////////
// NvServiceList class
class NvServiceList : public NvObject
{
private:
	NvService* InternalGetService(const char *Name);

public:
	std::map<std::string,void *> m_mServices;			// service list
	
	NvServiceList();
	virtual ~NvServiceList();

	// Removes all services from the list
	void Clear();
	
	// Clears service registry caches
	void ClearRegistryCache();
	
	// used to get an existing service
	// The function increments the in use counter of the corresponding service if found and if LockIt is set to true
	NvService* GetService(const char *Name, bool LockIt = true, bool CheckEnable = false);

	// Load the service list from registry
	void Load();

	// Saves the service list to the system registry
	bool Save();

	// Starts services having the load on start flag set
	void LoadOnStart();


	// Creates a new service with the given name
	// Or eventually modify the service parameters if it's not running
	bool CreateService(const char *Name, const char *Desc, const char *Path, bool Multithread = true, bool LoadOnStart = false, int Language = LANGUAGE_CPP, bool FromLoad = false);

	// Creates a skeleton of service source file
	bool CreateSourceSkeleton(const char *Name, bool WithSessions, int Language, const char *SessionId);

	// used to free an existing service
	// The function decrements the in use counter of the corresponding service
	void FreeService(const char *Name);
	
	// Remove the given service from the list
	bool RemoveService(const char *Name, const char *SessionId, int WaitTime = 60);

	// Enable a service (default)
	bool EnableService(const char *Name, int StartIt = START_MODE_NOSTART, bool FromLoad = false);

	// Disable a service (and unload it from memory)
	bool DisableService(const char *Name, const char *SessionId, int WaitTime = 60);

	// Checks if a service exists in the list
	bool ServiceExist(const char *Name);

	// Checks if a service is enabled or not
	bool IsServiceEnable(const char *Name);

	// Checks if a service is a renderer service or not
	bool IsRenderer(const char *Name);

	// Get sessions ids connected to the service
	bool GetSessions(const char *Name, NvObjStringList *ConnectedSessions);

	// Get service info
	bool GetInfo(const char *Name, NvObjTable *ServiceInfo);

	// Get service specific info
	bool GetInfoEx(const char *Name, NvObjIndStringList *ServiceInfo);

	// Package service
	bool Package(const char *Name, const char *DescriptionFile, const char *PackageFile, std::string& Info);

	// Install service
	bool Install(const char *Name, const char *PackageFile, bool bMountIt, std::string& Info);

	// Get all service permissions
	bool GetPermissions(NvObjTable *ServicePermissions);

	// Verify if service is required for some applicaitons
	bool CheckRequiredForApplications(const char *Name);
};




#endif //_SERVICE_H
