//*****************************************************************************
// File: transaction.h
// Description: declarations of the Nirva transaction classes
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************

#ifndef _TRANSACTION_H
#define _TRANSACTION_H


// Classes declared in this file
class NvTransaction;
class NvTransactionList;
class NvTransactionServer;
class NvTransactionClient;

#include "task.h"
#include "util.h"
#include "container.h"
#include "session.h"
#include <string>
#include <list>
#include <set>


// transaction status
#define NV_TRANSACTION_STATUS_NOT_STARTED			0
#define NV_TRANSACTION_STATUS_RUNNING				1
#define NV_TRANSACTION_STATUS_FAILED				2
#define NV_TRANSACTION_STATUS_COMPLETED				3
#define NV_TRANSACTION_STATUS_VALIDATED_OK			4
#define NV_TRANSACTION_STATUS_VALIDATED_KO			5

// transaction start mode
#define NV_TRANSACTION_START_MANUAL					0
#define NV_TRANSACTION_START_AUTO					1
#define NV_TRANSACTION_START_SCHEDULED				2

// transaction validation mode
#define NV_TRANSACTION_VALIDATION_MANUAL			0
#define NV_TRANSACTION_VALIDATION_AUTO				1
#define NV_TRANSACTION_VALIDATION_SCHEDULED			2

// transaction remove mode
#define NV_TRANSACTION_REMOVE_MANUAL				0
#define NV_TRANSACTION_REMOVE_AUTO					1
#define NV_TRANSACTION_REMOVE_SCHEDULED				2


/////////////////////////////////////////////////////////////////////////////
// NvTransaction class
class NvTransaction
{

private:
	int InUse;									// In use counter
	int m_iStatus;								// transaction status
	int m_iStartMode;							// transaction start mode
	int m_iValidateMode;						// transaction validation mode
	int m_iRemoveMode;							// transaction remove mode
	bool m_bSaveInitial;						// save initial context flag
	bool m_bSaveCurrent;						// save current context flag
	bool m_bInQueue;							// set to true when the transaction is in the server queue for processing
	std::string m_sIdent;						// transaction unique identifier
	std::string m_sName;						// Transaction name
	std::string m_sDesc;						// Transaction description
	std::string m_sAppName;						// Name of the application that owns the transaction
	std::string m_sProcTransaction;				// Transaction procedure
	std::string m_sProcValidation;				// Validation procedure
	std::string m_sProcRollbackC;				// Rollback completed procedure
	std::string m_sProcRollbackF;				// Rollback failed procedure
	std::string m_sProcFailed;					// Failed procedure
	std::string m_sProcCompleted;				// Completed procedure
	std::string m_sProcValidatedOk;				// ValidatedOk procedure
	std::string m_sProcValidatedKo;				// ValidatedKo procedure
	time_t tCreationTime;						// Time of transaction creation
	time_t tStartTime;							// Time of last transaction start or restart

public:
		
	NvMutex ProtectInUse;						// protect locking of transaction for updating the in use counter
	NvSession *Session;							// Current session
	NvError Error;								// Error object (used when transaction fails)

	NvTransaction();
	virtual ~NvTransaction();
	
	std::string GetRegistryDir(bool WithIdent = true, bool CreateMode = false);

	void SetTransactionId();
	bool Start(NvCommand *CommandToExecute, bool bAutoRollBack = false);
	bool Restart(NvCommand *CommandToExecute, bool bAutoRollBack = false, bool bFromServer = false);
	bool Rollback(NvCommand *CommandToExecute);
	bool Validate(NvCommand *CommandToExecute);
	bool SetStatus(NvCommand *CommandToExecute, int NewStatus);
	bool GetContainer(NvCommand *CommandToExecute, const char *Context, const char *SourceContainerName, NvContainer *Destination, int AppendMode, const char *Objects, bool WithSubContainers, bool WithData, const char *FileDir, int PersistTime);
	bool GetVariables(NvCommand *CommandToExecute, const char *Context, NvObjIndStringList *Variables);

	void SetError(NvCommand *CommandToExecute);

	bool IncInuse();
	void DecInUse();
	int GetInUse();

	void SetIdent(const char *Ident);
	std::string GetIdent();
	void SetAppName(const char *AppName);
	std::string GetAppName();
	void SetStatus(int Status);
	int GetStatus();
	void SetStartMode(int StartMode);
	int GetStartMode();
	void SetValidateMode(int ValidateMode);
	int GetValidateMode();
	void SetRemoveMode(int RemoveMode);
	int GetRemoveMode();
	void SetSaveInitial(bool SaveIt);
	bool GetSaveInitial();
	void SetSaveCurrent(bool SaveIt);
	bool GetSaveCurrent();
	void SetName(const char *Name);
	std::string GetName();
	void SetDesc(const char *Desc);
	std::string GetDesc();
	void SetProcTransaction(const char *Proc);
	std::string GetProcTransaction();
	void SetProcValidation(const char *Proc);
	std::string GetProcValidation();
	void SetProcRollbackC(const char *Proc);
	std::string GetProcRollbackC();
	void SetProcRollbackF(const char *Proc);
	std::string GetProcRollbackF();
	void SetProcFailed(const char *Proc);
	std::string GetProcFailed();
	void SetProcCompleted(const char *Proc);
	std::string GetProcCompleted();
	void SetProcValidatedOk(const char *Proc);
	std::string GetProcValidatedOk();
	void SetProcValidatedKo(const char *Proc);
	std::string GetProcValidatedKo();
	void SetCreationTime(time_t Time);
	time_t GetCreationTime();
	void SetStartTime(time_t Time);
	time_t GetStartTime();
	void SetInQueue(bool InQueue);
	bool GetInQueue();

	// Load the transaction parameters from registry
	bool Load();

	// Saves the transaction parameters to the application registry
	bool Save();

	// Removes the transaction information from registry
	void Remove();
};


/////////////////////////////////////////////////////////////////////////////
// NvTransactionList class
class NvTransactionList : public NvObject
{
private:
	NvTransaction* InternalGetTransaction(const char *Ident);

public:
	std::map<std::string,void *> m_mTransactions;			// transaction list
	std::string m_sAppName;									// Name of the application that owns the transaction list
	bool bListLoaded;										// Flag to load the transaction list only on request
	
	NvTransactionList();
	virtual ~NvTransactionList();

	// Removes all transactions from the list
	void Clear();
	
	// used to get an existing transaction
	// The function fails if the transaction is in use so only one task at a time can use the transaction
	NvTransaction* GetTransaction(const char *Ident, NvSession *Session = NULL);

	// Load the transaction list from registry
	bool Load();

	// Creates a new transaction with the given parameters
	// returns the transaction identifier
	std::string CreateTransaction(const char *Name, const char *Desc, const char *ProcName);

	// used to free an existing transaction
	// The function decrements the in use counter of the corresponding transaction
	void FreeTransaction(const char *Ident);
	
	// Remove the given transaction from the list if not in use
	bool RemoveTransaction(const char *Ident, int WaitTime = 10);

	// Checks if a transaction exists in the list
	bool TransactionExist(const char *Ident);

	// Get transaction list info with search strategy on name
	bool GetInfo(const char *Name, const char *FromDate, const char *ToDate, NvObjTable *TransactionInfo, const char *Language = NULL);

	// Get transaction specific info
	bool GetInfoEx(const char *Ident, NvObjIndStringList *TransactionInfo, const char *Language = NULL);

	// Get scheduled transactions to process
	bool GetTransactionsToProcess(NvObjStringList *TransactionsToProcess);
};





	
// Transaction server class
class NvTransactionServer : public NvTask
{
// Attributes
public:
	NvEvent AskToTerminate;
	bool b_Initialized;
	int m_iMaxTransactions;
	int m_iCheckEvery;
	std::list<std::string> m_lTransactionsToProcess;		// transaction list to process (FIFO)
	std::set<std::string> m_lTransactionsInProcess;			// transaction in process

	NvTransactionServer();
	void GetTransactionsToProcess();
	bool IsInProcess(std::string tIdent);
	bool AddInProcess(std::string tIdent);
	bool RemoveInProcess(std::string tIdent);
	int GetInProcessSize();
	void SetMaxTransactions(int MaxTransactions);
	int GetMaxTransactions();
	void SetCheckEvery(int CheckEvery);
	int GetCheckEvery();


// Overridable callbacks
protected:
	virtual int Main();
	virtual void OnException(std::string sInfo);
	virtual bool OnInit();
	virtual void OnCleanup();

public:
	void Stop();

};



// Transaction client class
class NvTransactionClient : public NvTask
{
// Attributes
public:
	NvTransactionServer *Server;
	bool b_Initialized;
	std::string m_sTransactionId;		// Transaction identifier
	NvTransaction *m_tTransaction;		// Current transaction
	bool m_bTransactionRunning;			// Used for crash condition testing

	NvTransactionClient();


// Overridable callbacks
protected:
	virtual int Main();
	virtual void OnException(std::string sInfo);
	virtual bool OnInit();
	virtual void OnCleanup();

public:

	// Run the transaction
	bool RunTransaction();
};


#endif //_TRANSACTION_H
